#!/opt/LifeKeeper/bin/perl
########################################################################
#
# Copyright (c) 2019-2020 SIOS Technology Corp
#
####################################
#   sap.pm
####################################
#
# Description: #
# Error Message Range:
#
# Modifications:
#
# Implementation date:
#
# Date:    Change#       By:  Description
# -------- ------------- ---- -------------------------------------------------
# 13Apr11    bug1703     cmr   Initial submit prior to CI
# 11Jul11    bug1841     cmr   fix the default tag for newInit
# 13Jul11    bug1868     cmr   fix the defaults for using NetApp for NFS
# 14Jul11    bug1861     cmr   quiet the message noise
# 14Jul11    bug1875     cmr   stop calling the startsapsrv every status check
# 14Jul11    bug1878     cmr   fix the command returned message
# 21Jul11    bug1868     cmr   add option to skip NAS FS for non NetApp FS
# 26Jul11    bug1936     cmr   make tunable values consistent between ARK versions
# 29Jul11    bug1869     cmr   kill -9 sapstart leaves orphans
# 02Aug11    bug1975     cmr   too much chatter in logs b/c of SAP system cmds
# 03Aug11    bug1968     cmr   takes too long for create (11minutes or more)
# 03Aug11    bug1975     cmr   too much chatter in logs b/c of SAP system cmds
# 09Aug11    bug1978     cmr   nfs hang causes sap hang of lkcheck
# 09Aug11    bug2003     cmr   1868 breaks non-NAS configurations
# 12Aug11    bug2050     cmr   use kill -15 and kill of pid in loop
# 15Aug11    bug2053     cmr   pingnfs use single message in hang check
# 01Dec11    bug2388     cmr   make expertMode a ENV or cli option LKSAPEXPERTMODE
# 19Jan12    bug2463     cmr   changes for cluster_connector
# 12Mar12    bug2588     emm   changes for refactoring connector modules
# 15Mar12    bug2604     emm   Filehandler will chown the file and do a restore -b in fra()
# 02Aug12    bug3046     jef   Fix incorrect regular expression used in CleanUp
# 26Mar13    bug3014     cmr   Fix name lookup for the sap instance
# 13Nov15    bug7170     cmr   Fix namespace polution for SAP
# 17Dec15    bug7206     cmr   Fixes required for SAP certification based on CC testing
# 19Jun17    bug7336     cmr   Fixes required for SAP in LDAP config
# 19Jun17    bug7223     cmr   Fixes required for SAP Web Dispatch
# 13Mar18    bug3260     cmr   Simplify LocalNFSMount and add debug
# 03Apr18    bug7427     cmr   Skip RKS for ERS and if set in tunable
# 03Apr18    bug7434     cmr   StopWait and StartWait for Enterprise Environment
# 09Apr18    bug7406     cmr   startsapsrv before create and checking for service status
# 23Apr18    bug7406     cmr   check for sapcontrol executable
# 01Nov18    bug7527     cmr   remove lock flags on sigHandler call
# 12Feb19    bug7622     emm   Added maintenance mode code
# 14Dec18    bug7523     jbb   Add more info to FlagInProgress header
# 20Feb19    bug7631     cmr   Start and Stop changes
# 05Mar19    bug7134     jef   Limit NFS checks to NFS mounts SAP dirs
# 14Mar19    bug7664     emm   Filter out what to show for SAP instances
# 21Mar19    bug7669     emm   Find out if we are connected to SAP connector manager, also new cpa code
# 28Mar19    bug7678     emm   Change the doc to static link
# 28Mar19    bug7677     ysa   SSCC package name change
# 15May19    bug7693     jbb   Use HAGetFailoverConfig to determine if HA connector is configured
# 05Jun19    bug7631     jbb   Add TYPE_NEW_ERS to handle ERS v1 and v2 in an independent hierarchy
# 14Jun19    bug7723     cew   Add TYPE_ERS, TYPE_NEW_ERS and TYPE_CS to SAPDisplay
# 23Jul19    bug7769     jbb   ENSA2/ERSv2 avoidance in fraNextNode
# 05Aug19    bug7780     cew   Do not hang when NFS mounts not available
# 06Aug19    bug7779     jbb   Updated behavior in init, hcc, and fra when a cluster node is down
# 14Aug19    bug7775     jbb   Add KillInstanceProcesses routine to forcefully stop instance
# 14Aug19    bug7759     PRC   fix pre-9.3.2 rolling upgrade issue
# 20Aug19    bug7781     jbb   Updates to support fra migrate for ENSA2/ER2
# 23Aug19    bug7789     jbb   Fix in-progress flag handling in cpa
# 28Aug19    bug7781     jbb   Call fraNextNode before setting resource OSU in fra migrate
# 28Aug19    bug7799     jbb   Modify checkFileandPid to only check current fra action
# 26Sep19    bug7819     jbb   Append tag to flag names in sigHandler
# 17Jul20    PL-6050     jef   Delay ERS v2 quickCheck/recover if comm_up running
#
################################################################################
#####################--- Start SAP Package Header Section ---###################
################################################################################
BEGIN { require '/etc/default/LifeKeeper.pl'; }
package sap;
use strict;
use LK;
use LKBase;
use Data::Dumper;
################################################################################
#####################--- End SAP Package Header Section ---#####################
################################################################################


################################################################################
####################--- Start SAP Generic Variables Section ---#################
################################################################################

our $appType;
*appType = \('appsuite');
our $resType;
*resType = \('sap');
our $me;
*me = \(LK::lcduname());

(our $cmd = $0 ) =~ s@.*/@@; # get basename
our $lkroot="$ENV{LKROOT}";
our $lkadm="$lkroot/lkadm";
our $APPBIN="$lkadm/subsys/$appType/$resType/bin";
our $APPSTORE="$lkroot/subsys/$appType/resources/$resType";
our $APPACTIONS="$lkroot/subsys/$appType/resources/$resType/actions";
our $APPRECOVER="$lkroot/subsys/$appType/resources/$resType/recovery/sap";
our $FSBINDIR="$lkroot/lkadm/subsys/gen/filesys/bin";
our $NFSBINDIR="$lkroot/lkadm/subsys/gen/nfs/bin";
our $PINGNFS="$lkroot/lkadm/bin/pingnfs";
my $PERL="/etc/default/LifeKeeper-perl";
our $varlookup="$lkroot/lkadm/bin/varlookup";
my $CORE_PKG="steeleye-lk";
my $SSCC_PKG="steeleye-lkHACONNECTOR-for-SAP";
my $SSCC_SCRIPT="sap_sios_cluster_connector";
our $signalTag="unset$$";


our $noise=0;

# sapcontrol: EXITCODES
#0  Last webmethod call successful
#1  Last webmethod call failed, invalid parameter
#2  StartWait, StopWait, WaitforStarted, WaitforStopped, RestartServiceWait timed out
#3  GetProcessList succeeded, all processes running correctly
#4  GetProcessList succeeded, all processes stopped
my $ALLSTARTED=3;
my $ALLSTOPPED=4;

my $BADPID=-5;
our %OS_RET= ( 1 => 'FAILED', 0 => 'SUCCESS' );
my %systoperl=(1 => 0, 0 => 1 );
our $nilValue="_INTERNAL_NO_OP_";

# These defines are for the protection level
# A protection level determines how much of a LifeKeepr
# action this resource is capable of handling
# All, Partial, Basic, or Minimum
# Example:
# Parent-SAPa
#    Child-SAPb
#        Child-SAPc
#           Child-IP
#           Child-NFS
#           Child-FS
# Resource SAPa may have a level of Basic b/c it is redundant but we want
# to log if it fails
# Resource SAPb may have a level of Standard b/c it should be started,
# monitored and recovered, but not stopped an example would be
# the replicated enqueue service
# Resource SAPc may have a level of FULL, meaning it should follow
# all normal LK rules of start, stop, monitor and recovery enabled

our $P_LEVEL_3="FULL";
our $P_LEVEL_2="STANDARD";
our $P_LEVEL_1="BASIC";
our $P_LEVEL_0="MINIMUM";

our %PROTECTION_LEVELS_TEXT = (
        $P_LEVEL_3 => 'Start, Stop, Monitor, Recover',
        $P_LEVEL_2 => 'Start, Monitor, Recover',
        $P_LEVEL_1 => 'Start, Monitor',
        $P_LEVEL_0 => "Start"
);

our %PROTECTION_LEVELS = (
        $P_LEVEL_3 => 3,
        $P_LEVEL_2 => 2,
        $P_LEVEL_1 => 1,
        $P_LEVEL_0 => 0
);

our %XREF_PROTECTION_LEVELS = (
        3 => "$P_LEVEL_3",
        2 => "$P_LEVEL_2",
        1 => "$P_LEVEL_1",
        0  => "$P_LEVEL_0"
);

# These defines are for the recovery levels
# A recovery level determines
# 1. If a resource is important enough to cause a hierarchy to failover
# 2. If the resource should be restarted (local or remote)
# Example:
# Parent-SAPa
#    Child-SAPb
#        Child-SAPc
#           Child-IP
#           Child-NFS
#           Child-FS
# Resource SAPa may have a level of local only, which
# indicates that local recovery should be performed, but the resource
# does not require a failover if local recovery fails.  Examples of this
# type of resource might include partially redundant secondary servers.
# Resource SAPb may have a level of remote only, which indicates that if
# the resource fails it should not be retried locally b/c but cause
# the entire stack to fail to a remote system.
# Resource SAPc may hae a level of Full, which indicates that if the
# resource fails it should be retried locally first, then remote

our $R_LEVEL_3="FULL";
our $R_LEVEL_2="REMOTE";
our $R_LEVEL_1="LOCAL";
our $R_LEVEL_0="OFF";
our %RECOVERY_LEVELS = (
        $R_LEVEL_3 => 3,
        $R_LEVEL_2 => 2,
        $R_LEVEL_1 => 1,
        $R_LEVEL_0 => 0
);

our %XREF_RECOVERY_LEVELS = (
        3 => $R_LEVEL_3,
        2 => $R_LEVEL_2,
        1 => $R_LEVEL_1,
        0 => $R_LEVEL_0
);


# We will create a file to store additional resource information
# values and configuration defaults
# The hash defines all the parameters of the file and their
# default values.  Each Key in the hash serves as the function
# name in sap.pm as well as the parameter it stores
# The array defines values that should be reset in the file on a
# successful restore and/or remove
my %SAP_INFO_FILE_DEFAULTS = (
        "SAPRED" => 1,
        "SAPYELLOW" => 0,
        "SAPRECOVER_ATTEMPTS" => 0,
        "SAPRECOVER_ATTEMPTS_MAX" => 4,
        "SAPRECOVER_SUCCESS" => 1,
        "SAPRECOVER_FAILED" => 0,
        "SAPRECOVER_TIMEPERIOD" => 120,
        "SAPRECOVER_TIMEELAPSED" => 0,
        "SAPRECOVER_CURRENTTIME" => 101200002281976052419761108119905232011,
        "SAPRESTORE_DATE" => -524
);

my @SAP_INFO_FILE_RESET_LIST = (
        "SAPRECOVER_ATTEMPTS",
        "SAPRECOVER_SUCCESS",
        "SAPRECOVER_FAILED",
        "SAPRECOVER_TIMEELAPSED",
        "SAPRESTORE_DATE"
);

################################################################################
####################--- End SAP Generic Variables Section ---###################
################################################################################


################################################################################
########################--- Start SAP varlookup Section ---#####################
################################################################################

chomp (our $PASSWDFILE=`$varlookup /etc/passwd 2>/dev/null`);
chomp (our $SAPDIR=`$varlookup /usr/sap 2>/dev/null`);
chomp (our $FSTAB=`$varlookup /etc/fstab 2>/dev/null`);
chomp (our $ETAB=`$varlookup /var/lib/nfs/etab 2>/dev/null`);
chomp (our $MTAB=`$varlookup /etc/mtab 2>/dev/null`);

our $REDIRECT=">&2";

# Based on CC feedback, trans can also be defined in a non-standard
# location and identified by DEFAULT.PFL value DIR_TRANS
our $SAPTRANSDIR="$SAPDIR/trans";

################################################################################
########################--- End SAP varlookup Section ---#######################
################################################################################

################################################################################
########################--- Start SAP TMP Files Section ---#####################
################################################################################

# lock and status files used during create/extend and actions/*
our $lockfile = "/tmp/lk-lock.SAPLKTMP$$";
our $tmpfile="/tmp/lk-tmp.SAPLKTMP$$";
our $resfile="/tmp/lk-res.SAPLKTMP$$";

# storage files that are used by kit to keep
# resource or SAP info.  The file format is
# <type>_<tag>.
our $INFO_FILE_STORE="$APPSTORE/INFO_";         #
our $AS_FILE_STORE="$APPSTORE/AS_";                     # needed for delete after upgrade
our $REPENQ_FILE_STORE="$APPSTORE/REPENQ_";     # needed for delete after upgrade

################################################################################
##########################--- End SAP TMP Files Section ---#####################
################################################################################

################################################################################
######################--- Start SAP Utility Names Section ---###################
################################################################################

#Mon Jun  6 11:46:45 EDT 2011
# These are some of the tools and server processes
# These are 7.3 specific per installs
our $SAP_START="sapstart";
our $SAP_SRV="sapstartsrv";
our $SAP_HOSTEXEC="saphostexec";
our $SAP_HOSTCTRL="saphostctrl";
our $SAPOSCOL="saposcol";
our $SAP_CONTROL="sapcontrol";
our $DPFL="DEFAULT.PFL";
our $SAP_SVCS="$SAPDIR/sapservices";
our $SAP_HOSTCTL_PATH_DEFAULT="$SAPDIR/hostctrl/exe";
our $SAP_HOSTCTL_PROFILE_DEFAULT="$SAP_HOSTCTL_PATH_DEFAULT/host_profile";
our $SAP_SRVHOST_CMD_DEFAULT="$SAP_HOSTCTL_PATH_DEFAULT/$SAP_HOSTEXEC";
our $SAP_HOSTCTL_CMD_DEFAULT="$SAP_HOSTCTL_PATH_DEFAULT/$SAP_HOSTCTRL";
our $SAP_CONTROL_CMD_DEFAULT="$SAP_CONTROL -format script -prot NI_HTTP";
our $SAP_OSCOL_CMD_DEFAULT="$SAPDIR/hostctrl/exe/$SAPOSCOL";

# Executables and utilities related to enqueue server/replication
our $SAP_ENSERVER="enserver";             # ENSA1 executable
our $SAP_ENQ_SERVER="enq_server";         # ENSA2 executable
our $SAP_ENREPSERVER="enrepserver";       # ERSv1 executable
our $SAP_ENQ_REPLICATOR="enq_replicator"; # ERSv2 executable
our $SAP_ENSMON="ensmon";                 # monitoring utility for ENSA1/ERSv1
our $SAP_ENQ_ADMIN="enq_admin";           # monitoring utility for ENSA2/ERSv2

# Hashes mapping executable names to version numbers, used in GetEnqVersion
our %SAP_ENQ_VER_HASH = (
        $SAP_ENSERVER => 1,
        $SAP_ENQ_SERVER => 2
);

our %SAP_ENQREP_VER_HASH = (
        $SAP_ENREPSERVER => 1,
        $SAP_ENQ_REPLICATOR => 2
);

# Profile parameters used for enabling/disabling enqueue replication
our $SAP_ENSA1_ENQREPENABLED_PARAM = "enque/server/replication"; # ENSA1 profile parameter
our $SAP_ENSA2_ENQREPENABLED_PARAM = "enq/server/replication/enable"; # ENSA2 profile parameter

# Hash mapping ENSA version to profile parameter used to enable/disable
# replication, used in EnqRepEnabled
our %SAP_ENQREPENABLED_PARAM_HASH = (
        1 => $SAP_ENSA1_ENQREPENABLED_PARAM,
        2 => $SAP_ENSA2_ENQREPENABLED_PARAM
);

#Fri May  6 10:56:45 EDT 2011
# To support older version we define tunable values that can
# override the 7.3 defaults
our $SAP_HOSTCTL_PATH=($ENV{SAP_HOSTCTL_PATH} eq '' )? $SAP_HOSTCTL_PATH_DEFAULT:"$ENV{SAP_HOSTCTL_PATH}";
our $SAP_HOSTCTL_PROFILE=($ENV{SAP_HOSTCTL_PROFILE} eq '' )? $SAP_HOSTCTL_PROFILE_DEFAULT:"$ENV{SAP_HOSTCTL_PROFILE}";
our $SAP_SRVHOST_CMD=($ENV{SAP_SRVHOST_CMD} eq '' )? $SAP_SRVHOST_CMD_DEFAULT:"$ENV{SAP_SRVHOST_CMD}";
our $SAP_HOSTCTL_CMD=($ENV{SAP_HOSTCTL_CMD} eq '' )? $SAP_HOSTCTL_CMD_DEFAULT:"$ENV{SAP_HOSTCTL_CMD}";
our $SAP_CONTROL_CMD=($ENV{SAP_CONTROL_CMD} eq '' )? $SAP_CONTROL_CMD_DEFAULT:"$ENV{SAP_CONTROL_CMD}";
our $SAP_OSCOL_CMD=($ENV{SAP_OSCOL_CMD} eq '' )? $SAP_OSCOL_CMD_DEFAULT:"$ENV{SAP_OSCOL_CMD}";

################################################################################
#######################--- End SAP Utility Names Section ---####################
################################################################################



################################################################################
#######################--- Start SAP Log Info Section ---#######################
################################################################################

our $_log;
logset('default');

################################################################################
#######################--- End SAP Log Info Section ---#########################
################################################################################

################################################################################
###################--- Start SAP Default Flags Section ---#################
################################################################################

# Flags for various actions in progress.  The flag format is
# !volatile!SAP_<action>_IN_PROGRESS_<tag>.
our $SAP_FLAG_RESTORE_IN_PROGRESS="!volatile!SAP_RESTORE_IN_PROGRESS_"; # + tag
our $SAP_FLAG_REMOVE_IN_PROGRESS="!volatile!SAP_REMOVE_IN_PROGRESS_";   # + tag
our $SAP_FLAG_QCK_IN_PROGRESS="!volatile!SAP_QUICKCHECK_IN_PROGRESS_";  # + tag
our $SAP_FLAG_RECOVER_IN_PROGRESS="!volatile!SAP_RECOVER_IN_PROGRESS_"; # + tag

# Flag to signal the recover script for a TYPE_NEW_ERS resource to fail
# and initiate failover of the ERS hierarchy
our $SAP_FLAG_MOVE_ERS="!volatile!SAP_MOVE_ERS_"; # + tag

# Flag to disable automatic hierarchy relocation for a TYPE_NEW_ERS resource
our $SAP_FLAG_NO_ERS_RELOCATION="sap_no_ers_relocation_"; # + tag

# Flag to force a TYPE_CS resource with SAPENQ_VERSION=1
# to come in-service on any node even if the corresponding
# ERS instance is running remotely
our $SAP_FLAG_CS_FORCE_RESTORE="sap_cs_force_restore_"; # + tag

# Flag to force a TYPE_CS resource with SAPENQ_VERSION=1
# to locally recover even if the corresponding ERS instance
# is running remotely
our $SAP_FLAG_CS_FORCE_RECOVER="sap_cs_force_recover_"; # + tag

################################################################################
###################--- End SAP Default Flags Section ---#################
################################################################################

################################################################################
###################--- Start SAP Default File Entry Section ---#################
################################################################################

our $REMOTETIMEOUT="$ENV{REMOTETIMEOUT}";
our $LKCHECKTIMEOUT="$ENV{LKCHECKINTERVAL}";

# Tunable values for changing retries, sleep intervals, and action
# settings
our $SAP_CREATE_NAS=($ENV{SAP_CREATE_NAS } =~ /(^\d+)$/ )?$ENV{SAP_CREATE_NAS}:1;

# determine how low in a hierarchy will we search for a common child resource
# Tune lower for a faster, less accurate create.  Tune higher for a more accurate
# but slower create
our $SAP_DEPENDENCY_DEPTH=($ENV{SAP_DEPENDENCY_DEPTH } =~ /(^\d+)$/ )?$ENV{SAP_DEPENDENCY_DEPTH}:7;

our $SAP_EXTERNAL_NFS =($ENV{SAP_EXTERNAL_NFS } =~ /(^\d+)$/ )?$ENV{SAP_EXTERNAL_NFS}:1;
our $SAP_DEBUG =($ENV{SAP_DEBUG } =~ /(^\d+)$/ && $ENV{SAP_DEBUG} <= 5 )?$ENV{SAP_DEBUG}:0;
our $SAP_CONNECTOR_DEBUG =($ENV{SAP_CONNECTOR_DEBUG } =~ /(^\d+)$/ && $ENV{SAP_CONNECTOR_DEBUG} <= 5 )?$ENV{SAP_CONNECTOR_DEBUG}:0;

# if SAP_*DEBUG is turned on, then we need to set the core debug value for LK_DEBUG
# this is a woraround for the namespace polution in bug7170
if ($SAP_DEBUG || $SAP_CONNECTOR_DEBUG ){
        $ENV{'LK_DEBUG'} = 1;
}
our $SAP_MSG_THROTTLE=($ENV{SAP_MSG_THROTTLE} =~ /(^\d+)/ ) ? $ENV{SAP_MSG_THROTTLE} : 25;
our $SAP_SMALL_SLEEP_INTERVAL=($ENV{SAP_SLEEP_INTERVAL} =~ /(^\d+)/ ) ? $ENV{SAP_SLEEP_INTERVAL} : 5;
our $SAP_RKS_SLEEP_INTERVAL=($ENV{SAP_RKS_INTERVAL} =~ /(^\d+)/ ) ? $ENV{SAP_RKS_INTERVAL} : 15;

# determine if we need to customize create or other actions
our $SAP_VMNODEMODE =($ENV{SINGLENODE} =~ /(^\d+)$/)?$ENV{SINGLENODE}:0;
our $SAP_EXPERTMODE =($ENV{SAP_EXPERTMODE } =~ /(^\d+)$/)?$ENV{SAP_EXPERTMODE}:0;
our $internalMode=76;

# SAP TIMEOUTS

our $SAP_STOP_WAIT=($ENV{SAP_STOP_WAIT} =~ /(^\d+)/ ) ? $ENV{SAP_STOP_WAIT} : 228;
our $SAP_START_WAIT=($ENV{SAP_START_WAIT} =~ /(^\d+)/ ) ? $ENV{SAP_START_WAIT} : 228;
our $SAP_KILL_INSTANCE_PROCESSES_WAIT=($ENV{SAP_KILL_INSTANCE_PROCESSES_WAIT} =~ /(^d+)/ ) ? $ENV{SAP_KILL_INSTANCE_PROCESSES_WAIT} : 5;
our $SAP_CONTROL_DELAY=($ENV{SAP_CONTROL_DELAY} =~ /(^\d+)/ ) ? $ENV{SAP_CONTROL_DELAY} : 5;
our $SAP_CONN_RETRIES=($ENV{SAP_CONN_RETRIES} =~ /(^\d+)/ ) ? $ENV{SAP_CONN_RETRIES} : 10;
our $SAP_CLEAN_RETRIES=($ENV{SAP_CLEAN_RETRIES} =~ /(^\d+)/ ) ? $ENV{SAP_CLEAN_RETRIES} : 3;
our $SAP_SRV_RETRIES=($ENV{SAP_SRV_RETRIES} =~ /(^\d+)/ ) ? $ENV{SAP_SRV_RETRIES} : 1;

# ACTION TIMEOUTS
our $SAP_QCK_TIMEOUT=($ENV{SAP_QUICKCHECK_TIMEOUT} =~ /(^\d+)/ && $ENV{SAP_QUICKCHECK_TIMEOUT} > 60 ) ? $ENV{SAP_QUICKCHECK_TIMEOUT} : 60;

my $SAPCALC=(2 * $SAP_START_WAIT) + $SAP_QCK_TIMEOUT;
our $SAP_RESTORE_TIMEOUT=($ENV{SAP_RESTORE_TIMEOUT} =~ /(^\d+)/ ) ? $ENV{SAP_RESTORE_TIMEOUT} : $SAPCALC;

my $SAPCALC=(3 * $SAP_STOP_WAIT) + (2 * $SAP_QCK_TIMEOUT);
our $SAP_REMOVE_TIMEOUT=($ENV{SAP_REMOVE_TIMEOUT} =~ /(^\d+)/ ) ? $ENV{SAP_REMOVE_TIMEOUT} : $SAPCALC;

our $SAPCALC=($SAP_RESTORE_TIMEOUT + $SAP_REMOVE_TIMEOUT);
our $SAP_RECOVER_TIMEOUT=($ENV{SAP_RECOVER_TIMEOUT} =~ /(^\d+)/ && $ENV{SAP_RECOVER_TIMEOUT} > $SAPCALC ) ? $ENV{SAP_RECOVER_TIMEOUT} : $SAPCALC;

# Per CC/ and testing it appears that sapcontrol will issue the start DB commands
# if needed.  Therefore we need to allow varied behavior
our $SAP_DB_CHECK_CRITICAL=($ENV{SAP_DB_CHECK_CRITICAL } =~ /(^\d+)/ ) ? $ENV{SAP_DB_CHECK_CRITICAL } : 0;

# DB/NFS check is not required always, allow override
our $SAP_DB_CHECK_IGNORE=($ENV{SAP_DB_CHECK_IGNORE } =~ /(^\d+)/ ) ? $ENV{SAP_DB_CHECK_IGNORE } : 0;
our $SAP_RKS_CHECK_IGNORE=($ENV{SAP_RKS_CHECK_IGNORE } =~ /(^\d+)/ ) ? $ENV{SAP_RKS_CHECK_IGNORE } : 0;

# Previous kit does stop in recover to force an cleanipc.
# Do this b/c the sapcontrol will not always handle this
our $SAP_STOP_IN_RECOVER=($ENV{SAP_STOP_IN_RECOVER } =~ /(^\d+)/ ) ? $ENV{SAP_STOP_IN_RECOVER } : 1;


my $SAPVHOSTSTRING="SAPLOCALHOST";      # string for looking up the VIP/hostname for the instance
our $SAP_SSCC_DOCS_VERSION=($ENV{SAP_SSCC_DOCS_VERSION} =~ /(^\d+)/ ) ? $ENV{SAP_SSCC_DOCS_VERSION} : 9.0.1;

# NFS remote mounts to check
my $SAP_NFS_CHECK_DIRS=($ENV{SAP_NFS_CHECK_DIRS}) ? $ENV{SAP_NFS_CHECK_DIRS} : "";

# bug7678 - hardcoded static link for docs.
my $SAP_DOCS_PAGE="docs.us.sios.com/Linux/current/LK4L/SAP/index.htm";


################################################################################
#####################--- End SAP Default File Entry Section ---#################
################################################################################


################################################################################
#####################--- Start SAP Hash Variables Section ---###################
################################################################################


# These are all of the keys used by the object.
# Keeping the list allows avoidance of name clashing

my $iThrottle="SAP_CMD_THROTTLE";
my $KEYSAPINRECOVER="SAP_IN_RECOVER";          # key if we are in recover
my $KEYSAPSID="SAP_SID";                                        # key of SAP SID
my $KEYSAPSRVCMD="SAP_SRV_CMD";                         # key of SAP SRV CMD/per sid
my $KEYSAPSRVPF="SAP_SRV_PF";                           # key of instance profile location from sapservices file
my $KEYSAPINSTTYPE="SAP_INST_TYPE";                     # key of SAP Instance type
my $KEYSAPINST="SAP_INST";                                      # key of SAP Instance Name
my $KEYSAPINSTNUM="SAP_INST_NUM";                       # key of SAP Instance Num (derived)
my $KEYSAPREPENQINST="SAP_REP_ENQ_INST";        # key of rep. enq
my $KEYSAPVIP="SAP_VIP";                                        # key of my SAP virtual host name
my $KEYSAPTAG="Tag";                                            # key of my resource tag
my $KEYSAPFLAGS="ACTION_IN_PROGRESS_FLAGS";     # key of my resource ACTION_IN_PROGRESS_FLAGS
my $KEYSAPSMALLWAIT="SAP_SMALL_WAIT";           # key of my resource SMALL SLEEP INTERVAL
my $KEYSAPRKSWAIT="SAP_RKS_WAIT";               # key of my resource RKS SLEEP INTERVAL
my $KEYSAPMNT="SAP_MNT";                                        # key of sap root dir (/sapmnt)
my $KEYSAPPROTECTION="SAP_PROTECTION";          # key of protection level
my $KEYSAPRECOVERY="SAP_RECOVERY";                      # key of recovery level
my $KEYSAPADMIN="SAP_ADMIN";                            # sap admin user
my $KEYSAPADMINHOME="SAP_ADMIN_HOME";           # sap admin user home
my $KEYSAPUID="SAP_UID";                                        # sap admin uid
my $KEYSAPGID="SAP_GID";                                        # sap admin gid
my $KEYSAPSHELL="SAP_SHELL";                            # sap admin shell
my $KEYSAPDBNAME="SAP_DB_NAME";                         # sap instance db name
my $KEYSAPDBTYPE="SAP_DB_TYPE";                         # key of inst db type (ora, db2...)
my $KEYSAPDBHOST="SAP_DB_HOST";                         # sap instance db host
my $KEYSAPDIRTRANS="SAP_TRANS_DIR";             # sap instance trans dir
my $KEYSAPDIR="SAP_DIR";                                        # sap install dir (/usr/sap)
my $KEYSAPCOLLPID="SAP_COLLECTOR_PID";          # key of os collector pid using ps
my $KEYSAPSRVPID="SAP_SRV_PID";                         # key of srv pid using ps
my $KEYSAPHOSTEXECPID="SAP_HOSTEXEC_PID";       # key of hostexec pid using ps
my $KEYSAPLOCALSYS="SAP_LOCAL_SYS";             # key of local sys name
my $KEYSAPACTIONS="SAP_ACTIONS";                        # key of LK actions dir
my $KEYSAPPASSWORD="SAP_PASSWORD_FILE";         # key of password file location
my $KEYSAPLOCKFILE="SAP_LOCK_FILE";             # key of LK lock file
my $KEYSAPFROMTAG="SAP_FROM_TAG";                       # key of internal new source
my $KEYSAPASSTORE="SAP_AS_STORE";                       # key of tmp file of AS srv(s)
my $KEYSAPINFOSTORE="SAP_INFO_STORE";           # key of tmp file of Additional Info
my $KEYSAPREPENQSTORE="SAP_REP_ENQ_STORE";      # key of tmpfile of Rep Enq srv(s)
my $KEYSAPTMPFILE="SAP_TMP_FILE";                       # key of LK tmp file
my $KEYSAPRESULTFILE="SAP_RESULT_FILE";         # key of LK results file
my $KEYSAPCHECK="LK_CHECK_STATUS";                      # key of UI display and qChk hash
my $KEYSAPNAMES="LK_CHECK_STATUS_NAMES";        # key of UI display and qChk hash
my $KEYSAPLKSAPINFO="LK_SAP_INFO";                      # entire resource info field
my $KEYSAPLKSAPID="LK_SAP_ID";                          # key of LK SAP resource ID
my $KEYSAPLKSAPSTATE="LK_SAP_STATE";            # key of LK SAP resource state
my $KEYSAPLKSAPLK="LK_SAP_LK";                          # key of object exists ie {LK}
my $KEYSAPPARENT="LK_CALLING_CMD";                      # key of call command
my $KEYSAPRED="SAP_RED_EVENT";                          # key of SAP Red display state
my $KEYSAPGREEN="SAP_GREEN_EVENT";                              # key of SAP GREEN display state
my $KEYSAPYELLOW="SAP_YELLOW_EVENT";            # key of SAP Yellow display state

# Keys related to the additional INFO file values
#%SAP_INFO_FILE_DEFAULTS.  Updates to SAP_INFO_FILE_DEFAULTS
# should be matched by keys here, and functions below
my $KEYSAPRECOVER_ATTEMPTS="LK_RECOVER_ATTEMPTS";                       # num. recover attempts since restore
my $KEYSAPRECOVER_ATTEMPTS_MAX="LK_RECOVER_ATTEMPTS_MAX";       # max num. recover attempts / timeperiod
my $KEYSAPRECOVER_SUCCESS="LK_RECOVER_SUCCESS";                         # num. of successful recovers
my $KEYSAPRECOVER_FAILED="LK_RECOVER_FAILED";                           # num. of failed recovers
my $KEYSAPRECOVER_TIMEPERIOD="LK_RECOVER_TIMEPERIOD";           # recover timeperiod for max attempts
my $KEYSAPRECOVER_TIMEELAPSED="LK_RECOVER_TIMEELAPSED";         # time elapsed since restore
my $KEYSAPRECOVER_CURRENTTIME="LK_RECOVER_CURRENTTIME";         # current epoch at restore or remove
my $KEYSAPRESTORE_DATE="LK_RESTORE_DATE";               # current lk date/time
my $KEYSAPENQ_VERSION="SAP_ENQ_VERSION";         # key of enqueue server version
my $KEYSAPENQREP_VERSION="SAP_ENQREP_VERSION";   # key of enqueue replication server version

################################################################################
######################--- End SAP Hash Variables Section ---####################
################################################################################

################################################################################
##################--- Start SAP Info Variables Section ---######################
################################################################################

# Defines for pulling default, most used values from the info field of a LK
# resource
our $DFLTINFOLOC=0;
our $NFSINFOLOC=0;
our $IPINFOLOC=2;

# The location of all the vaues in the current resource info field.
our $SIDINFOLOC=0;
our $INSTINFOLOC=1;
our $VIPINFOLOC=2;
our $MNTINFOLOC=3;
our $PROTECTINFOLOC=4;
our $RECOVERINFOLOC=5;
our $INSTTYPEINFOLOC=6;
our $CITAGINFOLOC=98;   # not used
our $REPENQINFOLOC=99;  # not used

################################################################################
######################--- End SAP Type Variables Section ---####################
################################################################################

# These are related to the Primary Application Servers
# PAS.  We may need these to sort and weed out values
our $ABAPCI="DVEBMGS";
our $JCI="J";
our @SAPCI=("SCS", "ASCS");

# Define types of each SAP component
our $TYPE_CS=1;
our $TYPE_ERS=2;
our $TYPE_CI=3;
our $TYPE_AS=4;
our $TYPE_NEW_ERS=5; # bug7631: Added to support ERS v1 and v2 in an independent hierarchy

################################################################################
######################--- Start SAP Type Variables Section ---##################
################################################################################


################################################################################
#####################--- Start SAP Debug/Log Function Section ---###################
################################################################################


# Set of log routines to make transition to LK.* logging easier
sub dbg
{
        return 0 unless ($SAP_DEBUG);
        do_dbg(@_);
}

sub do_dbg
{
        my $msg=shift;
        my $tabs;
        if ($msg !~ /trace/i && $msg !~ /dbg/i ){
                $msg="\tTrace[$$]:$msg";
        }

        if ($msg =~ /\n$/ ){
                $msg=~s/(.*)\n$/$1/;
        }
        LKBase::LogMessage($_log, calledby(), 'RKBase', 'PlainMessage', $msg);
}

# debug print function.
sub OBJECT_DEBUG_PRINT
{
        my $self=shift;
        return 0 unless ($SAP_DEBUG && !$self->{'CMR_DEBUG'});
        dbg "Dumping the Object Information: ". Dumper ($self) ."\n";
        return 1;
}

####
# Description
#       Set the logging level
####
# Input
#       Logging - all (STDERR, STDOUT, LOGFILE), default (STDERR, LOGFILE), other
####
# Output
#       None
####
# Returns
#       1 - always
sub logset
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        my $Logging=shift;

        # When debug is on - it doesn't make since to silence logging
        # or log anywhere but the defaults
        if ($SAP_DEBUG ){
                dbg "!!!! ($$) !!!! Manual override of logset required b/c you have selected debug on\n";
                $Logging='default';
        }

        if ($Logging =~ /all/i ){
                $_log = {'STDERR' => 1, 'LOGFILE' => 1 , 'STDOUT' => 1 };
        }elsif ($Logging =~ /default/i ){
                $_log = {'STDERR' => 1, 'LOGFILE' => 1 };
        }else{
                my %l;
                foreach my $v (split /\s+/, $Logging ){
                        $v=uc($v);
                        $l{$v} = 1;
                }
                $_log = \%l;
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}


sub do_stack
{
    my $die = shift;
    my $l = 0;
    my $msg = ($_[0] ? "@_" : ($die ? "Died" : "Error")) . " at ";

    while (++$l) {
        @_ = caller($l);
        last if (!defined($_[0]));
        $_[1] =~ s@^.*/@@;
        $msg .= ", " if ($l > 1);
        $msg .= "$_[1]:$_[2] ($_[3])";
    }

    if ($die) {
       # msg $msg;
        die("$msg\n");
    }
    dbg $msg;
}

sub stack
{
    do_stack(0, @_);
}

sub stack_die
{
    do_stack(1, @_);
}

# return $0 if it's not '-e', otherwise return calling subroutine
sub calledby
{
    return $cmd if ($cmd ne '-e');
    my $level = 2; # 2 means the caller of the caller
    # skip over message helper functions to the "real" calling routine
    $level++ while (((caller($level))[3] =~ /(dbg|output|msg|fatal|error|base)2*$/));
    return (caller($level))[3];
}

sub erroranddie
{
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        error(@_); stack_die();
}

sub msganddie
{
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        msg(@_); stack_die();
}

sub error
{
        my $error=1;
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        return msg1($resType,$error,@_);
}

sub msg1
{
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        my $caller = calledby();
        my $msgFile=shift;
        my $type=shift;
        my $msgString=shift;
        ($noise) && return 1;
        if ( !$type) {
                LKBase::LogMessage($_log, $caller, $msgFile, $msgString, @_);
        }else{
                LKBase::LogError($_log, $caller, $msgFile, $msgString, @_);
        }
        return 1;
}

sub msg
{
        my $error=0;
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        msg1($resType,$error,@_);
}

sub baseerroranddie
{
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        baseerror(@_);
        stack_die();
}

sub baseerror
{
        my $error=1;
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        return msg1('RKBase', $error, @_);
}

sub basemsg
{
        my $error=0;
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        return msg1('RKBase', $error, @_);
}

sub info
{
        (ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ ) && shift;
        my $caller = calledby();
        LKBase::LogMessage($_log, $caller, 'RKBase', 'PlainMessage', @_);
        return 1;
}

################################################################################
#######################--- End SAP Debug/Log Function Section ---###################
################################################################################

################################################################################
##################--- Start SAP Get/Set (Getter/Setter) Section ---#############
################################################################################

# General Description
# These functions get items from the object or set a value in the
# object.
# Get - when called with no options or values
# Set - when called with a value
# To make the changes permanent, the user should call the helper/UI
# function SAPUpdateLKInfoField();  which takes the object values
# and uses ins_setinfo and/or call SAPEND(1) which will write current
# values of the file to make them permanent
# Typical setup is function name matches keyname function=SAPSID key=KEYSAPSID
sub SAPOBJECT
{
        my $self=shift;
        my $key=shift;
        my $value=shift;
        $self->{$key}=$value;
        return 1;
}

# set info field into object; retrieve info field from object
sub SAPLKSAPINFO
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPLKSAPINFO, @_);
        return $self->{$KEYSAPLKSAPINFO};
}

# set state field into object; retrieve state field from object
sub SAPLKSAPSTATE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPLKSAPSTATE, @_);
        return $self->{$KEYSAPLKSAPSTATE};
}

# set lk id field into object; retrieve lk id field from object
sub SAPLKSAPID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPLKSAPID, @_);
        return $self->{$KEYSAPLKSAPID};
}

# set ins_list field into object; retrieve ins_list field from object
sub SAPLKSAPLK
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPLKSAPLK, @_);
        return $self->{$KEYSAPLKSAPLK};
}

# Note: This function sets and retrieves a temporary file
# related to the ENQ Server resource
sub SAPREPENQSTORE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPREPENQSTORE, @_);
        return $self->{$KEYSAPREPENQSTORE};
}

# Note: This function sets and retrieves a temporary file
# related to the Info Fiedl
sub SAPINFOSTORE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPINFOSTORE, @_);
        return $self->{$KEYSAPINFOSTORE};
}

# Note: This function sets and retrieves a temporary file
# related to the AS Server resource
sub SAPASSTORE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPASSTORE, @_);
        return $self->{$KEYSAPASSTORE};
}

# get/set pid of SAPSRV
sub SAPSRVPID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSRVPID, @_);
        return $self->{$KEYSAPSRVPID};
}
# set gid field into object; retrieve gid field from object
sub SAPGID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPGID, @_);
        return $self->{$KEYSAPGID};
}

# set user shell field into object; retrieve user shell field from object
sub SAPSHELL
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSHELL, @_);
        return $self->{$KEYSAPSHELL};
}

# set uid field into object; retrieve uid field from object
sub SAPUID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPUID, @_);
        return $self->{$KEYSAPUID};
}

# set admin home field into object; retrieve admin home field from object
sub SAPADMINHOME
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPADMINHOME, @_);
        return $self->{$KEYSAPADMINHOME};
}

# set admin name field into object; retrieve admin name field from object
sub SAPADMIN
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPADMIN, @_);
        return $self->{$KEYSAPADMIN};
}

# set tmpfile field into object; retrieve tmpfile name field from object
sub SAPTMPFILE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPTMPFILE, @_);
        return $self->{$KEYSAPTMPFILE};
}

# set SAP default dir into object; retrieve SAP default dir from object
sub SAPDIR
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPDIR, @_);
        return $self->{$KEYSAPDIR};
}

# set dbname into object; retrieve dbname from object
sub SAPDBNAME
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPDBNAME, @_);
        return $self->{$KEYSAPDBNAME};
}

# set dbtype into object; retrieve dbtype from object
sub SAPDBTYPE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPDBTYPE, @_);
        return $self->{$KEYSAPDBTYPE};
}

# set DIR_TRANS into object; retrieve DIR_TRANS from object
sub SAPDIRTRANS
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPDIRTRANS, @_);
        return $self->{$KEYSAPDIRTRANS};
}

# set dbhost into object; retrieve dbhost from object
sub SAPDBHOST
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPDBHOST, @_);
        return $self->{$KEYSAPDBHOST};
}

# set original $0 into object; retrieve original $0 from object
sub SAPPARENT
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPPARENT, @_);
        return $self->{$KEYSAPPARENT};
}

# set os collector pid into object; retrieve os collector pid from object
sub SAPCOLLPID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPCOLLPID, @_);
        return $self->{$KEYSAPCOLLPID};
}

# set small wait into object; retrieve small wait from object
sub SAPSMALLWAIT
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSMALLWAIT, @_);
        return $self->{$KEYSAPSMALLWAIT};
}

# set small wait into object; retrieve small wait from object
sub SAPRKSWAIT
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRKSWAIT, @_);
        return $self->{$KEYSAPRKSWAIT};
}


# set flags into object; retrieve flags from object
sub SAPFLAGS
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPFLAGS, @_);
        return $self->{$KEYSAPFLAGS};
}

# set tag into object; retrieve tag from object
sub SAPTAG
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPTAG, @_);
        return $self->{$KEYSAPTAG};
}

# set recovery_level into object; retrieve recovery_level from object
sub SAPRECOVERY
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVERY, @_);
        return $self->{$KEYSAPRECOVERY};
}

# set protection_level into object; retrieve protection_level from object
sub SAPPROTECTION
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPPROTECTION, @_);
        return $self->{$KEYSAPPROTECTION};
}

# set <sapmnt> into object; retrieve <sapmnt> from object
sub SAPMNT
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPMNT, @_);
        return $self->{$KEYSAPMNT};
}

# set <vip> into object; retrieve <vip> from object
sub SAPVIP
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPVIP, @_);
        return $self->{$KEYSAPVIP};
}

# Note: This function sets and retrieves the actual
# ENQ Instance that is associated with the actively
# protected CI Instance
sub SAPREPENQINST
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPREPENQINST, @_);
        return $self->{$KEYSAPREPENQINST};
}

# set inst_type into object; retrieve inst_type from object
sub SAPINSTTYPE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPINSTTYPE, @_);
        return $self->{$KEYSAPINSTTYPE};
}

# set SAP SRV CMD into object; retrieve SAP SRV CMD from object
sub SAPSRVCMD
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSRVCMD, @_);
        return $self->{$KEYSAPSRVCMD};
}

# set profile location obtained from sapservices file into object; retrieve profile location from object
sub SAPSRVPF
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSRVPF, @_);
        return $self->{$KEYSAPSRVPF};
}

# set SAP SID into object; retrieve SAP SID from object
sub SAPSID
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPSID, @_);
        return $self->{$KEYSAPSID};
}

# set SAP INSTNUM into object; retrieve SAP INSTNUM from object
sub SAPINSTNUM
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPINSTNUM, @_);
        return $self->{$KEYSAPINSTNUM};
}
# set SAP INST into object; retrieve SAP INST from object
sub SAPINST
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPINST, @_);
        return $self->{$KEYSAPINST};
}

# set result_file into object; retrieve result_file from object
sub SAPRESULTFILE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRESULTFILE, @_);
        return $self->{$KEYSAPRESULTFILE};
}

# set lock_file into object; retrieve lock_file from object
sub SAPLOCKFILE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPLOCKFILE, @_);
        return $self->{$KEYSAPLOCKFILE};
}

# set or retrieve GREEN event response in or from object;
sub SAPGREEN
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPGREEN, @_);
        #return $self->{$KEYSAPGREEN}; #per design review
        # always return 1;
        return 1;
}


# set or retrieve RED event response in or from object;
sub SAPRED
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRED, @_);
        #return $self->{$KEYSAPRED}; #per design review
        # always return 1;
        return 1;
}

# set or retrieve YELLOW event response in or from object;
sub SAPYELLOW
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPYELLOW, @_);
        return $self->{$KEYSAPYELLOW};
}

# ***        Wed Jun  1 10:03:50 EDT 2011         ***
# Below the line features related to additional
# resource information objects.  These items
# pertain to flexible recover and additional
# recovery information stored in file SAP_INFO_FILE_DEFAULTS
# each key in the has matches a function in the pm the
# will set/retrieve the value

#set/retrieve current date of restore action on host
sub  SAPRESTORE_DATE
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRESTORE_DATE, @_);
        return $self->{$KEYSAPRESTORE_DATE};
}

#set/retrieve current time of restore action on host
sub  SAPRECOVER_CURRENTTIME
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_CURRENTTIME, @_);
        return $self->{$KEYSAPRECOVER_CURRENTTIME};
}

#set/retrieve recover attempts made on host
sub  SAPRECOVER_ATTEMPTS
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_ATTEMPTS, @_);
        return $self->{$KEYSAPRECOVER_ATTEMPTS};
}

#set/retrieve max attempts allowed on host per time period
sub  SAPRECOVER_ATTEMPTS_MAX
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_ATTEMPTS_MAX, @_);
        return $self->{$KEYSAPRECOVER_ATTEMPTS_MAX};
}

#set/retrieve recover success attempt
sub  SAPRECOVER_SUCCESS
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_SUCCESS, @_);
        return $self->{$KEYSAPRECOVER_SUCCESS};
}

#set/retrieve recover failed attempt
sub  SAPRECOVER_FAILED
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_FAILED, @_);
        return $self->{$KEYSAPRECOVER_FAILED};
}

#set/retrieve max attempts time period on this host
sub  SAPRECOVER_TIMEPERIOD
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_TIMEPERIOD, @_);
        return $self->{$KEYSAPRECOVER_TIMEPERIOD};
}

#set/retrieve recover time elapsed since restore/recover
sub  SAPRECOVER_TIMEELAPSED
{
        my $self=shift;
        ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPRECOVER_TIMEELAPSED, @_);
        return $self->{$KEYSAPRECOVER_TIMEELAPSED};
}

# set/retrieve whether we are in a recover operation
sub SAPINRECOVER
{
       my $self=shift;
       ( @_ != 0 ) && return $self->SAPOBJECT($KEYSAPINRECOVER, @_);
       return $self->{$KEYSAPINRECOVER};
}

# set or retrieve enqueue server version in or from object
# only applicable for TYPE_CS instances
sub SAPENQ_VERSION
{
        my $self=shift;
        ((@_ != 0) && ($_[0] ne 'getvalue')) && return $self->SAPOBJECT($KEYSAPENQ_VERSION, @_);
        $self->SAPOBJECT($KEYSAPENQ_VERSION, $self->GetEnqVersion()) if ($self->{$KEYSAPENQ_VERSION} eq '');
        return $self->{$KEYSAPENQ_VERSION};
}

# set or retrieve enqueue replication server version in or from object
# only applicable for TYPE_ERS or TYPE_NEW_ERS instances
sub SAPENQREP_VERSION
{
        my $self=shift;
        ((@_ != 0) && ($_[0] ne 'getvalue')) && return $self->SAPOBJECT($KEYSAPENQREP_VERSION, @_);
        $self->SAPOBJECT($KEYSAPENQREP_VERSION, $self->GetEnqVersion()) if ($self->{$KEYSAPENQREP_VERSION} eq '');
        return $self->{$KEYSAPENQREP_VERSION};
}

################################################################################
###################--- End SAP Get/Set (Getter/Setter) Section ---##############
################################################################################


################################################################################
########################--- Start SAP Object Section ---#######################
################################################################################
# General
# These routines provide methods for creating an object necessary for calling
# functions within sap.pm and manipulating object variables.
# The three classes of new allow for object creation as necessitated by the
# various scripts in question.
# new() - for a fully loaded object that can be updated and add to
# newFromSys() - for an object based on some system values
# newFromTag() - for an object based on an existing resource tag
# newInit()  - load system values in a common spot

####
#  Description
#       - Creates a specific SAP object based on the SAP SID and Instance, or tag
#       - The object will be filled with data from the lookups
####
#  Input:
#       - sid - SAP SID
#       - inst - the instance for the SID
#       - tag - the associated tag for the sid & inst ( or '' if none)
#       - sys - system for the object to be created from
####
#  Output:
#       - None (values set in object)
####
#  Returns
#       - object reference handle
###
sub new
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $proto = shift;
        my $sid = shift;
        my $Instance = shift;
        my $tag = shift;
        my $sys = shift;
        if (  $sid eq '' || $Instance eq '' ){
                if ( $tag eq '' ){
                        dbg "<= Trace:Leaving $func on $me\n";
                        error ('missingparam', "<sid> && <instance> or <tag>", $func, $me );
                        return 0;
                }
        }
        my $class = ref($proto) || $proto;
        dbg "--- ($class) --- (".calledby().") --- ($sid) --- ($Instance) --- ($tag) --- ($$) --- ($me)\n";
        my $self = {};
        bless ($self, $class);

        # save a dummy default in case the passed in tag is bogus.
        my $DEFAULT_TAG="SAP-${sid}_${Instance}";
        my $initType=0;
        $self->{$KEYSAPFROMTAG}=1;
        my $ret;
        my $badTag = 0;
        # If a tag value was passed in, try using GetLK to get all of the stored values from the LifeKeeper resource
        # This call will be trapped in an eval so that we can retry using system values if no tag exists
        if ( $tag ne '' ){
                $ret = eval { $self->GetLK($tag,$sys) };
                if ($@){
                        dbg "The object function $func was called with tag $tag, but this tag does not exist in LifeKeeper on $me\n";
                        # If we were called with a tag and no sid or instance, this is a critical error
                        # but we cannot die explicitly or lkcheck breaks.  Instead return 0 instead of
                        # the object.  Now all calls to new need to check for a real object
                        if (  $sid eq '' || $Instance eq '' ){
                                dbg "<= Trace:Leaving $func on $me\n";
                                error ('missingparam', "<sid> && <instance> or <tag>", $func, $me );
                                return 0;
                        }
                        # set for later
                        $badTag=1;
                }
        }
        # If no tag was passed in, or the tag used was a dummy or placeholder, then we will call newInit
        # to fully populate the object using system values
        # TO DO: Consider using newInit and never using GetLK??
        if ($tag eq '' || $badTag ){
                $initType=1;
                $self->{$KEYSAPFROMTAG}=0;
                #
                dbg "Generate a tag using the sid and Instance that was passed in to ($func) on $me\n";
                $tag = $DEFAULT_TAG;
        }
        dbg "Attempting to finish object initialization using newInit ($initType) with a tag generated from the innput ($tag) on $me\n";

        # load as many defaults as possible for the object
        # by invoking the correct version of newInit
        $ret = eval { $self->newInit($initType,$sid,$Instance,$tag, $sys) };
        if ($@){
                dbg "The internal function $func/newInit was called with a bunch of junk!! This is a fatal exception on $me!!\n";
                dbg "<= Trace:Leaving $func on $me\n";
                error ('cantcreateobject', "$sid", "$Instance", "$tag", $me );
                return 0;
        }

        bless $self, $class;
        dbg "<= Trace:Leaving $func on $me\n";
        return $self;

}


####
#  Description
#       - Initializes the object with data from the ins_list or system
####
#  Input:
#   - type -
#               [0] from system
#               [1] from tag
#       - sid - sap SID
#       - Instance - sap instance
#       - tag - resource tag for the instance
#       - [sys] - the system to base the instance information on
####
#  Output:
#       - None (values set in object)
####
#  Returns
#       - object reference handle
####
sub newInit
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $initType=shift;
        my $sid=shift;
        my $Instance=shift;
        my $tag=shift;
        my $sys=shift;

        # These keys have very few uses  and are not
        # wrapped by any object wrapper get/set routine
        $self->SAPOBJECT($KEYSAPLOCALSYS,$me);
        $self->SAPOBJECT($KEYSAPACTIONS,$APPACTIONS);
        $self->SAPOBJECT($KEYSAPPASSWORD,$PASSWDFILE);
        $self->SAPOBJECT($KEYSAPFROMTAG,0);

        $self->SAPDIR($SAPDIR);
        $self->SAPLOCKFILE($lockfile);
        $self->SAPTMPFILE($tmpfile);
        $self->SAPRESULTFILE($resfile);

        # These are set to the defaults that are stored on the hash
        # they are reset whenever the real INFO_FILE is read
        $self->{$KEYSAPYELLOW}=$SAP_INFO_FILE_DEFAULTS{'SAPYELLOW'};
        $self->{$KEYSAPRED}=$SAP_INFO_FILE_DEFAULTS{'SAPRED'};

        $self->SAPPARENT(calledby());
        if ( $initType ){
                # Set object values from input values
                dbg "Object initialization ($initType) taking place using the sid=$sid and Instance=$Instance on $me\n";
                dbg "Object initialization will search for the tag because ($initType) assumes that the tag, $tag, may not be valid on $me\n";
                $self->SAPSID($sid);
                $self->SAPINST($Instance);
                (my $haveATag, undef, undef)=isWhatProtected($me, $sap::appType, $sap::resType, $Instance, $sap::INSTINFOLOC);
                if ( $haveATag ne '' ){
                        dbg "$func was called with tag ($tag), the tag for the instance $Instance was found ($haveATag)\n";
                        $tag=$haveATag;
                        my $ret = eval { $self->GetLK($tag, $sys) };
                }
                $self->SAPINSTTYPE($self->GetSAPInstType());
                $self->SAPTAG($tag);
        }
        #May 11 2011:
        # The SAPREPENQINST is a hold over from a previous prototype
        # If we do not store replicate enqueue in the info field
        # then we can probably remove this key and code. May 11 2011
        if ($self->SAPINSTTYPE() == $TYPE_ERS ){
                $self->SAPREPENQINST($self->SAPINST());
        }
        $self->SAPINSTNUM($self->GetSAPInstNum());
        # These are functions to load additional object data
        my $ret=$self->NFSCheck();
        if ( $ret == 0 ){
                error('nonfs', $me );
                dbg "<= Trace:Leaving $func on $me\n";
                erroranddie ('nonfs', $me );
        }
        # NFS Mounts
        dbg "[7406]: INFORMATION: newInit Execute LocalNFSMounts before checking for /sapmnt system information\n";
        $self->LocalNFSMounts();
        $self->GetSysSAPMntDir();
        $self->GetSysAdminInfo();
        $self->SAPDIRTRANS($SAPTRANSDIR);
        # bug7622 - Don't run startsap on backup node on extend
        if ($cmd ne "extend" && $cmd ne "canextend"){
        	dbg "[7406]: INFORMATION:Before you go getting all the additional instance information, make sure sapstartsrv is running\n" ;
        	my $msgRet=$self->StartSapServer($sid,$Instance);
        	if ($msgRet == 0 ){
                	erroranddie ('sapstartsrvfailed', $sid, $Instance,$me );
        	}
        	if ( !$self->GetSysDFLTInfo()){
                	# Using erroranddie triggers the eval to fail
                	# which is needed to indicate a failed object create
                	erroranddie ('errorgetting', 'default (DFLTInfo)', $DPFL, $me );
        	}
        }

        $self->GetSysSAPSRV();
        $self->GetSysSAPSRVPF();
        $self->SAPSMALLWAIT($SAP_SMALL_SLEEP_INTERVAL);
        $self->SAPRKSWAIT($SAP_RKS_SLEEP_INTERVAL);

        # check for Autostart = 1
        $self->AutoConfigCheck();

        my $SAP_FLAG_QCK_IN_PROGRESS_tag = "${SAP_FLAG_QCK_IN_PROGRESS}${tag}";
        my $SAP_FLAG_RECOVER_IN_PROGRESS_tag = "${SAP_FLAG_RECOVER_IN_PROGRESS}${tag}";
        my $SAP_FLAG_RESTORE_IN_PROGRESS_tag = "${SAP_FLAG_RESTORE_IN_PROGRESS}${tag}";
        my $SAP_FLAG_REMOVE_IN_PROGRESS_tag = "${SAP_FLAG_REMOVE_IN_PROGRESS}${tag}";
        my $SAP_FLAG_MOVE_ERS_tag = "${SAP_FLAG_MOVE_ERS}${tag}";

        #Notes: Per the CI we are adding the kitchen sink to the initial object
        # this includes the action flags.  The flags will be stored in a hash
        # with the hash stored in the object.
        # To retrieve the values from this hash try one of the following:
        #        %a=%{$self->SAPFLAGS()};
        #       foreach my $f (keys %a ){
        #               print STDERR "key=$f val=$a{$f}\n";
        #       }
        #       my $flag=${$self->SAPFLAGS()}{$Type};
        #
        my %Flags = (
                0 => $SAP_FLAG_RECOVER_IN_PROGRESS_tag,
                1 => $SAP_FLAG_QCK_IN_PROGRESS_tag,
                2 => $SAP_FLAG_RESTORE_IN_PROGRESS_tag,
                3 => $SAP_FLAG_REMOVE_IN_PROGRESS_tag,
                4 => $SAP_FLAG_MOVE_ERS_tag
        );
        $self->SAPFLAGS(\%Flags);

        # Since we are created with a tag, update the
        # tmp files and storage files to include the tag
        # <type>_<tag>
        $self->SAPASSTORE("$AS_FILE_STORE$tag");
        $self->SAPREPENQSTORE("$REPENQ_FILE_STORE$tag");
        $self->SAPINFOSTORE("$INFO_FILE_STORE$tag");

        # Read any extra info parameters from the INFO storage file
        $self->SAPREADER();
        dbg "=> Trace:Leaving $func on $me\n";
        return 1;
}

#### GetLK
# Description
#       - Obtain the LK resource instance information
####
#  Input:
#       - tag - The LK resource tag for the protected instance
#       - [sys] - The system to get the instance information on
####
#  Output:
#       - None (set in object)
####
#  Returns
#       - 0 - errors
#       - 1 - success
####
sub GetLK
{

        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag=shift;
        my $sys=shift;
        if ($tag eq ''){
                dbg "<= Trace:Leaving $func on $me\n";
                msganddie ('missingparam', "<tag>", $func, $me );
        }
        $self->SAPTAG($tag);
        my @info;
        if (defined $self->{$KEYSAPLKSAPLK}){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        $self->{$KEYSAPLKSAPLK} = undef;        # start clean
        @_ = LK::ins_list($sys, '', '', '', $tag , '');
        if (@_ == 0){
                dbg "<= Trace:Leaving $func on $me\n";
                # based on CI and newInit, this is not a error, but
                # we trigger die to cause the newInit to try an
                # alternate action of the eval
                msganddie ('lkinfoerror', $func, $tag, $me );
        }
	$signalTag=$tag; #add the tag so that is needed for signal

        # storage info field items in the resource object
        @_ = split /\001/, $_[0];

        $self->SAPLKSAPLK(\@_);         # similar to the {LK} in other kits
        $self->SAPLKSAPINFO($_[5]);
        $self->SAPLKSAPID($_[4]);
        $self->SAPLKSAPSTATE($_[6]);

        #Jun 6 2011:
        # Keep the old info field in the .pm for upgrade reference
                # Note: The old info field appeared in two formats.  One including a java flag
                # Note: The new format will include additional information
                # Note: OLD[1]: #-v"$SID$InstanceID$VirtualSysName$SAPRootJS-$JavaFlg"
                # Note: OLD[2]: #-v"$SID$InstanceID$VirtualSysName$SAPRoot"

        @info = split /\002/, $self->SAPLKSAPINFO();

        # Info field items are positional and we use the numbers
        # in case we need to add or rearrange
        $self->SAPSID($info[$SIDINFOLOC]);
        $self->SAPINST($info[$INSTINFOLOC]);
        $self->SAPVIP($info[$VIPINFOLOC]);
        $self->SAPMNT($info[$MNTINFOLOC]);
        $self->SAPPROTECTION($info[$PROTECTINFOLOC]);
        $self->SAPRECOVERY($info[$RECOVERINFOLOC]);
        $self->SAPINSTTYPE($info[$INSTTYPEINFOLOC]);

        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Object destructor
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - always
####
sub DESTROY
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self=shift;
        my $exitRequested=$?; #save off exit code that triggered DESTROY
        $self->CleanUp($$);
        foreach my $file ("$lockfile",  "$tmpfile", "$resfile"){
                next if ( ! -f $file );
                unlink <$file*>;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        $? = $exitRequested; #reset exit code, it was (may have been) lost by system calls
        return 0;
}

################################################################################
########################--- End SAP Object Section ---#######################
################################################################################


################################################################################
########################--- Start HostExec Section ---##########################
################################################################################
# General
# The HostExec is a utility/server that must be running for many of the
# saphostctrl, startsapsrv, and sapcontrol commands to run successfully
# The kit will need to start this utility/server and also have the
# ability to monitor and restart the command when needed


####
#  Description
#       - Start, stop or get status of the saphostexec utility on
#       - the current host.
#       - This uses the -start function of the utility
####
#  Input:
#       - action - action to perform (status, start, stop)
####
#  Output:
#       - None
####
#  Returns
#       - 1 - running
#       - 0 - not running
sub HostExec
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;
        my $wait=$self->SAPSMALLWAIT();
        my $pf="pf=$SAP_HOSTCTL_PROFILE";
        my $endcriteria=($action =~ /stop/)? 0:1;
        my $retries=-1;         # how many retries for this action; default none
        # what other command needs running before I do an action
        my %beginCmd;
        # what other command needs running after I do an action
        my %endCmd;     # set in stop and start blocks below
        my $ret=0;

        if ( ! -e $SAP_SRVHOST_CMD ){
                dbg "No $SAP_SRVHOST_CMD exists on $me\n";
                print STDERR "INFORMATION: The $SAP_SRVHOST_CMD command does not exist on $me\n";
                return $endcriteria;
        }

        msg ('userstatus', "$SAP_SRVHOST_CMD $pf -status", $me );
        # This command returns 0 if it is running, 1 if not
        system ("$SAP_SRVHOST_CMD $pf -status $REDIRECT"); #could send to /dev/null
        my $sret=($? >> 8 );
        my $pid=GetPids($SAP_SRVHOST_CMD);
        dbg "\t$SAP_SRVHOST_CMD command returned ". ($OS_RET{$ret} || $ret ) . ". The process (ps) search returned pid=$pid on $me.\n";
        if ( $pid > 1 && $sret == 0  ){
                $ret=1;
                msg ('running', $SAP_SRVHOST_CMD, $me );
        }else{
                msg ('notrunning', $SAP_SRVHOST_CMD, $me );
        }

        if ($action =~ /start/ ){
                %endCmd= ( "OSCollector" => "start" );
                if ($ret == 0){
                        msg ('start', $SAP_SRVHOST_CMD, $me );
                        $retries=1;
                }
        }elsif ($action =~ /stop/ ){
                %endCmd= ( "OSCollector" => "stop" );
                if ($ret){
                        $retries=$SAP_SRV_RETRIES;
                        msg ('stop', $SAP_SRVHOST_CMD, $me );
                }
        }

        dbg "$func: pre-entry status ret=$ret, retries=$retries, action=$action, endcriteria=$endcriteria on $me\n";
        my $cnt=0;
        my $update=0;
        while ( $cnt <= $retries ){
                msg ('userstatus', "$SAP_SRVHOST_CMD $pf -$action", $me );
                system ("$SAP_SRVHOST_CMD $pf -$action $REDIRECT"); #could send to /dev/null
                sleep $wait;
                $ret=$self->HostExec('status');
                dbg "$func: Retry Loop. retries=$retries - ret=$ret - cnt=$cnt - endcriteria=$endcriteria - action=$action - on $me\n";
                last if ($ret == $endcriteria );
                $cnt++;
                $update++;
                # the loop counter has exceeded the throttle counter
                # reset the counter, and allow messages on next iteration
                if ($update > $SAP_MSG_THROTTLE ){
                        loud();
                        $update=0;
                }else{
                        quiet();
                }
        }
        loud();
        foreach my $bCmd (keys %endCmd ){
                $self->$bCmd($endCmd{$bCmd});
        }
        dbg "$func ret=$ret - endcriteria=$endcriteria for action=$action on $me\n";
        dbg "<= Trace:Leaving $func on $me\n";
        return 1 if ($ret == $endcriteria );
        return 0;
}


################################################################################
########################--- End HostExec Section ---############################
################################################################################


################################################################################
#############--- Start SRV Section & sapstartsrv Section ---############
################################################################################
# General
# In addition to the hostexec utility/server, the startsapsrv
# utility/server must be running for many of the
# sapcontrol commands to run successfully.
# The kit will need to start this utility/server and also have the
# ability to monitor and restart the command when needed

# START bug7406
# The kit will need to start this utility/server and also have the
# ability to monitor and restart the command when needed

# Currently we expect sapcontrol to be in the user's path
# this function will verify sapcontrol can be found
# and that it is executable
# 0 - not found or configured in the user path
# 1 - found and executable
sub FindSapServerCmd
{
        my $self=shift;
        my $resfile=$self->SAPRESULTFILE();
        my $sapadmin=$self->SAPADMIN();
        my $sapcmd=$self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C which $SAP_CONTROL; echo \\\$? > $resfile' \"");
        chomp ($sapcmd);
        chomp (my $findReturn=`cat $resfile`);
        unlink $resfile;
        if ($findReturn || $sapcmd eq '' ){
                dbg "ERROR: The 'which' command for the $SAP_CONTROL command failed";
                return 0;
        }
        # we use a relative path when we call $SAP_CONTROL, but based on customer issue
        # we should give ourselves a bread crumb and force a hard stop when the full path to the command
        # cannot be found or is not executable
        if ( ! -x $sapcmd ){
                dbg "ERROR: Unable to find the full path to the command $SAP_CONTROL (sapcmd=$sapcmd).\n";
                return 0;
        }
        return 1;
}

# 0 - not running
# 1 - running
sub StatusSapServer
{
        my $self=shift;
        my $sid=shift;
        my $Instance=shift;
        if ($sid eq '' ){
                $sid = $self->SAPSID();
        }
        if ($Instance eq '' ){
                $Instance = $self->SAPINST();
        }
        if ($sid eq '' || $Instance eq '' ){
                dbg "[7406]: ERROR: Unable to find instance and sid parameters\n";
                msg ('missingparam', "SID and Instance", 'sap::StatusSapServer', $me );
                return 0;
        }
        my $pid=GetPids("$sid $Instance $SAP_SRV");
        if ( $pid eq '' || $pid <= 1 ){
                dbg "The $SAP_SRV command for this instance $Instance is not running\n";
                return 0;
        }
        $self->SAPSRVPID($pid);
        dbg "The $SAP_SRV command for this instance $Instance is already running (pid=$pid)\n";
        return 1;
}

# 0 - not started
# 1 - started
sub StartSapServer
{
        my $self=shift;
        my $sid=shift;
        my $Instance=shift;
        my $instNum=shift;
        my $wait=$self->SAPSMALLWAIT();
        my $resfile=$self->SAPRESULTFILE();
        my $sapadmin=$self->SAPADMIN();
        my $startRet;

        my $Type=$self->SAPINSTTYPE();
        $wait=($wait > 1 ) ? $wait : 5 ;
        if ($sid eq '' ){
                $sid = $self->SAPSID();
        }
        if ($Instance eq '' ){
                $Instance = $self->SAPINST();
        }
        if ($instNum eq '' ){
                $instNum=$self->SAPINSTNUM();
        }
        if ($sid eq '' || $Instance eq '' || $instNum eq ''){
                dbg "[7406]: ERROR: Unable to find instance, instNum and sid parameters\n";
                msg ('missingparam', "SID and Instance", 'sap::StartSapServer', $me );
                return 0;
        }
        dbg "[7406]: Start HostExec from StartSapServer\n";
        # Try to start HostExec.  If it will not start, move on.  It is not critical
        $self->HostExec('start');
        # If we can't find the full path to the command, the environment is off and likely
        # to fail the sap instance start, so flag it now
        return 0 if (!$self->FindSapServerCmd());
        my $startSapServerCmd="$SAP_CONTROL_CMD -nr $instNum -function StartService $sid";
        foreach my $loops (0..$SAP_SRV_RETRIES){
                if ( $self->StatusSapServer($sid, $Instance) == 1 ){
                        return 1;
                }
                msg ('userstatus', $startSapServerCmd, $me );
                dbg "Running [$loops of $SAP_SRV_RETRIES] command $startSapServerCmd\n";
                $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $startSapServerCmd; echo \\\$? > $resfile' \" >&2");
                chomp ($startRet=`cat $resfile`);
                unlink $resfile;
                dbg "[7406]: Start returned $startRet\n";
                sleep $wait;
        }
        my $ret=$self->StatusSapServer($sid, $Instance);
        return 1 if ($ret);
        # go old school
        my $srvcmd=$self->SAPSRVCMD();
        dbg "[7406]: INFORMATION: Unable to start using $startSapServerCmd using $srvcmd\n";
        system ("$srvcmd $REDIRECT");
        sleep $wait;
        # end old school
        $ret=$self->StatusSapServer($sid, $Instance);
        return 1 if ($ret);
        error ('sapstartsrvfailed', $sid, $Instance,$me );
        return 0;
}



#  Description
#       - Stops the startsapsrv script used with web interfaces and sapcontrol
#  Returns
#       - 0 - error
#       - 1 - success
sub StopSapServer
{
        my $self=shift;
        my $sid=shift;
        my $Instance=shift;
        my $instNum=shift;
        my $sig=15;
        my $sapadmin=$self->SAPADMIN();
        my $resfile=$self->SAPRESULTFILE();
        my $stopRet;
        my $wait=$self->SAPSMALLWAIT();
        $wait=($wait > 1 ) ? $wait : 5 ;
        if ($sid eq '' ){
                $sid = $self->SAPSID();
        }
        if ($Instance eq '' ){
                $Instance = $self->SAPINST();
        }
        if ($instNum eq '' ){
                $instNum=$self->SAPINSTNUM();
        }
        if ($sid eq '' || $Instance eq '' || $instNum eq ''){
                msg ('missingparam', "SID and Instance", 'sap::StopSapServer', $me );
                return 0;
        }

        # If we can't find the full path to the command, the environment is off and likely
        # to fail the sap instance stop, so flag it now
        return 0 if (!$self->FindSapServerCmd());
        my $stopSapServerCmd="$SAP_CONTROL_CMD -nr $instNum -function StopService $sid";
        foreach my $loops (0..$SAP_SRV_RETRIES){
                if ( $self->StatusSapServer($sid, $Instance) == 0 ){
                        return 1;
                }
                msg ('userstatus', $stopSapServerCmd, $me );
                dbg "Running [$loops of $SAP_SRV_RETRIES] command $stopSapServerCmd\n";
                $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $stopSapServerCmd; echo \\\$? > $resfile' \" >&2");
                chomp ($stopRet=`cat $resfile`);
                unlink $resfile;
                dbg "[7406]: Stop returned $stopRet\n" ;
                sleep $wait;
        }
        my $ret=$self->StatusSapServer($sid, $Instance);
        return 1 if (!$ret);
        # go old school
        my $pid=GetPids("$sid $Instance $SAP_SRV");
        return 1 if ( $pid <= 1 );
        my $srvcmd="kill -$sig $pid";
        dbg "[7406]: INFORMATION: Unable to stop using $stopSapServerCmd using $srvcmd\n";
        system ("$srvcmd $REDIRECT");
        sleep $wait;
        $ret=$self->StatusSapServer($sid, $Instance);
        # end old school
        return 1 if (!$ret);
        error ('sapstopsrvfailed', $sid, $Instance,$me );
        return 0;
}

####
#  Description: Starts the startsapsrv script used with web interfaces and sapcontrol
####
#  Input: status, start, or stop
####
#  Output:
####
#  Returns
#       - 0 - error on action
#       - 1 - success on action
####
# bug7406 (Temporarily keep SAPSRVUtils)
sub SAPSRVUtils
{
        my $func = (caller(0))[3];
        my $self=shift;
        my $action=shift;
        my $sid = $self->SAPSID();
        my $Instance = $self->SAPINST();
        if ($action =~ /start/i ){
                return $self->StartSapServer($sid,$Instance);
        }elsif ($action =~ /stop/i ){
                return $self->StopSapServer($sid,$Instance);
        }
        return $self->StatusSapServer($sid,$Instance);
}
#END bug7406

####
#  Description
#       - get the sapstartsrv script used for this instance
####
#  Input:
####
#  Output:
#       - None
####
#  Returns
# 0 - not found
# 1 - found
sub GetSysSAPSRV
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $sid=$self->SAPSID();
        my $Instance=$self->SAPINST();
        if ( ! -f $SAP_SVCS ){
                msg ('nosuchfile', "$SAP_SVCS", $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Server is not running, use the sapservices file to locate the correct entry
        # using the sapservices fil, start this instances sapstartsrv command
        my @srv=grep /$sid\// && /$Instance\//, `cat $SAP_SVCS 2>/dev/null`;
        # Note: There should be one sapstartsrv per sid/instance, use @srv to
        # log.  Per CI, we will log and error and die
        #
        # 524:Fix for Web Dispatch:
        # If we did not find the sid and Instance in the sapservices file as part of
        # the run path, try finding just the sid and instance in the _Instance_ format
        # this is the format most often found for the Web Dispatch instance
        # TO DO: Can this be refactored
        if (@srv == 0 ){
                # check if the format is Web Dispatch service
                msg ('userstatus', "First pass! No match on 'cat $SAP_SVCS' searching for $sid and $Instance", $me );
                @srv=grep /\/$sid\// && /${sid}_${Instance}_/, `cat $SAP_SVCS 2>/dev/null`;
        }

        if (@srv > 1 ){
                msg ('multisap', $SAP_SVCS, $sid, $Instance, $me);
        }
        chomp ($srv[0]);
        $self->SAPSRVCMD($srv[0]);
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Get the instance profile location for the instance by
#       - parsing the sapstartsrv command from the sapservices file
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - file not found
#       - 1 - file exists and location stored by SAPSRVPF
sub GetSysSAPSRVPF
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;

        # Get the start profile for the instance from the sapstartsrv command in the sapservices file.
        my $srvcmd = $self->SAPSRVCMD();
        (my $srvpf = $srvcmd) =~ s/.*pf=(.*).*$/$1/g;
        $srvpf = (split /\s+/, $srvpf)[0];
        $self->SAPSRVPF($srvpf);
        if (! -f $srvpf) {
                error('nosuchfile', $srvpf, $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

################################################################################
##################--- End SRV Section & sapstartsrv Section ---#################
################################################################################


################################################################################
######################--- Start Helpers Section ---#############################
################################################################################
#General
# Various scripts and checks for assistance with the resource create
# restore, monitoring, remove, etc


####
#  Description
#       - This function takes the object values from the object and uses
#       ins_setinfo to make the permanent in the info field
####
#  Input:
#       - [sys] - system to update info on
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error
#       - 1 - success
sub SAPUpdateLKInfoField
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $sys=shift;

        # block any resource that has not updated the tag using $self->SAPTAG(<value>)
        if ( $self->SAPTAG() eq '' ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # build the required info field using the existing object values
        my $infoField=join "\x02", $self->SAPSID(), $self->SAPINST(), $self->SAPVIP(), $self->SAPMNT, $self->SAPPROTECTION(), $self->SAPRECOVERY(), $self->SAPINSTTYPE();
        my $ret=LK::ins_setinfo($sys, $self->SAPTAG(), $infoField);

        dbg "\tThe command (ins_setinfo) returned ". ($OS_RET{$ret} || $ret ) . " on $me.\n";
        dbg "<= Trace:Leaving $func on $me\n";

        if ($ret){
                msg ('updatefailed', $self->SAPTAG(), $me );
                $ret=0;
        }else{
                msg ('updatesuccess', $self->SAPTAG(), $me );
                $ret=1;
        }
        return $ret;
}

####
#  Description
#       - Set, Remove or Check for the in progress flag
####
#  Input:
#       - Type - flag to check
#               0 - recover
#               1 - quickCheck
#               2 - restore
#               3 - remove
#               4 - move ERS
#       - check_code - is this a check, set or remove (optional)
#               not passed - check flag
#               0 - remove flag
#               1 - set flag
####
#  Output:
#       - None
####
#  Returns
# 0 - flag not found, not created or not removed
sub FlagInProgress
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Type=shift;
        my $check_code=shift;
        my $actionFlag=${$self->SAPFLAGS()}{$Type};

        if ( $check_code ne '' && $check_code ){
                dbg "Creating flag $actionFlag on $me\n";
                LK::flg_create('',$actionFlag);
        }elsif ( $check_code ne '' && !$check_code ){
                dbg "Removing flag $actionFlag on $me\n";
                LK::flg_remove('',$actionFlag);
        }else{
                dbg "\t\tChecking only flag $actionFlag on $me\n";
        }
        my $ret=LK::flg_test('',$actionFlag);
        $ret=$systoperl{$ret};
        $ret=!$ret if ($check_code ne '' && !$check_code );

        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

# clear all flags set
# Input
# - flag to skip (optional)
# Output
# - None
sub ClearAllFlags
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $flagToSkip = shift;
        # If no flag to skip was given, set flagToSkip to -1 to allow == comparison below.
        # Otherwise it will always exclude type 0 because of how == works.
        $flagToSkip = ((defined($flagToSkip) && ($flagToSkip ne '')) ? $flagToSkip : -1);
        my @flagTypes = keys %{$self->SAPFLAGS()};
        foreach my $type (@flagTypes){
                next if ($type == $flagToSkip);
                $self->FlagInProgress($type,0);
        }
        dbg "<= Trace:Leaving $func on $me\n";
}

####
#  Description
#       - Determine the level of monitoring and if this resource
#       - should be monitored by quickCheck
#       - Levels: 0 = Start Only, >=1 includes monitoring
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
# 0 - no monitoring of this resource
# 1 - yes, monitor this resource
sub SAPCanMonitor
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        if ($self->SAPPROTECTION() !~ /$P_LEVEL_0/i ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        msg ('disabled', 'monitoring', 'quickCheck', $self->SAPTAG(), $me );
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Determine the level of recovery and if this resource
#       - should be failed over to a remote system by recover
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
# 1 - yes remote recovery can be initiated by this resource
# 0 - no, this resource should not cause a remote failover
sub SAPCanFailover
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $rec=$self->SAPRECOVERY();

        if ( $RECOVERY_LEVELS{uc($rec)} > $RECOVERY_LEVELS{$R_LEVEL_1} ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        msg ('disabled', 'remote recovery', 'recover', $self->SAPTAG(), $me );
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Determine if a local recover is attempted
#       Recovery Level
#               Full - Local and Remote Recovery
#               LocalOnly - Local Recovery Only
#               RemoteONly - No local Recovery, immediate remote failover
#               Off - No recovery, no failover
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
# 1 - yes local recovery can be initiated for this resource
# 0 - no, local recovery should not be inititated for this resource
sub SAPCanLocalRecover
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $rec=$self->SAPRECOVERY();
        if ($rec !~ /$R_LEVEL_2/ ){ #ignore 0, which is caught before this call
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        msg ('disabled', 'recovery', 'local recover', $self->SAPTAG(), $me );
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Determine the level of recovery and if this resource
#       - can be recovered using sendevent and recover
#       Recovery Level
#               Full - Local and Remote Recovery
#               LocalOnly - Local Recovery Only
#               RemoteONly - No local Recovery, immediate remote failover
#               Off - No recovery, no failover
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
# 1 - yes recovery can be initiated for this resource
# 0 - no, recovery should not be inititated for this resource
sub SAPCanRecover
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $rec=$self->SAPRECOVERY();
        my $pro=$self->SAPPROTECTION();
        if ($rec !~ /$R_LEVEL_0/ && $PROTECTION_LEVELS{uc($pro)} > $PROTECTION_LEVELS{$P_LEVEL_1} ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        msg ('disabled', 'recovery', 'recover', $self->SAPTAG(), $me );
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Determine the level of protection and if this resource
#       - should be stopped during a remove of the resource hierarchy
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
# 0 - don't stop on remove
# 1 - yes stop this instance on remove
sub SAPCanStop
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        if ($self->SAPPROTECTION() !~ /$P_LEVEL_3/i ){
                msg ('disabled', 'shutdown during remove', 'remove', $self->SAPTAG(), $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Stop SAP in the recover process
####
#  Input:
#       - None
####
#  Output
#       - None
####
#  Returns
# 1 - stop SAP during the recover proces
# 0 - Don't stop
sub SAPCanStopInRecover
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $ret=0;

        if ($SAP_STOP_IN_RECOVER ){
                $ret=1;
        }else{
                msg ('disabled', 'shutdown during local recovery', 'recover', $self->SAPTAG(), $me );
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

####
#  Description
#       - Determine whether the given resource has an available
#       - live backup node to attempt to failover to, and whether
#       - conditions appear favorable for a successful failover.
####
#  Input:
#       - None
####
#  Output
#       - None
####
#  Returns
#       0 - no viable remote failover target available
#       1 - viable remote failover target available
sub CanFailoverToBackupNode
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $recovery_lvl = $self->SAPRECOVERY();

        # If SAPRECOVER < 2 (i.e., OFF or LOCAL), then failover is not allowed.
        if ($RECOVERY_LEVELS{uc($recovery_lvl)} < $RECOVERY_LEVELS{$R_LEVEL_2}) {
                dbg "The recovery level for resource $tag is $recovery_lvl on $me. The hierarchy for resource $tag cannot failover to any backup node.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Call the generic hierarchy-level canfailover script to determine
        # whether any remote system that this resource has been extended to
        # is a viable failover target for the hierarchy containing this resource
        my @eqv = LK::eqv_list($me, '', $tag);
        my @remote_eqv_sys = map { (split /\001/, $_)[2] } @eqv;
        foreach my $sys (@remote_eqv_sys) {
                system("$lkroot/lkadm/bin/canfailover -t $tag -s $sys");
                if ($? == 0) {
                        msg('resourcecanfailover', $sys, $tag);
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }

        # If we fall through the loop, then no remote system is a viable failover target.
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Get the resource equivs for a specified tag
####
#  Input:
#       - [hashref] - hash to return equivs
#       - [tag] - resource tag
####
#  Output
#       - None
####
#  Returns
# 1 - has eqvs and got them
# 0 - otherwise
sub GetLKEquiv
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $hashref=shift;
        my $tag=shift;
        my $fullref=shift;
        if ( $tag eq '' ){
                $tag=$self->SAPTAG();
        }
        if ($tag eq '' ){
                error ('valueempty', 'Tag', $func, $me );
                dbg "=> Trace:Entering $func on $me\n";
                return 0;
        }
        my @eqvs=LK::eqv_list($me,'',$tag);
        foreach my $eqv (@eqvs){
                chomp ((my $otherSys, my $otherTag, my $pri, my $otherPri) = (split /\x01/, $eqv)[2,3,5,6]);
                $$hashref{$otherSys}=$otherTag;
                $$fullref{$otherSys}{'Tag'}=$otherTag;
                $$fullref{$otherSys}{'Pri'}=$otherPri;
                $$fullref{$me}{'Pri'}=$pri;
                $$fullref{$me}{'Tag'}=$tag;
        }

        dbg "<= Trace:Leaving $func on $me\n";
        (@eqvs) && return 1;
        return 0;
}

####
#  Description
#       - Check if a tag is in-service
####
#  Input:
#       - tag - tag to check
#       - sys - system to test on
####
#  Output:
#       - None
####
#  Returns
#       1 - tag is ISP
#       0 - tag is not ISP or not found
sub IsTagISP
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $tag=shift;
        my $sys=shift;
        $sys=($sys ne '' )? $sys: $me;
        dbg "\tChecking for $sys and $tag on $me\n";
        my @existins = LK::ins_list($sys, '', '', '', $tag, '');
        my $state;
        # for each instance get the protected value
        foreach my $lkins (@existins){
                chomp( ($state) = (split /\001/, $lkins)[6]);
                dbg "\tFound state=$state on $sys for $tag on $me\n";
                if ( $state =~ /ISP/ ){
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Check if a particular key item is already LifeKeeper protected
#       - This is a generic is a unique item already under LK protection
####
#  Input:
#       - sys - system
#       - app - app type
#       - res - res type
#       - value - value to check if protected
#       - loc - location of value in info field
#       - sep - separator used in info field
#
####
#  Output:
#       - None
####
#  Returns
#       - '' - if not found
#       - <tag> <id> <state> - tag and state of resource protecting value
####
sub isWhatProtected
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $sys=shift;
        my $apptype=shift;
        my $restype=shift;
        my $value=shift;

        my $infolocation=shift;
        my $sep=shift;
        $infolocation=($infolocation ne '' ) ? $infolocation : $DFLTINFOLOC;
        $sep=($sep ne '' ) ? $sep : '\002';

        my @existins;
        my $protectedvalue;
        my $ret;
        if ($value eq ''){
                dbg "<= Trace:Leaving $func on $me\n";
        return '';
        }
        dbg "$func called to search for app=$apptype, res=$restype, value=$value, sep=$sep, and internal LK location=$infolocation on $me.\n";
        # get list of existing resources of app=$apptype and res=$restype
        @existins = LK::ins_list($sys, '', $apptype,$restype, '', '' );
        my $tag;
        my $state;
        my $id;

        # for each instance get the protected value
        foreach my $lkins (@existins){
                chomp( ($tag, $id, $protectedvalue,$state) = (split /\001/, $lkins)[3,4,5,6]);
                chomp($protectedvalue = (split /$sep/, $protectedvalue)[$infolocation]);
                if ($protectedvalue eq $value ){
                        dbg "<= Trace:Leaving $func on $me \n";
                        return ("$tag", "$id", "$state");
                }
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return '';
}

####
#  Description
#       - Determine whether there is already an LK resource
#       - on the given system with the same LK ID. If so,
#       - check whether it has the same app, type, and
#       - SAP instance type. Used by the SAP canextend/extend
#       - scripts.
####
#  Input:
#       - targetTag - tag chosen for the resource on the target system during extend
#       - sys - (optional) system to check for an existing resource on
####
#  Output:
#       - None
####
#  Returns
#       - undefined value - no resource found with same ID
#       - "!similarresourceerror!" - resource found with same ID but
#       -                            different app, type, or SAP instance type
#       - tag name - resource found with same ID, app, type, and SAP instance type
####
sub ResourceAlreadyExtended
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $targetTag = shift;
        my $sys = shift;

        $sys = ((defined($sys) && $sys ne '') ? $sys : $me);
        my $id = $self->SAPLKSAPID();
        my $type = $self->SAPINSTTYPE();

        my $existingIns = LK::ins_list($sys, '', '', '', '', $id);
        if ($existingIns ne '') {
                chomp (my ($existingApp, $existingType, $existingTag) = (split /\001/, $existingIns)[1,2,3]);
                my $existingInstType = (split /\002/, (split /\001/, $existingIns)[5])[6];
                if (($existingApp eq $sap::appType) && ($existingType eq $sap::resType) && ($existingInstType == $type)) {
                        dbg "The resource tag $targetTag already exists with correct app, type, and SAP instance type on $me.";
                        dbg "<= Trace:Leaving $func on $me\n";
                        return $existingTag;  # Found existing resource with correct app, type, and SAP instance type
                } else {
                        baseerror ('SimilarResourceAlreadyExists', $targetTag, $id, $sys, $existingApp, $existingType);
                        dbg "<= Trace:Leaving $func on $me\n";
                        return '!similarresourceerror!';  # Existing resource had incorrect app, type, or SAP instance type
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return;  # No resource found with the same ID
}

####
#  Description
#       - Get and set in the object the SAP administrator's information
#       - the information is retrieved primarily from the passwd file
####
#  Input:
####
#  Output:
#       - None (sets values in the object)
####
#  Returns
#       - 1 - always
####
sub GetSysAdminInfo
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $sid=$self->SAPSID();

        if ($self->SAPADMIN() ne '' && $self->SAPADMINHOME() ne '' && $self->SAPUID() ne '' ){
                msg ('valueempty', 'SAP User Info', $func, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        my $SAPAdmin=lc("${sid}adm");
        $self->SAPADMIN($SAPAdmin);
        my $shell;
        my $SAPAdminHome;
        my $uid;
        my $gid;
        chomp ((my $pwname, my $pwpass, $uid, $gid, my $pwquota, my $pwcomment, my $pwgcos, $SAPAdminHome, $shell)=(getpwnam($SAPAdmin)));

        $self->SAPADMINHOME($SAPAdminHome);
        $self->SAPUID($uid);
        $self->SAPGID($gid);
        $self->SAPSHELL($shell);

        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

#---------------------------------------------------------------------------
# GetSysAdminHome(SID)
# Description:  Returns the SAP administrator's home directory which is
#   retrieved from the /etc/passwd file.
# Arguments:
# Return Values:
#   1 - Always
#
sub GetSysAdminHome
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        if ($self->SAPADMINHOME() eq '' ){
                $self->GetSysAdminInfo();
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}


# Description:
# Take the Instance passed in and determine if it is
# a PAS/CI (DVEBGMS/J*), a CS (SCS/ASCS), a ERS (ERS*)
# or a SAS
# Description:
#       Take the Instance passedin by the caller, or from the
#       object and get the instance number.
#
sub GetSAPInstType
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Type;
        my $Instance=$self->SAPINST();

        # 1. Is it ERS
        if ( $Instance =~ /^ERS/ ){
                $Type = $TYPE_NEW_ERS;
        }elsif ( $Instance =~ /^SCS/ || $Instance =~ /^ASCS/ ){
                $Type = $TYPE_CS;
        }elsif ( $Instance =~ /^$ABAPCI\d{2}/ || $Instance =~ /^$JCI\d{2}/ ){
                $Type = $TYPE_CI;
        }else{
                $Type = $TYPE_AS;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $Type;
}

sub GetSAPInstNum
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Instance=$self->SAPINST();

        (my $instNum=$Instance)=~s/\S+(\d{2})/$1/g;
        dbg "<= Trace:Leaving $func on $me\n";
        return $instNum;
}

# Description:
#       Get the running pids from ps -ewwwf and sort
#       based on the inclusions and exclusions parameters
#       Example: GetPids("LifeKeeeper lcm", "tty");
#                                       matching_criteria, exclusions
#       1. get what matches items in list (0..N)
#       2. exclude items from search in exclusions (0..N)

sub GetPids
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $in=shift;
        my $ex=shift;
        my @in=split /\s+/, $in;
        my @ex=split /\s+/, $ex;
        my @list=`ps --no-headers -ewwwf`;
        my @exc;

        # loop over ps list and keep only what is in the
        # inclusion list
        foreach my $l (@in){
                @exc=grep /$l/, @list;
                @list=@exc;
        }

        # loop over the included list and throw out what
        # should be excluded
        foreach my $l (@ex){
                @list=grep !/$l/, @exc;
                @exc=@list;
        }

        my $pid=(split /\s+/, $list[0])[1];
        if ($pid ne '' && $pid > 1 ){
                #Uncomment next two lines for additional Trace
                        #my @out=`ps -ewwwwwf|grep $pid`;
                        #@out=join "\t", @out;
                        #dbg "\tTrace:GetPids pid found pid=$pid for inclusions=$in (out=@out) on $me.\n";
                dbg "\tTrace:GetPids pid found pid=$pid for inclusions=$in on $me\n";
                dbg "<= Trace:Leaving $func on $me\n";
                return $pid;
        }

        dbg "\tTrace:Nothing found for process $in with exclusions $ex on $me\n";
        dbg "<= Trace:Leaving $func on $me\n";
        return $BADPID;
}

####
#  Description
#       - Find real IP address of a dnsname
#####
#  Input:
#       - Instance - instance to check
#       - loc - field location
####
#  Output:
#       - None
####
#  Returns
#       null - unable to lookup ip by dns name or resolve address
#       address
sub GetRealIP
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        use Socket;
        my $self=shift;
        my $ipordns=shift;
        dbg "Input value for ipordns=$ipordns on $me\n";
        my $aip;
        $aip=gethostbyname($ipordns);
        if ($aip ne '' ){
                $ipordns=inet_ntoa($aip);
                if ($ipordns ne '' ){
                        dbg "<= Trace:Leaving $func on $me\n";
                        return $ipordns;
                }
        }else{
                msg ('cannotfind', $ipordns, $me );
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return '';
}

################################################################################
###########################--- End Helpers Section---###########################
################################################################################



################################################################################
####################--- Start SAP Main Section ---##############################
################################################################################

# NOTE: do NOT call this method (see bug7759)
# this stub method is necessary for rolling upgrades from LK 9.3.1 and earlier,
# since a remote call from the 9.3.1 node to the upgraded node calls this method
# NOTE: this code can be removed once upgrade from 9.3.1 is no longer supported
sub SAPInstanceCmd
{
        my $self = shift;
        my $action = shift;
        my $run_local = shift;
        my $type = $self->SAPINSTTYPE();

        dbg "SAPInstanceCmd $action $run_local $type\n";

        if ($run_local == 0) { # this should never happen
                error('usage', "SAPInstanceCmd(<action>, 1)", "This function exists only to handle pre-9.3.2 rolling upgrades. It should not be used for any other purpose.");
                return 0;
        }

        # during a rolling upgrade from 9.3.1 or earlier, we get an errant call to this method
        # that appears to be a request for local execution, but may be either a request
        # for local or remote execution -- since we don't know, and the most common
        # case (during restore) would be a remote request, we send the request back out
        # for remote execution (see the bug report for details)
        if ($type == $TYPE_ERS) {
                dbg "Calling SAPRemExec(\"SAPInstanceCmd\", \"$action\") to address pre-9.3.2 rolling upgrade issue.";
                return $self->SAPRemExec('SAPInstanceCmd', $action);
        }

        error('usage', "SAPInstanceCmd(<action>, 1)", "This function exists only to handle pre-9.3.2 rolling upgrades. It should not be used for any other purpose.");
        return 0;
}

# Allow extra messages
sub loud
{
        $sap::noise=0;
        $sap::REDIRECT=">&2";
}

# Turn off all extra messages
sub quiet
{
        # When debug is on - it doesn't make since to silence logging
        # so ignore calls to quiet
        if ($SAP_DEBUG){
                return;
        }
        $sap::noise=1;
        $sap::REDIRECT=">/dev/null 2>&1";
}

####
#  Description
#       - Use the sapcontrol utility -function GetProcessList
#       - to determine if the SAP Instance for the SID is running
#       - The sapcontrol utility will check all SAP Instance processes
#       - This requires that the startsapsrv is running first - it will
#       - start it possible
####
#  Input:
#       - [RunLocal] - run on this or another server
#       - [display] - populate object for sap-display
####
#  Output:
#       - sapcontrol output on status
####
#  Returns
#       - 0 - all required instance processes stopped (fully stopped)
#       - 1 - all required instance processes running (fully running)
#       - 2 - at least one required instance process stopped and at least one running (partially stopped)
sub IsInstanceRunning
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;       #added for consistency with interfaces
        my $RunLocal=shift;
        my $nodisplay=shift;
        my $wait=$self->SAPSMALLWAIT();
        my $Instance=$self->SAPINST();
        my $InstanceNum=$self->SAPINSTNUM();
        $nodisplay=($nodisplay eq '' ) ? 0 : $nodisplay;
        $RunLocal=($RunLocal eq '' ) ? 0: $RunLocal;
        my $sapadmin=$self->SAPADMIN();
        my $tmpfile=$self->SAPTMPFILE();
        my $sid=$self->SAPSID();
        my $Type=$self->SAPINSTTYPE();
        my $ret=0;
        @{$self->{$KEYSAPCHECK}{$KEYSAPNAMES}}=();

        dbg "Entry to $func: RunLocal=$RunLocal, Instance=$Instance, InstanceNum=$InstanceNum, sid=$sid and Type=$Type nodisplay=$nodisplay on $me\n";
        if ($Type == $TYPE_ERS && $RunLocal == 0 ){
                $ret=$self->SAPRemExec($func, $nilValue, $nodisplay);
                dbg "<= Trace:Leaving $func on $me\n";
                return $ret;
        }

        if ($InstanceNum eq '' ){
                error ('valueempty', 'InstanceNum', $func, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # 0. Make sure basic NFS shares are accessible
        # IsInstanceRunning is a common entry to most action routines.
        # Check if NFS is running before proceeding with additional SAP
        # checks
        #
        $ret=$self->NFSCheck();
        if ( $ret == 0 ){
                error('nonfs', $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1; #
        }
        $self->LocalNFSMounts();

        # 1. Start the required sapstartsrv
        $self->SAPSRVUtils('start') if ( !$self->{$iThrottle} );
        $self->{$iThrottle}=1;

        # 2. Get the VIP
        my $vip = ($RunLocal ? $me : $self->SAPVIP());

        # 3. GetProcessList
        my $sapProcessList="$SAP_CONTROL_CMD -host $vip -nr $InstanceNum -function GetProcessList";
        my $resfile=$self->SAPRESULTFILE();
        msg ('userstatus', $sapProcessList, $me );
        $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $sapProcessList > $tmpfile 2>&2; echo \\\$? > $resfile' \" $REDIRECT");
        my @out=`cat $tmpfile`;
        my @prettyout=join "\t", @out;
        chomp ($ret=`cat $resfile`);
        msg ('commandReturned', $sapProcessList, $ret, $me);
        dbg "The return code of the $sapProcessList was ret=$ret with output=@prettyout on $me\n";
        unlink $resfile;
        unlink $tmpfile;
        # Need to include a check for sapstart
        my $pid=GetPids("$SAP_START $sid $Instance", "$SAP_SRV");
        if ( $pid eq '' || $pid <= 1 ){
                dbg "Unable to find the pid for $SAP_START for SID $sid and Instance $Instance\n";
                msg ('notrunning', $SAP_START, $me );
        }

        # 4. Check if we can do a quick exit based on successful return;
        if ( $ret == $ALLSTOPPED && !$nodisplay ){
                dbg "\tAll processes are stopped on $me\n";
                msg ('status', $sid, $Instance, 'stopped', $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }elsif ( $ret == $ALLSTARTED && !$nodisplay ){
                msg ('status', $sid, $Instance, 'running', $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        $ret=0;

        # 5. Something was not completely running, or this was a quickCheck
        #       event.  Process the list from the GetProcessList
        my $status;
        my $cnt;
        my %p;
        my $name;
        my $found=0;
        my $last;
        my @total;

        # We are storing items in a prettier format in a list for the UI
        # or b/c of further use in quickCheck
LINE:foreach my $line(@out){
                # 5a. Start by getting the process name (key=name) from the output
                chomp $line;
                my $txt="name";
                if ($line =~ /\d+ $txt/ ){
                        $line=~s/(^\d+)\s+$txt:\s*(\S+)/$1:$2/g;
                        ($cnt, $name)=split /:/, $line;
                        $name=uc($name);        #prettier
                        dbg "\tTrace:The name=$name process was found for instance $Instance on $me\n";
                        push @{$self->{$KEYSAPCHECK}{$KEYSAPNAMES}}, $name;
                        next;
                }
                # 5b. Get the rest of the process values from the output
                foreach $txt ("textstatus", "dispstatus", "pid"){
                        if ($line =~ /\d+ $txt/ ){
                                $line=~s/(^\d+)\s+$txt:\s*(\S+)/$1:$2/g;
                                ($cnt, $status)=split /:/, $line;
                                $self->{$KEYSAPCHECK}{$name}{$txt}=$status;
                                dbg "\tTrace: The $txt status line was obtained for $name of $Instance with value=$status on $me\n";
                                next LINE;
                        }
                }
        }

        # 6. parse through the output of the process hash
        my @total=@{$self->{$KEYSAPCHECK}{$KEYSAPNAMES}};
        $cnt=@total;
        my @notrunning;
        my @running;
        my $ret=0;
        dbg "Trace: The total=@total on $me\n";
        foreach my $k (@total){
                chomp ($k);
                my $color=$self->{$KEYSAPCHECK}{$k}{'dispstatus'};
                $color=uc($color);
                if ( $color =~ /GREEN/i && $self->{$KEYSAPCHECK}{$k}{'pid'} > 1 ){
                                push @running, $k;
                }elsif ($color =~ /RED/i || $color =~ /YELLOW/i ){
                        my $internalColorCmd="SAP${color}";
                        if ( $self->$internalColorCmd() ){
                                dbg "Treating a status of $color returned by $internalColorCmd as a failure of the instance $Instance on $me\n";
                                push @notrunning, $k;
                                $cnt--;
                        }else{
                                dbg "Treating a status of $color returned by $internalColorCmd as okay for the instance $Instance on $me\n";
                                push @running, $k;
                        }
                }elsif ( $color =~ /GRAY/i || $self->{$KEYSAPCHECK}{$k}{'pid'} <= 1 ){
                        msg ('nopidnostatus', $Instance, $self->{$KEYSAPCHECK}{$k}{'textstatus'}, $me );
                        push @notrunning, $k;
                        $cnt--;
                }
        }
        # 7. Determine what our return values should be, and return
        msg ('checksummary', $func, $Instance, "@running", "@notrunning", "@total", $me );
        if ( $cnt == @total && @total > 0 ){
                msg ('instancerunning', $sid, $Instance, $me );
                $ret=1;
        }elsif ( $cnt == 0 || @total == 0 ){
                msg ('instancenotrunning', $sid, $Instance, $me );
        }else{
                # To fall into this block, we must have 0 < $cnt < @total
                msg ('instancepartiallystopped', $sid, $Instance, $me);
                $ret=2;
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

# clean up orphaned shared mem for InstanceNum
sub cleanipc
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $sapadmin=$self->SAPADMIN();
        my $InstanceNum=$self->SAPINSTNUM();
        dbg "$func: Running cleanipc $InstanceNum remove as $sapadmin on $me\n";
        system ("su - $sapadmin -c \"sh -c 'LANG=C cleanipc $InstanceNum remove' \" $REDIRECT");
        dbg "<= Trace:Leaving $func on $me\n";
}

# clean up orphaned pids for SID/Instance
sub cleanOrphans
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;
        my $wait=$self->SAPSMALLWAIT();
        my $pid;
        my $sig=15;
        my $sid=$self->SAPSID();
        my $Instance=$self->SAPINST();
        my $retries=$SAP_CLEAN_RETRIES;
        my @checker;
        my @list;
        my @ps;
        my $cnt=0;
        my $ret;
        dbg "$func: pre-entry status retries=$retries, action=$action on $me\n";
        while ($cnt <= $retries ){
                @list=();
                dbg "$func: Retry Loop:retries=$retries - cnt=$cnt - action=$action ($$) on $me\n";
                @ps=grep /$sid/ && /_$Instance/ && !/sapstart/ && /pf=/ && !/$lkroot/, `ps --no-headers -ewwwwwf 2>/dev/null`;
                if (@ps == 0 ){
                        dbg "$func: Unable to find the pids for sid=$sid and Instance=$Instance on $me\n";
                        last;
                }
                foreach my $pid (@ps){
                        $pid=(split /\s+/, $pid)[1];
                        @checker=grep /$pid/ && /$sid/ && /$Instance/, `ps --no-headers -wwwwwf -p $pid 2>/dev/null`;
                        if (@checker){
                                dbg "$func: The pid $pid is still running after intial $action attempt and will be added to a kill list on $me\n";
                                push @list, $pid;
                        }
                }
                if ( @list ){
                        msg ('usingkill', "@list", "misc. processes", $sid, $Instance, "kill $sig, $pid", $me );
                        kill $sig, @list;
                        sleep $wait;
                }
                $cnt++;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Generic Stop entry point for any Instance type
####
#  Input:
#       - RunLocal - run commands on local server (1) or remote (0)
####
#  Output:
#       - None
####
#  Returns
# 0 - failed
# 1 - ok
sub StopInstance
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;
        my $RunLocal=shift;
        if ($action eq '' ){
                $action="stop";
                dbg "INFORMATION: No action specified by " . calledby() . " . Action set to $action on $me (". __FILE__ .":$func:". __LINE__ .")";
        }
        my $wait=$self->SAPSMALLWAIT();
        $RunLocal=($RunLocal eq '' ) ? 0: $RunLocal;
        my $Instance=$self->SAPINST();
        my $InstanceNum=$self->SAPINSTNUM();
        my $Type=$self->SAPINSTTYPE();
        my $Tag=$self->SAPTAG();
        my $sid=$self->SAPSID();
        my $sapadmin=$self->SAPADMIN();
        my $resfile=$self->SAPRESULTFILE();
        my $cmdWait=$SAP_STOP_WAIT;
        my $endcriteria=0;      # what return code marks this action succesful
        my $retries=2;          # how many retries for this action; default none
        my $ret=0;

        my $viphost=$self->SAPVIP();
        dbg "$func: Arguments: Parameters: RunLocal=$RunLocal, Instance=$Instance, InstanceNum=$InstanceNum, sid=$sid and Instance action=$action Type=$Type on $me (". __FILE__ .":$func:". __LINE__ .")";

        if ($Type == $TYPE_ERS ) {
                if ( $RunLocal == 0 ){
                        $ret=$self->SAPRemExec($func, $action);
                        dbg "SAPRemExec call for action $action returned (ret=$ret) on $me (". __FILE__ .":$func:". __LINE__ .")";
                        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                        return $ret;
                }
                $retries=1;
                dbg "Type = $Type ($TYPE_ERS) with RunLocal=$RunLocal.  Setting retries on $me (". __FILE__ .":$func:". __LINE__ .")";
        }

        my $instanceCmd="$SAP_CONTROL_CMD -host $viphost -nr $InstanceNum -function StopWait $cmdWait $SAP_CONTROL_DELAY";
        dbg "Running $instanceCmd for $sid and $Instance via user $sapadmin on $me (". __FILE__ .":$func:". __LINE__ .")";
        # Reset throttling to enable the StopSapServer routine to retry starting the startsapsrv process
        # needed for the instance
        $self->{$iThrottle}=0;
        my $cnt=1;
        while ($cnt <= $retries ){
                $ret=$self->IsInstanceRunning($action,$RunLocal);
                if ( $ret == 0 ){
                        dbg "INFORMATION: IsInstanceRunning returned that the instance was stopped on $me (ret=$ret) (". __FILE__ .":$func:". __LINE__ .")";
                        last;
                }
                ($cnt) && dbg "INFORMATION: Issuing retry [ $cnt of $retries ] for sid=$sid and Instance=$Instance (". __FILE__ .":$func:". __LINE__ .")";
                msg ('userstatus', $instanceCmd, $me );
                $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $instanceCmd; echo \\\$? > $resfile' \" $REDIRECT");
                chomp (my $cmdRet=`cat $resfile`);
                unlink $resfile;
                msg ('commandReturned', $instanceCmd, ($OS_RET{$cmdRet} || $cmdRet ), $me);
                sleep $wait;
                $cnt ++;
                quiet();
        }

        # Handle falling out the loop with SAP running
        if ($ret){
                dbg "Loop fallout (ret=$ret).  Checking again on $me (". __FILE__ .":$func:". __LINE__ .")";
                $ret=$self->IsInstanceRunning($action,$RunLocal);
                if ($ret){
                        dbg "SAP instance did not stop cleanly (ret=$ret).  Calling cleanOrphans on $sid and $Instance on $me (". __FILE__ .":$func:". __LINE__ .")";
                        $self->cleanOrphans("stop");
                }
        }
        dbg "This is the end.  The ret=$ret.  We will finish with cleanipc and StopSapServer (". __FILE__ .":$func:". __LINE__ .")";
        if (($Type == $TYPE_CS) && (calledby() eq 'recover') && ($self->CheckHeldLocks())) {
                dbg "Skipping call to cleanipc during local recovery of resource $Tag since this would delete backup locks being stored in shared memory on $me.";
        } else {
                $self->cleanipc() if (!$ret); # Only call cleanipc if the instance was cleanly stopped
        }

        # bug7669/7693 - We only want to stop the sapstartsrv process for the instance if the HA
        # connector is not configured. This is to allow users using the SAP management console
        # to restart the instance directly from the console (which is not possible if we stop
        # the sapstartsrv process).
        my $HAActive = 1;
        my $instanceCmd = "$SAP_CONTROL_CMD -host $viphost -nr $InstanceNum -function HAGetFailoverConfig";
        my @HGFC = $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $instanceCmd'\" 2>/dev/null");
        $HAActive = 0 if (grep {/^HAActive\s*:\s*FALSE/i} @HGFC);
        # We only stop the sapstartsrv process if the HA connector is not configured (HAActive = 0).
        if (!$HAActive) {
             dbg "Based on HAGetFailoverConfig output, HAActive is $HAActive.  Running StopSapServer for SID $sid and instance $Instance on $me (". __FILE__ .":$func:". __LINE__ .")";
             $self->StopSapServer($sid,$Instance);
        }

        dbg "Ret=$ret on $me (". __FILE__ .":$func:". __LINE__ .")";
        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
        return 1 if ($ret == 0);
        return 0;
}

####
#  Description
#       - Generic Start entry point for any Instance type
####
#  Input:
#       - RunLocal - run commands on local server (1) or remote (0)
####
#  Output:
#       - None
####
#  Returns
# 0 - failed
# 1 - ok
sub StartInstance
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;
        my $RunLocal=shift;
        if ($action eq '' ){
                $action="start";
        }
        my $wait=$self->SAPSMALLWAIT();
        my $wait=$self->SAPSMALLWAIT();
        $RunLocal=($RunLocal eq '' ) ? 0: $RunLocal;
        my $Instance=$self->SAPINST();
        my $InstanceNum=$self->SAPINSTNUM();
        my $Type=$self->SAPINSTTYPE();
        my $Tag=$self->SAPTAG();
        my $sid=$self->SAPSID();
        my $sapadmin=$self->SAPADMIN();
        my $resfile=$self->SAPRESULTFILE();
        my $cmdWait=$SAP_START_WAIT;
        my $retries=2;
        my $ret=0;

        my $viphost=$self->SAPVIP();
        my $instanceCmd="$SAP_CONTROL_CMD -host $viphost -nr $InstanceNum -function StartWait $cmdWait $SAP_CONTROL_DELAY";
        dbg "$func: Arguments: Parameters: RunLocal=$RunLocal, Instance=$Instance, InstanceNum=$InstanceNum, sid=$sid and Instance action=$action Type=$Type on $me (". __FILE__ .":$func:". __LINE__ .")";

        if ($Type == $TYPE_ERS ){
                if ( $RunLocal == 0 ){
                        $ret=$self->SAPRemExec($func, $action);
                        dbg "SAPRemExec call for action $action returned (ret=$ret) on $me (". __FILE__ .":$func:". __LINE__ .")";
                        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                        return $ret;
                }
                $retries=1;
        }

        # Reset throttling to enable the StartSapServer routine to retry starting the startsapsrv process
        # needed for the instance
        $self->{$iThrottle}=0;
        dbg "INFORMATION: IsInstanceRunning returned $ret (". __FILE__ .":$func:". __LINE__ .")";
        msg ('startinstance', $sid, $Instance, $me );
        my $cnt=1;
        while ($cnt <= $retries ){
                # We will check if the database is running on and accessible
                # If the database is not running.  If the DB_CHECK_CRITICAL variable
                # is not set, we log a message.  When the start SAP command
                # is executed, the database is started, if at all possible.
                # Note: We will skip DB checks for ERS, any stop action
                # and after the intial try
                if ( $Type != $TYPE_ERS && $cnt == 0 ){
                        if ( ! $self->CheckDB()){
                                msg ('dbnotopen', $self->SAPDBNAME(), $sid, $Instance, $me );
                                if ($SAP_DB_CHECK_CRITICAL){
                                        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                                        return 0;
                                }
                        }
                }
                $ret=$self->IsInstanceRunning($action,$RunLocal);
                if ( $ret == 1 ){
                        dbg "INFORMATION: Start successful.  IsInstanceRunning returned running and started on $me (ret=$ret) (". __FILE__ .":$func:". __LINE__ .")";
                        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                        return 1;
                }

                dbg "Running [$cnt of $retries] of $instanceCmd for $sid and $Instance via user $sapadmin on $me (". __FILE__ .":$func:". __LINE__ .")";
                msg ('userstatus', $instanceCmd, $me );
                $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $instanceCmd; echo \\\$? > $resfile' \" $REDIRECT");
                chomp (my $cmdRet=`cat $resfile`);
                unlink $resfile;
                msg ('commandReturned', $instanceCmd, ($OS_RET{$cmdRet} || $cmdRet ), $me);
                sleep $wait;
                $cnt ++;
        }
        # Handle fall out after the final start
        $ret=$self->IsInstanceRunning($action,$RunLocal);
        dbg "Ret=$ret on $me (". __FILE__ .":$func:". __LINE__ .")";
        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
        return 1 if ($ret == 1);
        return 0;
}

###
#  Description:
#        - This routine forcefully kills the processes associated to an instance by
#          performing the following sequence of actions:
#        - 1. Call sapcontrol GetProcessList to get a list of PID's of running processes
#             for the instance
#        - 2. Kill (SIGKILL) the 'sapstart pf=<instance pf>' process if it is running
#        - 3. Kill (SIGKILL) the running processes found in step 1
#        - 4. Stop the 'sapstartsrv pf=<instance pf>' process
#  Input
#       - run_local - Whether to run this on the local node or the backup node for TYPE_ERS
#           - 0 - Run on the backup node (default)
#           - 1 - Run locally
#  Returns
#       - 0 - Error, not all processes could be killed
#       - 1 - All processes killed
###
sub KillInstanceProcesses
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me (". __FILE__ .":$func:". __LINE__ .")";
        my $self = shift;
        my $run_local = shift;
        $run_local = ((defined($run_local) && ($run_local ne '')) ? $run_local : 0);

        my $inst = $self->SAPINST();
        my $inst_num = $self->SAPINSTNUM();
        my $sid = $self->SAPSID();
        my $vip = $self->SAPVIP();
        my $sap_admin = $self->SAPADMIN();
        my $type = $self->SAPINSTTYPE();
        my $tmp_file = $self->SAPTMPFILE();

        my $ret = 1;

        if (!$run_local && ($type == $TYPE_ERS)) {
                $ret = $self->SAPRemExec($func, '');
                dbg "SAPRemExec call for routine $func returned (ret=$ret) on $me (". __FILE__ .":$func:". __LINE__ .")";
                dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                return $ret;
        }

        my $retries = 2;

        # Check that sapstartsrv is running, start it if not.
        if (!$self->StatusSapServer()) {
                if (!$self->StartSapServer()) {
                        dbg "Unable to start the sapstartsrv process for instance $inst, returning 1 from $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                        dbg "=> Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
                        return 0;
                }
        }

        # Call sapcontrol GetProcessList to get a list of PID's of running processes
        # associated with the instance.
        my $process_list_cmd = "$SAP_CONTROL_CMD -format script -host $vip -nr $inst_num -function GetProcessList";
        $self->NFSCheckBeforeCmd("su - $sap_admin -c \"sh -c 'LANG=C $process_list_cmd > $tmp_file 2>&2' \" $REDIRECT");
        my @out = `cat $tmp_file`;
        unlink $tmp_file;

        # If a line in the GetProcessList output has the form:
        # [int] textstatus : <state>
        # where <state> is not 'Stopped', then we look for the next line of the form:
        # [int] pid : <pid>
        # and add <pid> to the @pid array.
        my @pid = ();
        my $found_pid;
        my @split_line;
        my $get_next_pid = 0;
        foreach my $line (@out) {
                @split_line = split /:/, $line;
                # In this block, we are checking whether the current line has the form
                # [int] textstatus : <state>
                # If <state> is not 'Stopped', then we will add the next PID we find to @pid.
                if (($split_line[0] =~ /^\s*\d+\s*textstatus\s*$/i) && ($split_line[1] !~ /^\s*Stopped\s*$/i)) {
                        $get_next_pid = 1;
                        next;
                }
                # In this block, we are checking whether the current line has the form
                # [int] pid : <pid>
                # If get_next_pid = 1 and we find a line like this, we add <pid> to @pid.
                if ($get_next_pid && ($split_line[0] =~ /^\s*\d+\s*pid\s*/i)) {
                        # Trim leading/trailing whitespace from the substring containing the pid.
                        ($found_pid = $split_line[1]) =~ s/^\s*(\d+)\s*$/$1/g;
                        dbg "Adding pid $found_pid to running process array in $func (". __FILE__ .":$func:". __LINE__ .")";
                        push @pid, $found_pid if (($found_pid ne '') && ($found_pid > 1));
                        $get_next_pid = 0;
                }
        }

        # Kill the sapstart process associated to the instance.  We need to do this first
        # because otherwise sapstart will immediately restart any processes that use the
        # Restart_Program parameter in the instance profile (e.g., msg_server in ASCS).
        my $sapstart_pid = GetPids("$SAP_START $sid $inst", "$SAP_SRV");
        if (($sapstart_pid ne '') && ($sapstart_pid > 1)) {
                for (my $i=1; $i <= $retries; $i++) {
                        dbg "Killing $SAP_START process for SID $sid and instance $inst with PID $sapstart_pid, try $i of $retries (". __FILE__ .":$func:". __LINE__ .")";
                        kill 'KILL', $sapstart_pid;
                        sleep $SAP_KILL_INSTANCE_PROCESSES_WAIT;
                        last if (! kill 0, $sapstart_pid); # Break out of the loop if the process is no longer running
                }
                # If we were unable to kill the sapstart process then we will return 0 since
                # sapstart may automatically restart processes that use Restart_Program.
                if (kill 0, $sapstart_pid) {
                        error('unabletokill', "the $SAP_START process", $sid, $inst, $sapstart_pid);
                        $ret = 0;
                }
        }

        dbg "Killing the following running processes for instance $inst: @pid (". __FILE__ .":$func:". __LINE__ .")";
        if (@pid) {
                for (my $i=1; $i <= $retries; $i++) {
                        dbg "Killing running processes for SID $sid and instance $inst with PIDs @pid, try $i of $retries (". __FILE__ .":$func:". __LINE__ .")";
                        kill 'KILL', @pid;
                        sleep $SAP_KILL_INSTANCE_PROCESSES_WAIT;
                        last if (! kill 0, @pid); # Break out of the loop if the processes are no longer running
                }
                # If any process is still running after the previous loop then we return 0.
                if (kill 0, @pid) {
                        my $pid_list = join ", ", @pid;
                        error('unabletokill', 'at least one running process', $sid, $inst, $pid_list);
                        $ret = 0;
                }
        }

        # Stop sapstartsrv.
        $self->StopSapServer();

        dbg "<= Trace:Leaving $func on $me (". __FILE__ .":$func:". __LINE__ .")";
        return $ret;
}

################################################################################
######################--- End SAP Main Section ---##############################
################################################################################

################################################################################
######################--- Start SAP DB Section ---##############################
################################################################################
####
#  Description
#       - Simple SAP Check determines if the database for the SAP SID & Instance
#       - is accessible.  This check is used during the StartInstance function
#       - b/c errors are caused if the database is not running during start up
#       - This function does not start the database, that should be done by
#       - A LifeKeeper resource on the database host
#       -
#       - [PreReq]: The saphostexec command must be running before these checks can be done
####
#  Input:
#       - None (object only)
####
#  Output:
#       - None
####
#  Returns
#       - 0 - database not running
#       - 1 - database running
####
sub CheckDB
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;

        if ( $SAP_DB_CHECK_IGNORE ){
                msg ('disabled', 'database checks', $func, $self->SAPTAG(), $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        my $sapadmin=$self->SAPADMIN();
        my $dbCmd=$sap::SAP_HOSTCTL_CMD;
        my $outfile=$self->SAPTMPFILE();

        my $db=$self->SAPDBNAME();
        my $dbtype=$self->SAPDBTYPE();
        my $dbhost=$self->SAPDBHOST();
        my $fnd="Database Status:";
        my $ret=0;
        my %eqvs;
        my %fulleqvs;

        # 1. We need to find out the real address of the ip used for the database
        # host, and if it is LK protected in this cluster
        chomp (my $ip=$self->GetRealIP($dbhost));
        (my $iptag, my $ipid , my $ipstate ) = isWhatProtected($me, 'comm', 'ip', $ip, $IPINFOLOC, '\s+');

        # 2. If it is LK protected, then we check if it is ISP on $me
        #       and issue the start command on $me
        #       otherwise use lcdremexec to issue the start command on one
        #       of the equivalent systems in the cluster until all are
        #       exhausted
        if ( $iptag ne '' ){
                if ( $ipstate =~ /IS/) {
                                $self->HostExec('restart');
                }else{
                        $self->GetLKEquiv(\%eqvs, $iptag, \%fulleqvs);
                        my $tag=$self->SAPTAG();
                        my $Instance=$self->SAPINST();
                        foreach my $sys (keys %eqvs){
                                dbg "Checking if the iptag is ISP on $sys on $me\n";
                                next if ($self->IsTagISP($eqvs{$sys}, $sys) == 0 );
                                dbg "The tag $iptag is ISP on $sys on $me\n";
                                my $isalive=LK::sys_getstate('', $sys );
                                if ($isalive != $LK::ALIVE ){
                                        dbg "\tThe clustered pair with equivalency to $tag is not alive on $me\n";
                                        error ('pairdown', $sys, $tag, $me);
                                        next;
                                }
                                system ("$lkroot/bin/lcdremexec -e -d $sys -- \"$APPBIN/remoteControl $tag $Instance \"'HostExec'\" \"'restart'\" \"$me\" \"$tag\" \"$nilValue\" 2>&2;\"");
                        }
                }
        }else{
                # Note: If the database is not local or in cluster, then the
                # function calls will probably fail
                msg ('nolkdbhost', $dbhost, $me );
        }

        # 3. Once the hostexec is started on the correct server, or all have been
        #       exhausted, try the status command
        my $tmpCmd="$dbCmd -function GetDatabaseStatus -dbname $db -dbtype $dbtype -host $dbhost";
        dbg ('userstatus', $tmpCmd, $me );
        system ("su - $sapadmin -c \"sh -c 'LANG=C $tmpCmd > $outfile 2>&2; ' \" $REDIRECT");
        chomp (my @out=`cat $outfile`);
        unlink $outfile;
        foreach my $st (grep /$fnd/i, @out){
                chomp ($st);
                if ( $st =~ /$fnd\s+Running/i ){
                        msg ('dbup', $db, $self->SAPINST(), $me );
                        $ret=1;
                        last;
                }elsif ($st =~ /$fnd\s+Warning/){
                        msg ('dbdown', $db, $self->SAPINST(), $me );
                }
        }
        @out=join "\t\t", @out;
        msg ('commandReturned', $tmpCmd, $ret, $me );
        dbg "The database check command $tmpCmd produced a status output=@out on $me\n";
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

################################################################################
########################--- End SAP DB Section ---##############################
################################################################################


################################################################################
#####################--- Start SAP OSCOLLECTOR Section ---######################
################################################################################

####
#  Description
#       - Start/Monitor/Stop the SAP OS Collector
####
#  Input:
#       - action - start, status, stop
####
#  Output:
#       - None
####
#  Returns
#       - 0 - failed to stop
#       - 1 - stopped
####
sub OSCollector
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;
        my $wait=$self->SAPSMALLWAIT();
        my $ret=1;
        my $endcriteria=1;
        my $OSCmd;
        my $endCmd;     # set in stop and start blocks below
        my $sig=9;
        my $retries=-1;
        my $tmpfile=$self->SAPTMPFILE();
        my $sapadmin=$self->SAPADMIN();


        my $pid=GetPids("$SAP_OSCOL_CMD $SAPOSCOL");
        $self->SAPCOLLPID($pid);
        dbg "\tThe $SAPOSCOL has pid=$pid on $me\n";

        my $statusCmd="$SAP_OSCOL_CMD -s pf=$SAP_HOSTCTL_PROFILE";
        msg ('userstatus', $statusCmd, $me );
        system ("LANG=C $statusCmd 2>&1 | grep -q 'not attached'"); #dog
        my $serr=($? >> 8 );
        if ( $pid < 1 || $serr == 0 ){
                $ret=0;
                dbg "\tThe $SAPOSCOL command is not running on $me\n";
        }else{
                dbg "\tThe $SAPOSCOL command is running on $me\n";
        }

        if ( $action =~ /start/i ){
                $OSCmd="$SAP_OSCOL_CMD -l pf=$SAP_HOSTCTL_PROFILE";
                if ($ret == 0 ){
                        $retries=1;
                }
        }elsif ( $action =~ /stop/i ){
                $OSCmd="$SAP_OSCOL_CMD -k pf=$SAP_HOSTCTL_PROFILE";
                $endcriteria=0;
                $endCmd= "kill";
                if ($ret ){
                        $retries=1;
                }
        }
        my $cnt=0;
        my $update=0;
        dbg "$func: pre-entry status ret=$ret, retries=$retries, action=$action, endcriteria=$endcriteria on $me\n";
        my $pid;
        while ( $cnt <= $retries ){
                dbg "$func Retry Loop:retries=$retries - cnt=$cnt - endcriteria=$endcriteria - action=$action on $me\n";
                msg ('userstatus', $OSCmd, $me );
                system ("LANG=C $OSCmd $REDIRECT"); #dog
                sleep $wait;
                $pid=$self->SAPCOLLPID();
                if ($endCmd ne '' && $pid > 1 ){
                        msg ('usingkill', $pid, $SAPOSCOL, $self->SAPSID(), $self->SAPINST(), "kill $sig, $pid", $me );
                        kill $sig, $pid;
                }
                $ret=$self->OSCollector("status");
                last if ( $ret == $endcriteria );
                $cnt++;
                $update++;
                # the loop counter has exceeded the throttle counter
                # reset the counter, and allow messages on next iteration
                if ($update > $SAP_MSG_THROTTLE ){
                        loud();
                        $update=0;
                }else{
                        quiet();
                }
        }
        loud();
        dbg "$func ret=$ret and the endcritiera=$endcriteria on $me\n";
        dbg "=> Trace:Leaving $func on $me\n";
        return 1 if ($ret == $endcriteria);
        return 0;
}

################################################################################
######################--- End SAP OSCOLLECTOR Section ---#######################
################################################################################


################################################################################
######################--- Start SAP File System Section ---#####################
################################################################################
#---------------------------------------------------------------------------
# FindActualPath(Path)
# Description:
#       Parses mounted Paths, looking for links and resolving them into actual paths.
#   Also looks for mounts to NFS mounted filesystems.  This function
#   Will return the actual path found.  Returned format for the
#   actual path is:
#       [NFS server:]<actual path>, or
#       <actual path>
#   [NFS server:] is omitted if the mount point does not resolve to an NFS
#   share.  Relative paths or paths containing relative directories,
#   i.e. /usr/sap/../ are not supported.  It also doesn't handle paths
#   containing directory names with spaces in them.
# Arguments:
#   Path - path being examined.
# Return:
#       success - physical path
#       failed  - ""
sub FindActualPath
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        use Cwd;
        my $self=shift;
        my $dir=shift;
        if ( $dir eq '' ){
                dbg "<= Trace:Leaving $func on $me\n";
                return '';
        }
        my @mnttab = grep /\s+$dir\s+/, `cat $MTAB 2> /dev/null`;
        my %index;
        my $elem;
        my @devices;
MTAB_ENTRY: foreach(@mnttab) {
                chomp ($_);
                (my $dev, my $mp, my $fs, undef) = split; # lklin000062
                # if a dev is a symbolic link then walk through each link
                while (defined($dev)) {
                        if (defined($index{$dev})) {
                                foreach $elem (@{$index{$dev}}) {
                                        next MTAB_ENTRY if($mp eq $elem);
                                }
                                push @{$index{$dev}}, $mp;
                        } else {
                                push @{$index{$dev}}, $mp;
                                push @devices, $dev;
                        }

                        my $real_dev = Cwd::realpath("$dev");
                        if ($real_dev eq $dev) {
                                undef $dev;
                        } else {
                                $dev=$real_dev;
                        }
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        # May 12 2011: multiple devices for the same mount point
        # are not allowed
        if (@devices > 1 ){
                error ('multimp', $dir, $me );
                return '';
        }
        my $Path=$devices[0];
        return $Path if (grep /\S+:\S+/, @devices );
        $Path=@{$index{$Path}}[0];
        return $Path;
}


#---------------------------------------------------------------------------
# GetMountedFS(Path)
# Description:  Parses Path, looking for a filesystem
#   that is mounted on the server.  If any part of the path is found to be
#   a mounted filesystem, the path is echo'ed.
# Arguments:
#   Path - path to filesystem being tested.
# Output:  mounted path or "" if links contain relative paths.
# Return Values: 0 = success, non-0 = failure.
#
sub GetMountedFS
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Path=shift;
        my $Sys=shift;
        my $required=shift;
        $Sys=($Sys ne '' )? $Sys : $me;
        $required=($required ne '' )? $required : 1;
        use lib "$ENV{LKROOT}/lkadm/subsys/gen/filesys/bin";
        use filesysBase;
        my $mountPoint;
        my $fsTag;
        msg ('userstatus', "$func, IsPathOnSharedFS for $Path", $me );
        my $ret=filesysBase::IsPathOnSharedFS($Sys, '', $Path, \$mountPoint, \$fsTag);
        if ($ret != 0 ){
                if ( $required ){
                        # If IsPathOnSharedFS fails the mountPoint will be null anyways
                        # but make it explicit that it is not shared when required to be so
                        $mountPoint='';
                        error ('notshared', $Path, $me );
                }else{
                        msg ('notshared', $Path, $me );
                }
                dbg "\tShared check for $Path on $Sys failed (ret=$ret) (fsTag=$fsTag) (mountPoint=$mountPoint) on $me\n";
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return $mountPoint;
}

####
#  Description
#       simple routine called by restore to get current
#       time of restore using date command
####
#  Input
#       - None
#  Output
#       - None
#  Return
#       - date output
####
sub GetSysCurrentTime
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        chomp (my $v=`date +%s`);
        dbg "<= Trace:Leaving $func on $me\n";
        return $v;
}

#---------------------------------------------------------------------------
####
#  Description
#   - Returns the SAP root directory (/sapmnt is the default
#   - path when you install SAP, but the user can change it).
#   - Find it by tracing a known link to the profile directory, i.e.
#   - /usr/sap/<SID>/SYS/profile -> <SAP_ROOT>/<SID>/profile.
####
#  Input:
####
#  Output:
#       - set SAPMNT to the SAP Root Directory
####
#  Returns
#       - 1 - ok
#       - 0 - error
#
####
sub GetSysSAPMntDir
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        if ($self->SAPMNT() ne '' ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        # this is looking for the mounted dir in /, use readlink?
        my $sapdir=$self->SAPDIR();
        my $sid=$self->SAPSID();
        if ($sid eq '' ){
                dbg "Error ++++ <sid> ++++ was empty on $me\n";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        dbg "performing readlink of $sapdir/$sid on $me\n";
        chomp (my $SAPMntDir=readlink "$sapdir/$sid/SYS/profile");
        $SAPMntDir=~s/\/$sid\/.*//g;
        if ($SAPMntDir eq '' && -f "$sapdir/$sid/SYS/profile" ){
                dbg "Error finding <sapmnt> directory on $me\n";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        $self->SAPMNT($SAPMntDir);
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

#---------------------------------------------------------------------------
# NFSCheck()
# Description:  Call this function prior to doing system operations that
#   tend to hang if NFS shares are not available due to a NFS server that
#   is down.  It uses the pingnfs utility.
# Arguments:
#   Tag - Tag name of the SAP hierarchy.
# Return Values:
#   1 - everything OK
#   0 - NFS not running.
#
sub NFSCheck
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Tag=$self->SAPTAG();

        # If we can't find pingnfs, then don't do anything else.
        if (! -x "$PINGNFS" ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;       #Can't find pingnfs, do we lie and say all is okay?
        }

        my %mtabServers;
        my $nfsTime=5;

        # Limit the nfs remote mount checking to mounts critical to SAP as
        # configured via the SAP_NFS_CHECK_DIRS tunable in /etc/default/LifeKeeper.
        # The tunable contains a comma separated list of directories to check
        # and see if a nfs remote mount exists at that location.  AWS EFS mounts
        # should not be include in the tunable list as they will by default look
        # just like any nfs mount but will fail the pingnfs test and result in
        # a failure.  - bug 7134
        return 1 if (!open(MTAB, $MTAB));
        my @mounts = <MTAB>;
        close MTAB;
        foreach my $mtabLine (@mounts) {
                chomp ($mtabLine);
                my ($mountDev, $mountLoc, $mountType) = (split /\s+/, $mtabLine)[0,1,2];
                # Skip entries that do not have a device in the nfs format of
                # server:directory and do not have the mount type of nfs.  This
                # eliminates mount types such as sshfs which matches the nfs
                # device format but has sshfs for the mount type vs. nfs.
                next if (($mountDev !~ /^\S+:\S+/) || ($mountType !~ /^nfs/));
                dbg "Checking nfs mount $mountDev that is mounted on $mountLoc\n";
                # Check if the mtab entry matches a directory listed in the
                # tunable SAP_NFS_CHECK_DIRS.
                foreach my $checkThis (split /,/, $SAP_NFS_CHECK_DIRS) {
                        next if ($checkThis ne $mountLoc);
                        my ($mountServer)=(split /:/, $mountDev)[0];
                        $mtabServers{$mountServer}=1;
                }
        }

        # Next, check to see if NFS is available for each one that is related
        # to SAP.
        my $ret=1;
        my $dbg=1;
        my $PINGCMD;
        my @serverList=keys %mtabServers;
        dbg ("status: $PINGNFS -s < @serverList > -t $nfsTime", $me );
        foreach my $server (@serverList){
                $PINGCMD="$PINGNFS -s $server -t $nfsTime";
                system ("$PINGCMD 2>/dev/null");
                $ret=($? >> 8 );
                if ( $ret != 0  ){
                        msg ('nfsdown', "$server", $self->SAPSID(), $self->SAPINST(), $me );
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 0;
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

#---------------------------------------------------------------------------
# LocalNFSMounts
# Description:  Call this function to locally mount NFS exports needed
#   by the SAP CI.
# Arguments:
#   Tag - Tag name of the SAP resource
# Return Values: 1
sub LocalNFSMounts
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $Tag=$self->SAPTAG();
        # Bug3260,7339: Fix LocalNFSMounts / Fix LocalNFSMounts (simplify)
        my $sid=$self->SAPSID();
        my $Instance=$self->SAPINST();
        my $sapmnt=$self->SAPMNT();
        my $sapmntSID="${sapmnt}/${sid}";
        my $usrmntSID="$SAPDIR/${sid}";
        my $usrmntInst="$SAPDIR/${sid}/${Instance}";
        my @ret;
        my $ret=1;
        my $mtab="/etc/mtab";
        my $procMounts="/proc/mounts";
        my $foundInFstab=0;
        my $sapmntToMount=$sapmntSID;
        my @mounts;
        # Check if there is an entry we can use in $FSTAB
        foreach my $testFstab ($sapmnt, $sapmntSID, $SAPDIR, $usrmntSID, $usrmntInst ){
                @mounts=grep /^\S+:\S+/ && /\s+$testFstab\s+/ && !/^#/, `cat $FSTAB`;
                if (@mounts == 0 ){
                        dbg "There was no mount listed or additional information detected in $FSTAB for $testFstab.\n";
                        next;
                }
                dbg "Found the mount listed in $FSTAB for $testFstab.\n";
                $sapmntToMount=$testFstab;
                # If it is already mounted, skip it
                foreach my $checkLocation ($procMounts, $mtab ){
                        if ( -f $checkLocation ){
                                @ret=grep /\s+$sapmntToMount\s+/, `cat $checkLocation 2>/dev/null`;
                                chomp (@ret);
                                dbg "[7339]: Information: Checking for $sapmntToMount in $checkLocation returned (ret='@ret')\n";
                                last if (@ret);
                        }
                }
                if ( @ret == 0 ){
                        dbg "$sapmntToMount was detected in $FSTAB, but was not mounted.  It will be mounted on $me\n";
                        msg ('userstatus', "mount $sapmntToMount", $me );
                        system ("mount $sapmntToMount $REDIRECT");
                        $ret=($? >> 8 );
                        msg ('userstatus', "mount $sapmntToMount returned $ret", $me);
                }else{
                        dbg "$sapmntToMount was already mounted on $me\n";
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

#---------------------------------------------------------------------------
# NFSCheckBeforeCmd
# Description:  Check NFS mounts before calling command.
# Arguments:
#   Command to run
# Return Values: die if NFS not mounted, cmd return otherwise
sub NFSCheckBeforeCmd
{
        my $self = shift;
        my $cmd = shift;
        my $func = (caller(0))[3];

        dbg "=> Trace:Entering $func on $me\n";
        my $ret = $self->NFSCheck();
        if ($ret == 0) {
                error('nonfs', $me);
                dbg "<= Trace:Leaving $func on $me\n";
                erroranddie ('nonfs', $me);
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return (`$cmd`);
}

#
################################################################################
#######################--- End SAP File System Section ---######################
################################################################################

################################################################################
#########################--- Start SAP UI/GUI Section ---#######################
################################################################################
# General: A host of UI related functions


####
#  Description
#       - Get all SAP resource tags that are ISP on this system
####
#  Input:
#       - Sys - system to check
####
#  Output:
#       - List of Tags
####
#  Returns
#       - List of Tags
####
sub GetAllSAPTags
{
        my $sys=shift;
        my @final;
        my @existins = LK::ins_list($sys, '', $appType,$resType, '', '' );
        foreach my $ins (@existins){
                ($ins, my $state)=(split /\001/, $ins )[3,6];
                next if ( $state !~ /IS/ );
                push @final, $ins;
        }
        push @final, "none";
        print "@final\n";
        return @final;
}


####
#  Description
#       - Get all FS resource tags and mounts
####
#  Input:
#       - Sys - system to check
#       - hash ref
####
#  Output:
#       - None
####
#  Returns
#       - 1
####
sub GetAllFS
{
        my $sys=shift;
        my $ret=shift;
        my @existins = LK::ins_list($sys, '', 'gen','filesys', '', '' );
        foreach my $ins (@existins){
                (my $tag, my $id)=(split /\001/, $ins )[3,4];
                $$ret{$id}=$tag;
        }
        return 1;
}

# walk list of deps and related items for matches
sub depChecker
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $allDeps=shift;
        my $previous=shift;
        my $tryTag=shift;
        my $ref=shift;
        my $cnt=shift;
        my @children;
        if ( $$allDeps{$tryTag} ){
                @children=@{$$allDeps{$tryTag}};
        }
        if (@children == 0 ){
                dbg "Potential customization of resource hierarchy containing $tryTag.  No children found for $tryTag\n";
        }
        if ($cnt > $SAP_DEPENDENCY_DEPTH ){
                dbg "=> Trace:Leaving $func on $me\n";
                return 1;
        }
        $cnt++;
        dbg "$func: Entry/Re-Entry for $tryTag from $previous with count=$cnt\n";
        # a child of mine is also in the list as a child of $tag, remove
        # my child from the dep_list of $tag
        foreach my $child (@children){
                if ( $$ref{$child} ){
                        dbg "\n\nTrace:The dependent resource $child is already a part of this hierarchy as a dependent of $tryTag. $child will not be added to the \@deps array on $me.\n";
                        $$ref{$child}=0;
                }
                depChecker($allDeps,$tryTag, $child,$ref, $cnt);
        }
        dbg "=> Trace:Leaving $func on $me\n";
        return 1;
}


####
#  Description
#       - Display status information of the SAP Instance
#       - and sap processes
####
#  Input:
#       - Instance - SAP Instance
####
#  Output:
#       - Display Output
####
#  Returns
#       - 1 - always
####
sub SAPDisplay
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        #$action is added for consistency with other interfaces
        # that require an action parameter b/c they perform
        # multiple actions.  This $action value is not used by this function
        my $action=shift;
        my $RunLocal=shift;
        $RunLocal=($RunLocal eq '' ) ? 0: $RunLocal;
        my $Instance=$self->SAPINST();
        my $InstanceNum=$self->SAPINSTNUM();
        my $Type=$self->SAPINSTTYPE();
        my $Version=$self->GetEnqVersion();
        $self->IsInstanceRunning($nilValue,$RunLocal,1);

        if ((($Type == $TYPE_ERS) || (($Type == $TYPE_NEW_ERS) && ($Version == 1))) && ($RunLocal == 0)){
                my $name = (@{$self->{$KEYSAPCHECK}{$KEYSAPNAMES}})[0];
                print "Process Name\t=\t$name\n";
                print "Version\t=\t$Version\n";
                $self->SAPDisplay($action, 1);
                $self->SAPRemExec($func, $nilValue);
                dbg "=> Trace:Leaving $func on $me\n";
                return 1;
        }

        foreach my $name (@{$self->{$KEYSAPCHECK}{$KEYSAPNAMES}}){
                if (($Type == $TYPE_ERS) || ($Type == $TYPE_NEW_ERS)) {
                        if ($Version == 1) {
                                print "Server Name\t=\t$me\n";
                        } else {        # version 2
                                print "Process Name\t=\t$name\n";
                                print "Version\t=\t$Version\n";
                        }
                } else {
                        print "Process Name\t=\t$name\n";
                }
                print "Pid\t=\t$self->{$KEYSAPCHECK}{$name}{'pid'}\n";
                print "Text Status\t=\t$self->{$KEYSAPCHECK}{$name}{'textstatus'}\n";
                print "Display Status\t=\t$self->{$KEYSAPCHECK}{$name}{'dispstatus'}\n";
                if (($Type == $TYPE_ERS) || (($Type == $TYPE_NEW_ERS) && ($Version == 1))) {
                        my $corr_cs_inst_num = $self->CorrespondingEnqInstanceNum();
                        my $locks_held = $self->CheckHeldLocks($me, $corr_cs_inst_num);
                        my $lock_status = ($locks_held ? "Yes" : "No");
                        print "Locks Held\t=\t$lock_status\n";
                }
        }
        if ($Type == $TYPE_CS) {
                my %enqueue_status = (0 => "Active", 1 => "Disabled", 2 => "Out of Sync", 3 => "Unknown", 4 => "Unknown");
                my $status = $enqueue_status{$self->EnqRepStatus()};
                print "Enqueue Replication Status\t=\t$status\n";
        }
        dbg "=> Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Show text of available protection levels for this resource
####
#  Input:
#       - None
####
#  Output:
#       - Display protection levels
####
#  Returns
#       - 1 - always
####
sub ShowProtectionLevels
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        foreach my $key (keys %XREF_PROTECTION_LEVELS ){
                print "$XREF_PROTECTION_LEVELS{$key}\n";
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Show text of available recovery levels for this resource
####
#  Input:
#       - None
####
#  Output:
#       - Display recovery levels
####
#  Returns
#       - 1 - always
####
sub ShowRecoveryLevels
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        foreach my $key (keys %XREF_RECOVERY_LEVELS ){
                print "$XREF_RECOVERY_LEVELS{$key}\n";
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Get All the SAP Instances for a particular SID as found on the system
#       - under /usr/sap/<SID> directory.
####
#  Input:
#       - sid - The SAP SID to look for instances to protect
#       - [FilterOn] - Include or Exclude
#               '' | 0 - include all instances
#               1 - include Central Services only
#               2 - include Additional AS only (non-J* and non-DV*)
#               3 - ERS only
#               4 - include (J* and DV*)
####
#  Output:
#       - Instance Names
####
#  Returns
#       - 1 - always
####
sub GetAllSAPInstances
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $sid=shift;
        my $FilterOn=shift;
        my @SAPInstances;
        my $isCreatedOK=shift;
        $isCreatedOK = 0 if ($isCreatedOK eq '' );
        my @SAPReturn;

        # May 12 2011: Some of these choices could be collapsed
        if ( $FilterOn eq '' ){
                @SAPInstances=grep !/^SYS/, `ls -1 $SAPDIR/$sid 2>/dev/null`;
        }elsif ( $FilterOn == 1 ){
                @SAPInstances=grep /^SCS/ || /^ASCS/, `ls -1 $SAPDIR/$sid 2>/dev/null`;
        }elsif ( $FilterOn == 2 ){
                @SAPInstances=grep !/^SYS/ && !/^SCS/ && !/^ASCS/ && !/^$ABAPCI/ && !/^$JCI/ && !/^ERS/, `ls -1 $SAPDIR/$sid 2>/dev/null`;
        }elsif ( $FilterOn == 3  ){
                @SAPInstances=grep /^ERS/, `ls -1 $SAPDIR/$sid 2>/dev/null`;
        }elsif ( $FilterOn == 4 ){
                @SAPInstances=grep /^$ABAPCI/ || /^$JCI\d{2}/ && !/^SYS/, `ls -1 $SAPDIR/$sid 2>/dev/null`;
        }
        # Sort out the choices that are already under protection as not valid
        foreach my $val (@SAPInstances){
                chomp ($val);
                if ( isWhatProtected($me, $sap::appType, $sap::resType, $val, $sap::INSTINFOLOC) ne '' ){
                        next if ($isCreatedOK == 0 );
                        push @SAPReturn, "$val";;
                }
                #bug7664 - Left the logic to get the values as-is up above, but -1 $SAPDIR/$sidwe want to filter to <characters>+<digit><digit> like ASCS00
                #We want to filter things like lost+found or ASCS00.tgz or junk
                #We kept to digits, in the future SAP may go to 3 digit instances
                next if ($val !~ m/\w+\d$/);

                next if ($isCreatedOK);
                print "$val\n";
        }

        dbg "<= Trace:Leaving $func on $me\n";
        if ($isCreatedOK){
            return @SAPReturn;
        }else{
            return 1;
        }
}

####
#  Description
#       - Search the SAP base directory /usr/sap for a list of SIDs
#       - A valid SID includes the SYS directory and has an associated
#       - SAP user <sid>adm
#       - Used primarily by the UI
####
#  Output:
#       - SIDs
####
#  Returns
#       - 1 - always
####
sub GetAllSids
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $val;
        my @dirs=`ls -1 $SAPDIR 2>/dev/null`;
        foreach my $dir (@dirs){
                chomp ($dir);
                next if ( ! -d "$SAPDIR/$dir/SYS" );
                next if ( IsSAPDiagAgent($dir) );

                chomp ($val=length $dir);
                next if ($val != 3 );
                $val=lc("${dir}adm");
                system ("id $val >/dev/null 2>&1");
                next if ($? != 0 );
                print "$dir\n";
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Determine if the SID is the Diag instance and needs to be
#       - filtered out of the GetAllSids list
####
#  Output:
#       - None
####
#  Returns
#       - 0 - not an agent
#       - 1 - is an agent
sub IsSAPDiagAgent
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $val=shift;
        my @globdir=glob "$SAPDIR/$val/*/*Agent";
        my $ret=0;
        foreach my $dir (@globdir)
        {
                dbg "Trace: Agent directory from glob of $val is $dir on $me.\n";
                if ( -d "$dir" ){
                        $ret=1;
                        dbg "Trace: Returning ret=$ret for SID $val.  If ret equals 1, then this is an agent on $me\n";
                        last;
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

####
#  Description
#       - Determine if instance is a core instance or not
#       - a core instance is ABAP or Java SCS/CI
####
#  Input:
#       - Instance
####
#  Output:
#       - None
####
#       - 0 - is not a CI and does not require Full Protection
#               nor Standard Recovery
#       - 1 - is a SAP CI and requires Full Protection Level
#               and Standard Recovery Level
sub IsSAPCore
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";


        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $Instance=shift;
        dbg "Checking if the SAP Instance, $Instance, is contained in the key:@SAPCI.\n";
        $Instance=~s/(\S+)\d{2}/$1/g;

        if ( grep /$Instance/, @SAPCI){
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Test if this is a valid version of SAP on this server
####
#  Input:
#       - None
####
#  Output:
#       - Display recovery levels
####
#  Returns
#       Note: These returns are different b/c of the UI
#       request
#       - 1 - utilities not found, version update needed
#       -  0 - utilities found
sub IsSAPValidVersion
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $ret=0;
        if ( $SAPDIR eq '' || ! -d $SAPDIR ){
                        error ('nosuchdir', $SAPDIR, "/usr/sap", $me );
                $ret=1;
        }
        if ( ! -f $SAP_SVCS ){
                error ('nosuchfile', $SAP_SVCS, $me );
                $ret=1;
        }
        my $support_ref=1;
        foreach my $util ($SAP_SRVHOST_CMD, $SAP_OSCOL_CMD, $SAP_HOSTCTL_CMD ){
                if ( $util eq '' ){
                        error ('valueempty', "$func support_ref#=$support_ref", $func, $me );
                        $ret=1;
                }
                if ( ! -x $util ){
                        error ('exenotfound', $util, $me );
                        $ret=1;
                }
                $support_ref++;
        }
        if ($ret == 1 ){
                error ('validversion', $me );
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}
# SAP connector function to check the sys_list
# 0 - okay
# 4 - one or more nodes is down
sub InspectCluster
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $isalive;
        my $sysToCheck=shift;
        my @sys=();
        my $rc=0;
        if ($sysToCheck eq '' ){
                @sys=LK::sys_list('');
        }else{
                push @sys, $sysToCheck;
        }
        if ( @sys == 0 ){ $rc = 4 };
        foreach my $sys (@sys){
                 $isalive=LK::sys_getstate('', $sys );
                 if ($isalive != $LK::ALIVE ){
                        $rc=4;
                 }
        }
        dbg "=> Trace:Leaving $func on $me\n";
        return $rc;
}
#get LK tag from ID, ID is unique in cluster
sub getTagFromID
{
        #ID format=SAPID-<SID>-<INSTANCE>
        my $res=shift;
        if (!defined($res) || ($res eq '')) {
                dbg "ERROR: res was empty, unable to find corresponding tag in getTagFromID on $me";
                return '';
        }
        my $tag = (split /\001/, LK::ins_list('','','','','',$res))[3];
        return "$tag";
}

################################################################################
##########################--- End SAP UI/GUI Section ---########################
################################################################################

################################################################################
#########################--- Start SAP Signal Section ---#######################
################################################################################

####
#  Description
#       - Default signal handler
####
#  Input:
#       - signal
####
#  Output:
#       - None
####
#  Returns
#       - $?
####
sub sigHandler
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $sig=shift;
        my $isError=shift;
        my $ecode=($isError ne '' && $isError =~ /\d+/ ) ? $isError : 0 ;
        #my $ecode=$?;
        #$ecode=($ecode ne '' && $ecode =~ /\d+/ ) ? $ecode : 1 ;
        my $time = localtime;

        if ($sig eq 'ALRM' ){
                # May 12 2011: Keep the print just in case the hang prevents
                # the use of LKBase::*
                print STDERR "\tThe command $0 with pid $$ has hung.  Forcibly terminating on $me\n";
                msg ('hung', "$0", "$$", $0, $me );
        }
	# Nov 1 2018: remove locking on an abort
	my $hungCmd=$0; # saving it just because
	print STDERR "Determining the flags to remove for script $hungCmd for $signalTag on $me\n";
	if ($hungCmd =~ /quickCheck/i ){
		LK::flg_remove('', "$SAP_FLAG_QCK_IN_PROGRESS${signalTag}");
	}elsif ($hungCmd =~ /recover/i ){
		LK::flg_remove('', "$SAP_FLAG_RECOVER_IN_PROGRESS${signalTag}");
	}elsif ($hungCmd =~ /remove/i ){
		LK::flg_remove('', "$SAP_FLAG_REMOVE_IN_PROGRESS${signalTag}");
	}elsif ($hungCmd =~ /restore/i ){
		LK::flg_remove('', "$SAP_FLAG_RESTORE_IN_PROGRESS${signalTag}");
	}else{
		print STDERR "Unknown command $hungCmd with pid $$ has hung for $signalTag on $me.  Unable to determine the flags to remove\n";
	}

        # May 12 2011: Keep the print just in case the signal prevents
        # the use of LKBase::*
        my $msg=LKBase::String($resType, 'terminated', 'cause');
        $msg=sprintf("$msg", $0, $$, $sig, $me );
        msg ('terminated', "$0", "$$", $sig, $me );
        $?=$ecode;
        # Potential workaround for NFS hang is to exit 0 on an
        # alarm instead of using the stack_die method
        stack("$msg");
        exit 0;
        #stack_die("$msg\n");
}

#---------------------------------------------------------------------------
# SetAlarm()
# Description:  Call this function to set a timer that will
#   kill the process after a designated period of time.
#   Use it for any executing script that might get hung, preventing
#   LifeKeeper from functioning properly.
# Arguments:
#   Timeout - number of seconds before script process will be killed and
#     optionally, a colon delimited variable name from
#     /etc/default/LifeKeeper that can be used to override this value
#     in the format <secs>:<varname>.
#   Func (optional) - Name of a function to call when timeout occurs.
# Return Values:
#       1 - Always
#
sub SetAlarm
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        ref($_[0]) && $_[0] =~ /^\w+=HASH\(/ && shift;
        my $sig=shift;
        my $Timeout=shift;
        $Timeout=($Timeout ne '' || $Timeout < 1 ) ? $Timeout : 30;
        $SIG{$sig} = \&sap::sigHandler;
        dbg "\tSetting the $sig handler for cmd $0 with timeout=$Timeout and with process $$ on $me\n";
        if ($sig eq 'ALRM'){
                alarm $Timeout;
        }
        return 1;
}

################################################################################
#########################--- Start SAP Signal Section ---#######################
################################################################################


################################################################################
########################--- Start SAP Enqueue Section ---#######################
################################################################################

####
#  Description
#       - Call a sap.pm function on a remote LifeKeeper system
####
#  Input:
#   action - action to call
#       actionOpt - option to pass to the action
#       Sys - system to run on
####
#  Output:
#       - None
####
#  Returns
#       - 0 - The given command failed on at least one remote system
#       - 1 - The given command was successful on all remote systems
sub SAPRemExec
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        # action: the action to call on the remote system
        #       the action will be a function in the sap.pm file
        #       typically *Instance or HostExec
        #       $ret=$self->SAPRemExec($func, $nilValue, $nodisplay);
        my $action=shift;
        # actionOption: the options to add to the action call
        my $actionOption=shift;
        my $actionOption2=shift;
        my $Sys=shift;

        my $tag=$self->SAPTAG();
        my $Instance=$self->SAPINST();
        if ( $Instance =~ /none/i || $Instance eq '' ){
                ($Instance eq '' ) && msg ('valueempty', "Instance", $func, $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        $action=~s/sap:://g;
        if ( $action eq '' ){
                msg ('valueempty', "$func [action]", $func, $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        my @failedhosts;
        my $ret=0;
        my %eqvs;
        my %fulleqvs;
        $self->GetLKEquiv(\%eqvs, $tag, \%fulleqvs);
        my @sys=keys %eqvs;
        if ( @sys == 0 ){
                msg ( 'noequiv', $action, $Instance, $me );
                return 1;
        }
        my $isalive=0;
        my $remoteCmd;
        foreach my $otherSys (keys %eqvs){
                $isalive=LK::sys_getstate('', $otherSys );
                if ($isalive != $LK::ALIVE ){
                        error ('pairdown', $otherSys, $tag, $me);
                        next;
                }
                if ($Sys ne '' && $Sys ne $otherSys ){
                        dbg "$func called to execute on the specific host $Sys, skipping because $otherSys is not $Sys on $me\n";
                        next;
                }
                $remoteCmd="$APPBIN/remoteControl $eqvs{$otherSys} $Instance \"$action\" \"$actionOption\" \"$otherSys\" \"$tag\" \"$actionOption2\" ";
                msg ('userstatus', "$remoteCmd", $me);
                my $cmd = "$lkroot/bin/lcdremexec -e -d $otherSys -- \"$APPBIN/remoteControl $eqvs{$otherSys} $Instance \"$action\" \"$actionOption\" \"$me\" \"$tag\" \"$actionOption2\" 2>&2;\"";
                my @output = `$cmd`;
                $ret = ($? >> 8 );
                print @output;
                msg ('commandReturned', $remoteCmd, $ret, $me );
                if ( $ret == 0 ){
                        push @failedhosts, $otherSys;
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        (@failedhosts == 0 ) && return 1;
        return 0;
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, this
#       - routine will determine which version of the enqueue or
#       - enqueue replication server being used by checking the
#       - output of sapcontrol GetProcessList for specific process names.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error - called for unsupported instance type (3 or 4)
#       - 1 - version unable to be found (defaults to 1), or version 1 found
#       - 2 - version 2 found
####
sub GetEnqVersion
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $type = $self->SAPINSTTYPE();
        my $sid = $self->SAPSID();
        my $inst = $self->SAPINST();
        if (($type != $TYPE_CS) && ($type != $TYPE_ERS) && ($type != $TYPE_NEW_ERS)) {
                error('unsupportedinstancetype', $type, '1, 2, and 5');
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # The keys of %ver_hash are process names and the correponding values are the versions of
        # the enqueue server or enqueue replication server being used.
        my %ver_hash = (($type == $TYPE_CS) ? %SAP_ENQ_VER_HASH : %SAP_ENQREP_VER_HASH);
        my @process_names = keys %ver_hash;

        my $sapsrv_running = $self->SAPSRVUtils('status');
        if (!$sapsrv_running) {
                dbg "The sapstartsrv process for $self->SAPINST() was not running, attempting to start it.";
                if (!$self->StartSapServer($sid, $inst)) {
                        dbg "Unable to start the sapstartsrv process for $self->SAPINST() to check enqueue version via sapcontrol GetProcessList, falling back to profile check.";
                        my $ret = $self->GetEnqVersionFromProfile();
                        return $ret;
                }
        }

        my $instance_num = $self->SAPINSTNUM();
        my $sapadmin = $self->SAPADMIN();

        my $sap_process_cmd = "$SAP_CONTROL_CMD -nr $instance_num -function GetProcessList";
        my $resfile = $self->SAPRESULTFILE();
        my $tmpfile=$self->SAPTMPFILE();
        msg('userstatus', $sap_process_cmd, $me);
        $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $sap_process_cmd > $tmpfile 2>&2; echo \\\$? > $resfile' \" $REDIRECT");
        my @out = `cat $tmpfile`;
        chomp (my $ret = `cat $resfile`);
        msg('commandReturned', $sap_process_cmd, $ret, $me);
        unlink $resfile;
        unlink $tmpfile;

        # This loop checks whether any line of the GetProcessList output starts with
        # the process name for ENSA1/2 or ERSv1/2. Typical output looks like:
        # 0 name : enserver
        foreach my $line (@out) {
                next if ($line eq '');
                foreach my $name (@process_names) {
                        if ($line =~ /^\d+\s*name\s*:\s*${name}\s*$/i) {
                                dbg "$func found $name in GetProcessList output, returning $ver_hash{$name}.";
                                $self->StopSapServer($sid, $inst) if (!$sapsrv_running);  # Stop sapstartsrv if it wasn't running before
                                dbg "<= Trace:Leaving $func on $me\n";
                                return $ver_hash{$name};
                        }
                }
        }

        dbg "Unable to obtain enqueue version from GetProcessList output, assuming version 1.";
        $self->StopSapServer($sid, $inst) if (!$sapsrv_running); # Stop sapstartsrv if it wasn't running before
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, this
#       - routine will determine which version of the enqueue or
#       - enqueue replication server being used by checking the
#       - start profile for the instance for specific process names.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error - called for unsupported instance type (3 or 4)
#       - 1 - version unable to be found (defaults to 1), or version 1 found
#       - 2 - version 2 found
####
sub GetEnqVersionFromProfile
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $type = $self->SAPINSTTYPE();
        if (($type != $TYPE_CS) && ($type != $TYPE_ERS) && ($type != $TYPE_NEW_ERS)) {
                error('unsupportedinstancetype', $type, '1, 2, and 5');
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # The keys of %ver_hash are process names and the correponding values are the versions of
        # the enqueue server or enqueue replication server being used.
        my %ver_hash = (($type == $TYPE_CS) ? %SAP_ENQ_VER_HASH : %SAP_ENQREP_VER_HASH);
        my @process_names = keys %ver_hash;

        # Get the start profile for the instance from the sapstartsrv command in the sapservices file.
        my $srvpf = $self->SAPSRVPF();

        if(!open(PFL, "<$srvpf")) {
                error('enqversionprofileerror', $srvpf, $me, $tag);
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        # The corresponding lines in the profile that we're looking for should be similar to:
        # ENSA1: Execute_05 = local ln -s -f $(DIR_EXECUTABLE)/enserver$(FT_EXE) $(_EN)
        # ENSA2: Execute_05 = local ln -s -f $(DIR_EXECUTABLE)/enq_server$(FT_EXE) $(_ENQ)
        # ERSv1: Execute_03 = local ln -s -f $(DIR_EXECUTABLE)/enrepserver$(FT_EXE) #(_ER)
        # ERSv2: Execute_03 = local ln -s -f $(DIR_EXECUTABLE)/enq_replicator$(FT_EXE) $(_ENQR)

        my $line;
        while (<PFL>) {
                chomp($line = $_);
                next if ($line eq '');
                next if ($line =~ /^\s*#/); # Ignore commented out lines.
                foreach my $name (@process_names) {
                        if ($line =~ /$name/) {
                                dbg "$func found $name in profile $srvpf, returning $ver_hash{$name}.";
                                close(PFL);
                                dbg "<= Trace:Leaving $func on $me\n";
                                return $ver_hash{$name};
                        }
                }
        }
        close(PFL);

        dbg "$func unable to determine enqueue version for resource $tag, assuming version 1.";
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, find the local tag for the
#         corresponding TYPE_ERS, TYPE_NEW_ERS, or TYPE_CS resource with the same SID
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - The local tag for the corresponding resource, or undefined
#         if not found or called for an unsupported instance type
sub CorrespondingEnqResourceTag
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $sid = $self->SAPSID();
        my $type = $self->SAPINSTTYPE();

        if (($type != $TYPE_CS) && ($type != $TYPE_ERS) && ($type != $TYPE_NEW_ERS)) {
                error('unsupportedinstancetype', $type, '1, 2, and 5');
                dbg "<= Trace:Leaving $func on $me\n";
                return;
        }

        my @ins_list = LK::ins_list($me, '', '', '', '', '');
        # TYPE_CS (1) could correspond to TYPE_ERS (2) or TYPE_NEW_ERS (5)
        # TYPE_ERS (2) and TYPE_NEW_ERS (5) correspond to TYPE_CS (1)
        my @corr_type = (($type == 1) ? (2,5) : (1));
        my @found;
        my @split_ins;
        my $ins_sid;
        my $ins_type;
        foreach my $ins (@ins_list) {
                @split_ins = split /\001/, $ins;
                next if ($split_ins[2] !~ /^sap$/i); # Skip any non-SAP resources.
                ($ins_sid, $ins_type) = (split /\002/, $split_ins[5])[0,6];
                if (($ins_sid eq $sid) && (grep /^$ins_type$/, @corr_type)) {
                        push @found, $split_ins[3]; # Push the local tag for the corresponding resource.
                }
        }

        if (@found == 0) {
                dbg "No corresponding enqueue resource was found.\n";
                dbg "<= Trace:Leaving $func on $me\n";
                return;
        } elsif (@found > 1) {
                # If more than one corresponding resource is found, this means
                # that the user has more than one CS instance or more than one
                # ERS instance installed under the same SID.
                my $found_list = join ', ', @found;
                error('multcorrenqresources', $tag, $me, $found_list);
                dbg "<= Trace:Leaving $func on $me\n";
                return;
        }

        # If only one corresponding resource was found, return its local tag.
        dbg "<= Trace:Leaving $func on $me\n";
        return $found[0];
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, find the instance name of the
#         corresponding TYPE_ERS, TYPE_NEW_ERS, or TYPE_CS resource with the same SID
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - The instance number for the corresponding resource, or undefined
#         if not found or called for an unsupported instance type
sub CorrespondingEnqInstance
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $corr_enq_tag = $self->CorrespondingEnqResourceTag();
        if (!defined($corr_enq_tag)) {
                dbg "Could not find a corresponding enqueue resource for $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return;
        }
        my @ins = LK::ins_list($me, '', '', '', $corr_enq_tag, '');
        my $corr_inst = (split /\002/, (split /\001/, $ins[0])[5])[1];
        dbg "The enqueue instance for the resource corresponding to $tag is $corr_inst";

        dbg "<= Trace:Leaving $func on $me\n";
        return $corr_inst;
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, find the instance number of the
#         corresponding TYPE_ERS, TYPE_NEW_ERS, or TYPE_CS resource with the same SID
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - The instance number for the corresponding resource, or undefined
#         if not found or called for an unsupported instance type
sub CorrespondingEnqInstanceNum
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $corr_inst = $self->CorrespondingEnqInstance();
        if (!defined($corr_inst)) {
                dbg "Could not find a corresponding enqueue instance for $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return;
        }
        (my $corr_inst_num = $corr_inst) =~ s/\S+(\d{2})/$1/g;
        dbg "The enqueue instance number for the resource corresponding to $tag is $corr_inst_num";

        dbg "<= Trace:Leaving $func on $me\n";
        return $corr_inst_num;
}

####
#  Description
#       - Check the given system to see if backup locks for ENSA2/ERSv1
#         are being held in shared memory.  This can be checked by calling
#         'cleanipc <ASCS inst #> show' and checking for a line
#         matching "Shared Memory Key: 66".
####
#  Input:
#       - sys (optional) - name of the system to perform the check on
#       - inst_num (optional) - instance number of the ASCS instance to check for
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error or locks not being held in shared memory on sys
#       - 1 - locks being held in shared memory on sys
sub CheckHeldLocks
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $sys = shift;
        my $inst_num = shift;

        $sys = (($sys ne '') ? $sys : $me);
        $inst_num = (($inst_num ne '') ? $inst_num : $self->SAPINSTNUM());
        my $sapadmin = $self->SAPADMIN();

        my $check_lock_cmd = "$lkroot/bin/lcdremexec -e -d $sys -- \"su - $sapadmin -c \\\"sh -c \\\\\\\"LANG=C cleanipc $inst_num show\\\\\\\"\\\"\"";
        my @output = `$check_lock_cmd`;
        if (!@output) {
                msg('cmdoutputempty', $check_lock_cmd, $me);
        }
        my @results = grep { /Shared Memory Key\s*:\s*66/i } @output;
        dbg "<= Trace:Leaving $func on $me\n";
        return ((@results) ? 1 : 0);
}

####
#  Description
#       - This routine attempts to determine whether enqueue replication
#       - is enabled for a given CS instance via the appropriate parameter
#       - in the instance profile. Calling this for any instance type
#       - other than TYPE_CS will result in an error.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - Enqueue replication disabled
#       - 1 - Enqueue replication enabled
#       - 2 - Internal error, enqueue replication status unknown
#             (unsupported instance type, error opening profile, etc)
sub EnqRepEnabled
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $tag = $self->SAPTAG();

        my $type = $self->SAPINSTTYPE();
        if ($type != $TYPE_CS) {
                error('unsupportedinstancetype', $type, '1');
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        my $srvpf = $self->SAPSRVPF();
        if (! -f $srvpf) {
                error('enqrepenabledprofileerror', $srvpf, $me, $tag);
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        my $sapadmin = $self->SAPADMIN();
        my $enqversion = $self->SAPENQ_VERSION();

        if (($enqversion != 1) && ($enqversion != 2)) {
                dbg "Unsupported SAPENQ_VERSION ($enqversion) for resource $tag on $me, unable to obtain enqueue replication status.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        # repenabled_param is the profile parameter used to determine whether
        # the user has activated enqueue replication for the enqueue server.
        # ENSA1 - enque/server/replication
        # ENSA2 - enq/server/replication/enable
        my $repenabled_param = $SAP_ENQREPENABLED_PARAM_HASH{$enqversion};

        dbg "The enqueue replication parameter to be checked is $repenabled_param";

        if (!open(PFL, "<$srvpf")) {
                error('enqrepenabledprofileerror', $srvpf, $me, $tag);
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        my $repenabled = 0;
        my $line;
        my @split_line;
        while (<PFL>) {
                chomp($line = $_);
                next if ($line eq '');
                next if ($line =~ /^\s*#/); # Ignore commented out lines
                next if ($line !~ /=/);
                @split_line = split "=", $line;
                if (($split_line[0] =~ /$repenabled_param/i) && ($split_line[1] =~ /true/i)) {
                        dbg "Enqueue replication was found to be enabled in profile $srvpf on $me.";
                        $repenabled = 1;
                        last;
                }
        }
        close(PFL);

        if (!$repenabled) {
                dbg "Enqueue replication was not found to be enabled in profile $srvpf on $me.";
        }

        dbg "<= Trace:Leaving $func on $me\n";
        return $repenabled;
}

####
#  Description
#       - This routine calls the appropriate monitoring utility (ensmon
#       - or enq_admin) to determine the current replication status
#       - of the lock table for the enqueue server which is part of the
#       - given CS instance. Calling this for any other instance type
#       - other than TYPE_CS will result in an error.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - OK: Replication is configured and active
#       - 1 - OK: Replication is not configured
#       - 2 - WARNING: Enqueue server is running and replication is configured, but there
#                      is no active connection to a replication server or replication
#                      is out of sync
#       - 3 - ERROR: Cannot connect to the server to obtain replication status
#       - 4 - ERROR: Internal error (unsupported instance type, profile not found, etc)
sub EnqRepStatus
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $tag = $self->SAPTAG();

        my $type = $self->SAPINSTTYPE();
        if ($type != $TYPE_CS) {
                error('unsupportedinstancetype', $type, '1');
                dbg "<= Trace:Leaving $func on $me\n";
                return 4;
        }

        my $srvpf = $self->SAPSRVPF();
        if (! -f $srvpf) {
                dbg "Profile $srvpf not found on $me, unable to obtain enqueue replication status.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 4;
        }

        my $sapadmin = $self->SAPADMIN();
        my $enqversion = $self->SAPENQ_VERSION();
        my $resfile = $self->SAPRESULTFILE();

        if (($enqversion != 1) && ($enqversion !=2)) {
                dbg "Unsupported SAPENQ_VERSION ($enqversion) for resource $tag on $me, unable to obtain enqueue replication status.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 4;
        }

        my %monitorcmd_hash = (
                1 => "$SAP_ENSMON pf=$srvpf 2",
                2 => "$SAP_ENQ_ADMIN --replication_state pf=$srvpf"
        );

        my $monitorcmd = ${monitorcmd_hash}{$enqversion};

        # Query the replication status using the appropriate monitoring utility.
        $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $monitorcmd > /dev/null; echo \\\$? > $resfile' \" $REDIRECT");
        chomp (my $ret = `cat $resfile`);
        unlink $resfile;

        # At this point, ret holds the return code of the monitoring utility that was called.
        # Return codes for "ensmon pf=<CS profile> 2":
        #     0 - Replication either not configured, or configured and active
        #     4 - Enqueue server running and replication configured, but not connected to
        #         replication server
        #     8 - Cannot connect to the enqueue server (may be an issue with the network or
        #         with the enqueue server itself
        #     12- Error opening the given profile
        # Return codes for "enq_admin --replication-state pf=<CS profile>":
        #     0 - Replication either not configured, or configured and active
        #     3 - Enqueue replication server cannot be reached (may be an issue with the
        #         network or with the replication server itself)
        #     4 - Replication configured, but out of sync

        # In the following hash, the key is the value of SAPENQ_VERSION and the value is
        # a hash mapping return codes of the appropriate monitoring utility to the
        # appropriate return code of this routine.
        my %result_hash = (
                1 => { 0 => 0, 4 => 2, 8 => 3, 12 => 4 },
                2 => { 0 => 0, 3 => 3, 4 => 2 }
        );

        if (! grep /^${ret}$/, keys %{${result_hash}{$enqversion}}) {
                dbg "Unexpected return code ret=$ret from \'$monitorcmd\', unable to determine enqueue replication status.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 4;
        }

        # Since a return code of 0 could either indicate that enqueue replication is not
        # enabled or that it is enabled but not active, we attempt to differentiate between
        # these cases by attempting to determine whether enqueue replication is enabled in
        # the instance profile.
        if ($ret == 0) {
                my $repenabled = $self->EnqRepEnabled();
                # If repenabled=1, then replication was found to be enabled in the profile.
                if ($repenabled == 1) {
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 0;
                } else {
                        # Note: If there was an error in EnqRepEnabled and it is unknown
                        # whether enqueue replication is enabled, this return code will
                        # indicate that it is disabled. This may cause the replication
                        # status in the properties panel for the resource to incorrectly
                        # display 'Disabled' instead of 'Active' if it is actually enabled.
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }

        # Map the return code from the monitoring utility to the appropriate return code for this routine.
        my $enqrepstatus_ret = ${result_hash}{$enqversion}{$ret};
        dbg "The return code from \'$monitorcmd\' was $ret which was mapped to return code $enqrepstatus_ret in $func on $me.";
        dbg "<= Trace:Leaving $func on $me\n";
        return $enqrepstatus_ret;
}

####
#  Description
#       - For a TYPE_NEW_ERS resource, determine whether the resource
#       - should be moved to another backup node. To do this, we check:
#       - 1. The sap_no_ers_relocation_<tag> flag is not set for this resource,
#       - 2. RecoveryLevel is not 0 (OFF) for this resource,
#       - 3. The corresponding CS resource is ISP on this node,
#       - 4. A backup node appears to be available to attempt failover to, and
#       - 5. The lock table replication status is either not configured or in-sync.
####
#  Input:
#       - None
####
#  Output
#       - None
####
#  Returns
#       0 - conditions do not appear favorable to move the ERS hierarchy, or
#           object represents an unsupported instance type
#       1 - conditions appear favorable to move the ERS hierarchy
sub ERSShouldMove
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self = shift;
        my $tag = $self->SAPTAG();
        my $type = $self->SAPINSTTYPE();
        if ($type != $TYPE_NEW_ERS) {
                error('unsupportedinstancetype', $type, '5');
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my $norelocate_flag = "$SAP_FLAG_NO_ERS_RELOCATION${tag}";
        if (LK::flg_test('', $norelocate_flag) == 0) {
                dbg "The $norelocate_flag flag is set on $me. Not attempting to move ERS resource $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my $corr_cs_tag = $self->CorrespondingEnqResourceTag();

        if (!defined($corr_cs_tag)) {
                dbg "No corresponding CS resource was found. Not attempting to move ERS resource $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        if (!$self->IsTagISP($corr_cs_tag)) {
                dbg "Corresponding CS resource $corr_cs_tag is not ISP on $me. Not attempting to move ERS resource $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        dbg "ERS resource $tag and corresponding CS resource $corr_cs_tag are both currently ISP on $me. Checking whether conditions are favorable to move the $tag resource hierarchy to a backup node.";

        if (!$self->CanFailoverToBackupNode()) {
                msg('resourcecannotfailover', $tag);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my $cs_obj = sap->new("", "", $corr_cs_tag);

        if (!ref($cs_obj) || ($cs_obj->SAPLKSAPLK() eq '')) {
                error('ErrorGettingResourceInfo', $corr_cs_tag, $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my $repl_status = $cs_obj->EnqRepStatus();
        if (($repl_status != 0) && ($repl_status != 1)) {
                msg('enqueuereplicationoutofsync', $corr_cs_tag);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # If we have gotten here, then all conditions are favorable to relocate the
        # ERS hierarchy.
        msg('ersshouldmove', $tag, $me, $corr_cs_tag, $me, $tag);
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - For a TYPE_CS, TYPE_ERS, or TYPE_NEW_ERS resource, this routine
#       - checks that the versions of the enqueue server and corresponding
#       - enqueue replication server match. If this is not the case (for
#       - example, running ENSA1 and ER2) then the configuration is not
#       - supported by SAP and will lead to unexpected behavior. Calling
#       - this for any instance type other than TYPE_CS, TYPE_ERS, or
#       - TYPE_NEW_ERS will result in an error.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - enqueue server and enqueue replication server have different versions
#       - 1 - enqueue server and enqueue replication server have the same version
#       - 2 - error obtaining version information (unsupported instance type or no
#             corresponding enqueue resource found)
sub EnqVersionsMatch
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $tag = $self->SAPTAG();

        my $type = $self->SAPINSTTYPE();
        if (($type != $TYPE_CS) && ($type != $TYPE_ERS) && ($type != $TYPE_NEW_ERS)) {
                error('unsupportedinstancetype', $type, '1, 2, and 5');
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        # Find the tag for the corresponding enqueue resource, if one exists.
        my $corr_enq_tag = $self->CorrespondingEnqResourceTag();
        if (!defined($corr_enq_tag)) {
                dbg "No corresponding enqueue resource found for resource $tag with instance type $type.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        my $version_cmd = (($type == $TYPE_CS) ? "SAPENQ_VERSION" : "SAPENQREP_VERSION");
        my $corr_version_cmd = (($type == $TYPE_CS) ? "SAPENQREP_VERSION" : "SAPENQ_VERSION");

        # Get the version of this resource.
        my $version = $self->${version_cmd}();

        # Determine whether the corresponding enqueue resource is ISP on any node.
        # If so, we will get its version on the node where it is ISP in case the
        # underlying filesystem is unavailable on this node.
        my %corr_eqv;
        $self->GetLKEquiv(\%corr_eqv, $corr_enq_tag);
        my $isp_node;
        foreach my $sys (keys %corr_eqv) {
                if ($self->IsTagISP(${corr_eqv}{$sys}, $sys)) {
                        $isp_node = $sys;
                        last;
                }
        }

        # Abort the check if the corresponding resource is not ISP on any node.
        if (!defined($isp_node)) {
                dbg "The corresponding enqueue resource is not ISP on any node, aborting version check.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        # The equivalent tag and instance name are needed for the remoteControl call below.
        my $eqv_tag = ${corr_eqv}{$isp_node};
        my $corr_inst = $self->CorrespondingEnqInstance();
        if (!defined($corr_inst)) {
                dbg "Unable to determine the corresponding enqueue instance for resource $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 2;
        }

        # Check the version of the corresponding resource on the node where it is ISP in
        # case the underlying filesystem is unavailable on this node.
        system("$ENV{'LKROOT'}/bin/lcdremexec -e -d ${isp_node} -- \"$ENV{'LKROOT'}/lkadm/subsys/appsuite/sap/bin/remoteControl ${eqv_tag} ${corr_inst} \"${corr_version_cmd}\" \"getvalue\" \"$me\" \"${corr_enq_tag}\" 2>&2;\"");
        my $corr_version = ($? >> 8);
        my $versions_match = (($version == $corr_version) ? 1 : 0);

        dbg "<= Trace:Leaving $func on $me\n";
        return $versions_match;
}

####
#  Description
#       - For a TYPE_CS resource this routine checks whether the
#       - corresponding ERS resource is not ISP locally, but is
#       - ISP remotely and backup locks are being held on the
#       - remote system. This is applicable to ENSA1/ERSv1
#       - configurations since the lock table will be lost if
#       - the TYPE_CS resource is brought in-service on a system
#       - other than the one where ERSv1 is running.
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - corresponding ERS resource either can't be found, is
#           - not TYPE_NEW_ERS, or is ISP on this node
#       - 1 - corresponding ERS resource is not ISP on this node but
#           - is ISP on a different node and backup locks are stored there
sub ERSResourceIsRemotelyISP
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $tag = $self->SAPTAG();
        my $type = $self->SAPINSTTYPE();

        if ($type != $TYPE_CS) {
                error('unsupportedinstancetype', $type, '1');
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Return 0 if no corresponding ERS resource is found.
        my $corr_ers_tag = $self->CorrespondingEnqResourceTag();
        if (!defined($corr_ers_tag)) {
                dbg "No corresponding ERS resource was found for resource $tag.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Return 0 if the corresponding ERS resource is not TYPE_NEW_ERS.
        my $corr_ers_ins = LK::ins_list('', '', '', '', $corr_ers_tag, '');
        my $corr_ers_type = (split /\002/, (split /\001/, $corr_ers_ins)[5])[6];
        if ($corr_ers_type != $sap::TYPE_NEW_ERS) {
                dbg "Corresponding ERS resource $corr_ers_tag for resource $tag has type $corr_ers_type, not TYPE_NEW_ERS (5).";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Return 0 if the corresponding ERS resource is ISP locally.
        if ($self->IsTagISP($corr_ers_tag)) {
                dbg "Corresponding ERS resource $corr_ers_tag for resource $tag is ISP on $me.";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        # Return 1 if the ERS resource is ISP on any remote system and backup locks are held there.
        my %ers_eqv = map { split /\x01/; $_[2] => $_[3] } LK::eqv_list('', $me, $corr_ers_tag);
        my $corr_ers_state;
        my $locks_held;
        foreach my $sys (keys %ers_eqv) {
                $corr_ers_state = (split /\001/, LK::ins_list($sys, '', '', '', ${ers_eqv}{$sys}, ''))[6];
                if (($corr_ers_state =~ /ISP/) && ($self->CheckHeldLocks($sys))) {
                        dbg "Corresponding ERS resource $corr_ers_tag for resource $tag is not ISP on $me, but is ISP on $sys and backup locks are being held there.";
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }

        # If we have gotten here, then the corresponding ERS resource is either
        # not ISP on any node or was not holding locks where it was ISP, so we return 0.
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

####
#  Description
#       - Check whether this is an ERSv1 resource that the user
#       - is attempting to extend beyond two nodes.
####
#  Input:
#       - templateSys - system that the resource is being extended from
#       - templateTag - resource tag on the template system
####
#  Output:
#       - None
####
#  Returns
#       - 0 - this is an ERSv1 resource which has already been extended to two nodes
#       - 1 - this is either an ERSv2 resource or an ERSv1 resource which currently
#           - only exists on one node
sub ERSThirdNodeExtendCheck
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self = shift;
        my $templateSys = shift;
        my $templateTag = shift;

        my $type = $self->SAPINSTTYPE();
        if ($type != $TYPE_NEW_ERS) {
                error('unsupportedinstancetype', $type, '5');
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        # We get the ERS version from templateSys in case the ERS filesystem has not
        # yet been extended to and mounted on the target system.
        my $inst = $self->SAPINST();
        system("$ENV{'LKROOT'}/bin/lcdremexec -e -d $templateSys -- \"$ENV{'LKROOT'}/lkadm/subsys/appsuite/sap/bin/remoteControl $templateTag $inst \"GetEnqVersion\" \"$nilValue\" \"$me\" \"\" \"$nilValue\" 2>&2;\"");
        my $enqrepversion = ($? >> 8);
        dbg "The enqueue replication server version for resource $templateTag was found to be $enqrepversion on system $templateSys.";
        if ($enqrepversion == 1) {
               my @eqv = LK::eqv_list($templateSys, '', $templateTag);
               if (@eqv) {
                       error('threenodeersextendfail', $templateTag, $me);
                       dbg "<= Trace:Leaving $func on $me\n";
                       return 0;
               }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       Tests to see if a comm_up event is in progress with quorum installed
#       to avoid a resource transfer overlapping with quorum_isp.
#
####
#  Input:
#       - None
####
#  Output
#       - None
####
#  Returns
#       0 - comm_up is not running or quorum is not installed
#       1 - comm_up is running and quorum is installed

sub IsCommupInProgress
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $quorumverify = $lkroot . "/lkadm/bin/quorum_verify";
        my $commupflag = "!volatile!comm_up:";

        # If quorum is not installed no overlap issue exists.
        if (! -x $quorumverify) {
                dbg "Quorum is not configured. No special action is needed for ERS v2.\n" ;
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my @flg_list = LK::flg_list('');

        # Check for comm_up flag.  If found return 1.
        if (grep (/^$commupflag/, @flg_list)) {
                dbg "Comm up is in progress.  Special handling is needed for ERS v2.\n" ;
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }

        dbg "Comm up is not in progress. No special action is needed for ERS v2.\n" ;
        dbg "<= Trace:Leaving $func on $me\n";
        return 0;
}

################################################################################
########################--- End SAP Enqueue Section ---#########################
################################################################################

################################################################################
########################--- Start SAP CleanUp Section ---#######################
################################################################################

####
#  Description
#       - Cleanup hung pids of LK and possibly SAP commands issued by LK
####
#  Input:
#   parent - PID of the parent.
#   cmd - name of the parent script (optional).
####
#  Output:
#       - None
####
#  Returns
#       - 1 - always
sub CleanUp
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $parent=shift;
        my $cmd=shift;
        my $cfound=1;

        if ( $parent eq '' || $parent == 1 ){
                error ('usage', $func, "<parent> [cmd]" );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        my $child;
        my $ppid;
        if ( $cmd ne '' ){
                my @ps=grep /$cmd/, `ps -fwwp $parent 2>/dev/null`;
                if (@ps == 0 ){
                        dbg "No processes for $parent and $cmd to cleanup on $me\n";
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }
        my @gchildren;
        my $gchild;
        my $sig=15;
        my @children=grep /\s*\d+\s+$parent$/, `ps -o pid,ppid ax`;
        foreach $child (@children){
                $child=(split /\s+/, $child)[0];
                next if ($child eq '' );
                dbg "Found child=$child for parent=$parent on $me\n";
                @gchildren=grep /\s*\d+\s+$child$/, `ps -o pid,ppid ax 2>/dev/null`;
                foreach $gchild (@gchildren){
                        $gchild=(split /\s+/, $gchild)[0];
                        dbg "Found grandchild=$gchild for parent=$parent on $me\n";
                        if ($cfound){
                                $cfound=0;
                                msg ('usingkill', $child, "$func ($parent)" , $self->SAPSID(), $self->SAPINST(), "kill $sig, $child", $me );
                                system ("ps -p $child");
                                kill $sig, $child;
                        }
                        msg ('usingkill', $gchild, "$func ($parent)" , $self->SAPSID(), $self->SAPINST(), "kill $sig, $gchild", $me );
                        system ("ps -p $gchild");
                        kill $sig, $gchild;
                }
                $cfound=1;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

################################################################################
#########################--- End SAP CleanUp Section ---########################
################################################################################


################################################################################
########################--- Start SAP Profile Section ---#######################
################################################################################

#
####
#  Description
#       - Check any profile has Autostart = 1
####
#  Input:
#
####
#  Output:
#       -
####
#  Returns
#       0 - value is set
#       1 - value is not set; we want it be unset
sub AutoConfigCheck
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $ret=1;
        my $sid=$self->SAPSID();
        my $Instance=$self->SAPINST();
        my $sapmnt=$self->SAPMNT();
        my $vip=$self->SAPVIP();
        my $srvpf=$self->SAPSRVPF();

        # build list of possible profile files
        my @profiles;
        push @profiles, $srvpf;
        if ( $SAP_DEBUG ){
                push @profiles, "$sapmnt/$sid/profile/${sid}_${Instance}_$me";
                push @profiles, "$sapmnt/$sid/profile/${sid}_${Instance}_$vip";
                push @profiles, "$SAPDIR/$sid/$Instance/profile/${sid}_${Instance}_$me";
                push @profiles, "$SAPDIR/$sid/$Instance/profile/${sid}_${Instance}_$vip";
        }

        foreach my $pf (@profiles){
                next if ( ! -f $pf );
                dbg "Checking pf=$pf on $me\n";
                system ("grep -iq 'autostart[[:space:]]=[[:space:]]1' $pf");
                if ( $? == 0 ){
                        dbg "Issuing a Warning for the $sid/$Instance because Autostart is enabled in $pf\n";
                        msg ('autostart', $pf, $sid, $Instance, $me );
                        $ret=0;
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

####
#  Description
#       - Get default information about the instance(s) installed
#       - by using the DEFAULT.PFL file.  This function will fill
#       - the object hash with this information
#       - This function can also be called when some object is already
#       - known (such as the SAPVIP set by GetLK)
####
#  Input:
#       - [sid] - the SAP SID to check
####
#  Output:
#       - None - set SAPDBTYPE, SAPDBNAME, SAPVIP, and SAPDBHOST in hash
####
#  Returns
#       - 0 - error no sid specified
#       - 1 - success
####
sub GetSysDFLTInfo
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $sid=$self->SAPSID();
        if ($sid eq '' ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        my $sapadmin=$self->SAPADMIN();
        if ($sapadmin eq '' ){
                dbg "<= Trace:Leaving $func (sapadmin is null) on $me\n";
                return 0;
        }
        my $tmphost;
        my $sapmnt=$self->SAPMNT();
        if ($sapmnt eq '' ){
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my @dfl=`cat $sapmnt/$sid/profile/$DPFL 2>/dev/null`;
        my %vips;
        foreach my $line (@dfl){
                chomp ($line);
                if ($line =~ /\/dbtype/ ){
                        $line=~s/.*\/dbtype\s+=\s+(\S+)/$1/g;
                        $self->SAPDBTYPE($line);
                }elsif ($line =~ /\/mshost/ || $line=~ /j2ee\/scs\/host/ ){
                        $line =~ s/\S+\s+=\s+(\S+)/$1/g;
                        $vips{$line}=1;
                }elsif ($line =~ /\/dbname/ ){
                        $line=~s/.*\/dbname\s+=\s+(\S+)/$1/g;
                        $self->SAPDBNAME($line);
                }elsif ($line =~ /\/dbhost/ ){
                        $line=~s/.*\/dbhost\s+=\s+(\S+)/$1/g;
                        $self->SAPDBHOST($line);
                }elsif ($line =~ /SAPDBHOST/ ){
                        $line=~s/^SAPDBHOST\s+=\s+(\S+)/$1/g;
                        $tmphost=$line;
                }elsif ($line =~ /\/DIR_TRANS/ ){
                        $line=~s/.*\/DIR_TRANS\s+=\s+(\S+)/$1/g;
                        $self->SAPDIRTRANS($line);
                }
        }

        # If there is no dbhost setting, then use the
        # SAPDBHOST value
        if ( $self->SAPDBHOST() eq '' ){
                $self->SAPDBHOST($tmphost);
        }

        # If the dbname is not set in file, default to the
        # SID name
        if ( $self->SAPDBNAME() eq '' ){
                $self->SAPDBNAME($sid);
        }

        if ($self->SAPVIP() eq '' ){
                # Use the sap control command to get the actual host name for the instance
                my $InstanceNum=$self->SAPINSTNUM();
                my $resfile=$self->SAPRESULTFILE();
                my $sapHostListCmd="$SAP_CONTROL_CMD -nr $InstanceNum -format list -function GetInstanceProperties | grep ^$SAPVHOSTSTRING";
                chomp (my $sapHostListOut=$self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $sapHostListCmd 2>/dev/null; echo \\\$? > $resfile'\""));
                chomp (my $ret=`cat $resfile`);
                msg ('commandReturned', $sapHostListCmd, $ret, $me);
                dbg "The return code of the $sapHostListCmd was ret=$ret with output=$sapHostListOut on $me\n";
                unlink $resfile;
                my $sapVHostName=(split /,/,$sapHostListOut)[2];
                $sapVHostName=~s/\s+//,;
                dbg "The $sapHostListCmd resulted in a VIP of $sapVHostName on $me\n";
                # if non-null use as the SAPVIP setting, otherwise default to previous behavior
                if ($sapVHostName ne '' ){
                        # If someone has set this VIP post create, or it was set while
                        # we were inside the parent if, do not override the VIP that has
                        # been set by GetLK, but we do need to dbg msg it
                        #
                        if ($self->SAPVIP() eq '' ){
                                $self->SAPVIP("$sapVHostName");
                        }else{
                                dbg "The VIP was set while we were in the check loop to (".$self->SAPVIP().").\nWe were going to set it to ($sapVHostName).\n";
                        }
                }else{
                        # If we are called, post create, do not override the VIP set by GetLK
                        my @vips=keys %vips;    # this should be a single host
                        if (@vips && $self->SAPVIP() eq '' ){
                                $self->SAPVIP("@vips");
                        }
                        #To Think About/To Negatively Test: What if @vips=0!
                }
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

################################################################################
##############################--- Below the line ---############################
################################################################################
# Wed Jun  1 10:03:50 EDT 2011
# Below the line features related to additional resource information objects.
# These items pertain to flexible recover and additional
# recovery information stored in file SAP_INFO_FILE_DEFAULTS

####
#  Description
#       - Write extra tunable values to the resource specific Info File
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error
#       - 1 - success
####
sub SAPWRITER
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $tag=$self->SAPTAG();
        my $type=$self->SAPINSTTYPE();

        my $file=$self->SAPINFOSTORE();
        (my $filename=$file) =~ s@.*/@@;
        if ( ! -f $file ){
                dbg "The SAP_INFO_FILE $file does not exist, it is possible that this is an initial write action on $me\n";
        }

        my $ret=0;
        dbg "Preparing to open the file $filename on $me\n";

        $ret=1 if (open (SAP, ">$file" ) );
        # if we cannot open the file for the initial write this means
        # the resource will always use the defaults for additional
        # actions defined by the SAP_INFO_FILE
        if ( ! $ret ){
                dbg "Unable to write the $file on $me\n";
                msg ('cantwrite', $file, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        my $val;
        my @vals=keys %SAP_INFO_FILE_DEFAULTS;
        foreach my $internalCmd (@vals){
                $val=$self->$internalCmd();     # the file format is function name specific
                dbg "The current internal function $internalCmd returned the value $val on $me\n";
                if ($val eq '' ){
                        $val=$SAP_INFO_FILE_DEFAULTS{$internalCmd};
                        dbg "No initial value was defined for $internalCmd.  Setting the initial value for $internalCmd to val=$val.  This is the default value on $me.\n";
                }
                dbg "Storing the value $val for function $internalCmd in the $filename on $me\n";
                print SAP "$internalCmd=$val\n";
        }

        if (($type == $TYPE_CS) || ($type == $TYPE_ERS) || ($type == $TYPE_NEW_ERS)) {
                my $versionCmd = (($type == $TYPE_CS) ? "SAPENQ_VERSION" : "SAPENQREP_VERSION");
                $val = $self->$versionCmd();
                dbg "Storing the value $val for $versionCmd in the file $filename on $me\n";
                print SAP "$versionCmd=$val\n";
        }

        close (SAP);
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - Read extra tunable values from the resource specific
#       - Info File and set in object
####
#  Input:
#       - None
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error
#       - 1 - success
####
sub SAPREADER
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $file=$self->SAPINFOSTORE();
        (my $filename=$file) =~ s@.*/@@;
        if ( ! -f $file ){
                dbg "The SAP_INFO_FILE $file does not exist on $me.  Unable to read the values on $me\n";
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }
        my $ret=0;
        dbg "Preparing to open the file $filename on $me\n";

        $ret=1 if (open (SAP, "<$file" ) );
        if ( ! $ret ){
                dbg "Unable to read or open the $filename on $me\n";
                msg ('nosuchfile', $file, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        while (<SAP>){
                chomp ($_);
                my $val=$_;
                next if ($val =~ /^#/ );
                next if ($val eq '' );
                next if ($val !~ /=/ );
                (my $internalCmd,$val)=(split /=/, $val);
                chomp ($internalCmd);
                chomp ($val);
                dbg "Found internalCmd=$internalCmd in $filename, value=$val on $me\n";
                if ($val eq '' ){
                        if (($internalCmd eq 'SAPENQ_VERSION') || ($internalCmd eq 'SAPENQREP_VERSION')) {
                                $val = $self->GetEnqVersion();
                        } else {
                                $val=$SAP_INFO_FILE_DEFAULTS{$internalCmd};
                        }
                        dbg "SET OBJECT=$internalCmd with default value=$val on $me\n";
                        dbg "$func did not find a suitable value in the $filename for $internalCmd, using default val=$val on $me\n";
                }
                $self->$internalCmd($val);
        }
        close (SAP);
        dbg "<= Trace:Leaving $func on $me\n";
        return 1;
}

####
#  Description
#       - At end of a successful action rewrite the
#       - Info File using object values and calculated
#       - values for CURRENTTIME and etc.
####
#  Input:
#       - Type
#               - 0 - restore/remove action
#                       zero out certain values
#               - 1 - non-restore action
#                       do not zero out certain values
####
#  Output:
#       - None
####
#  Returns
#       - 0 - error
#       - 1 - success
####
sub SAPEND
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $type=shift;
        if ($type eq '' || $type == 0 ){
                foreach my $internalCmd (@SAP_INFO_FILE_RESET_LIST){
                        dbg "$func is clearing the value of $internalCmd in the SAP_INFO_FILE on $me";
                        $self->$internalCmd(0);
                }
                my $v = $self->GetSysCurrentTime();
                dbg "$func is resetting the current time value in the SAP_INFO_FILE on $me";
                $self->SAPRECOVER_CURRENTTIME($v);
                $self->SAPRESTORE_DATE(LKBase::GetLocalizedDate);
        }
        my $ret=$self->SAPWRITER();
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

####
#  Description
#       - determine if this is an expertMode create
#       - determines if vAppKeeper type then default to
#       - expertMode
####
#  Input:
#       - Type
#               command line indicator of the mode
####
#  Output:
#       - None
####
#  Returns
#       - 0 - non-expert mode
#       - 1 - expert mode
####
sub LKSAPEXPERTMODE
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $type=shift;
        my $ret=0;
        if ( $type ne '' && $type == $internalMode ){
                dbg "\tThe expertMode has been enabled via the command line on $me.\n";
                $ret=1;
        }elsif ( $SAP_EXPERTMODE == 1 ){
                dbg "\tThe expertMode has been enabled via the defaults file on $me.\n";
                $ret=1;
        }elsif ( $SAP_VMNODEMODE == 1 ){
                dbg "\tThe expertMode has been enabled for virtualized environments on $me.\n";
                $ret=1;
        }
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

################################################################################
#############################  Connector Functions  ############################
################################################################################

####
#  Description
#       - write a message string to the specified file
####
#  Input:
#       - file - fully qualified file name
#       - existOK - [0|1] error if file
#       - msg - message to print to the file
#
####
#  Returns
#
# NOTE: These return codes match command line style returns
# 1 - if file write or open fails
# 0 - if success
####
sub FileHandler
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $file=shift;
        my $existOK=shift;
        my $msg=shift;
        my $self=shift;
        my $ret=0;
        (my $filename=$file) =~ s@.*/@@;
        if ( -f $file ){
                dbg "The $file exist on $me\n";
                if ($existOK == 0){
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 1;
                }
        }
        dbg "Preparing to open the file $filename on $me\n";
        $ret=1 if (open (SAP, ">$file" ) );
        if ( ! $ret ){
                dbg "Unable to write the $file on $me\n";
                msg ('cantwrite', $file, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1;
        }
        print SAP "$msg\n";
        dbg "Preparing to close the file $filename on $me\n";
        dbg "<= Trace:Leaving $func on $me\n";
        close (SAP);
        #chown the file that was created
        if ($self){
           my $sapadmin=$self->SAPADMIN();
           my (undef, undef, $uid, $gid) = getpwnam($sapadmin) or return 1;
           chown $uid, $gid, $file;
        }
        return 0;
}

####
#  Description
#       - init function
#       - If the HA connector is configured for an SAP instance, then this routine
#       - is called by sapstartsrv when it starts up (via the SAP SIOS HA Connector)
#       - to determine whether the HA Connector should be activated or not.  It will
#       - only be activated if init returns 0.
####
#  Input
#       - None
#  Returns
#       - 0 - Necessary commands are available and cluster software is running locally
#       - 1 - Error during detection of cluster state
#       - 2 - Bad parameters
#       - 3 - Necessary commands not installed correctly
#       - 4 - Cluster commands installed correctly but cluster software is not running
####
sub init() {
        my $ret = IsSAPValidVersion();
        if ($ret) { return 3; }
        my $notAlive = InspectCluster($me);
        my $lkInitDone = (-f "$ENV{'LKROOT'}/config/LK_INITDONE");
        if ($notAlive || !$lkInitDone) { return 4; }
        return 0;
}
 
####
#  Description
#       - dbgConnector function
####
sub dbgConnector($)
{
        return 0 unless ($SAP_CONNECTOR_DEBUG);
        do_dbg(@_);
}

####
#  Description
#       - getResFromSINO function
####
#  get resources from the sid/ino combo
sub getResFromSINO($$)
{
        my $sid=shift;
        my $ino=shift;
        (my @lsr)=GetAllSAPInstances($sid,'',1);
        my @matches;
        foreach my $inst (@lsr){
                (my $instNum=$inst)=~s/\S+(\d{2})/$1/g;
                push @matches, $inst if ($instNum==$ino );
        }
        return @matches;
}

####
#  Description
#       - getMsgFromID function
####
#  Get the message from the ID supplied
sub getMsgFromID {
        my %eqvs;
        my %fulleqv;
        my @sys;
        my $curr="NotActive";
        my %pri;
        my $sapObj;
        my $func=shift;
        my $res=shift;
        my $tag = getTagFromID($res);
        $sapObj=sap->new("","",$tag,'');
        #If we cant resolve the tag, then return internal error 1
        if ( !ref($sapObj) || $sapObj->SAPLKSAPLK() eq '' ){
                exit 1;
        }
        $sapObj->GetLKEquiv(\%eqvs, $tag,\%fulleqv);
        foreach my $sys (keys %fulleqv){
                $pri{$fulleqv{$sys}{'Pri'}}=$sys;
        }

        # Create an array of cluster nodes that the resource has been extended
        # to, ordered by priority
        foreach (sort { $a <=> $b } keys %pri) {
                push @sys, $pri{$_};
        }
        # handle not extended
        if (@sys == 0) {
                push @sys,$sap::me;
        }
        foreach my $s (@sys){
                if ($sapObj->IsTagISP($tag,$s)){
                        $curr=$s;
                        last;
                }
        }
        # list of nodes in failover priority order
        my $prinodes=join ',', @sys;
        #TBD: What do we want the CLUSTER GROUP NAME TO BE?
        # option 1: my $grp=join '_', @sys;
        # option 2:
        #my $grp=$sapObj->SAPLKSAPID();
        #lsr will append more info to the beginning of this string
        if ($func eq 'lsr') {
            return "$res:$res:$prinodes:";
        }else{
            return "$res:$res:$curr:$prinodes:";
        }
}

####
#  Description
#       - Connector lsn ("List SAP Nodes") function
#       - Given the cluster ID of an SAP resource, this
#       - routine returns the name of the node where the
#       - instance is currently running (or "NotActive")
#       - as well as a prioritized list of possible
#       - failover targets for the instance. This is
#       - written to the output file specified by --out
#       - in the format:
#       - RESID:GRPID:CURR_NODE:PRIO_NODES
#  Input
#       - out - Filename to write results to
#       - res - Cluster resource name
#  Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
####
sub lsn($$) {
        my ($out, $res) = @_;
        if (-e $out) { unlink $out; }
        my $msg=getMsgFromID("lsn",$res);
        if ($msg eq '') { exit 1; }
        dbgConnector "$msg";
        #if the file exists, we'll delete and it will override it
        my $tag = getTagFromID($res);
        my $sapObj=sap->new("","",$tag,'');
        return FileHandler($out, 1, "$msg", $sapObj);
}

####
#  Description
#       - Connector lsr ("List SAP Resources") function
#       - Given an SID and instance number, this routine
#       - determines the cluster resource ID corresponding
#       - to the instance and the cluster nodes that it
#       - could failover to (in priority order). This is
#       - written to the file specified by --out in the format:
#       - SID:INO:RESID:GRPID:CLUSTER_NODES
#  Input
#       - out - Filename to write results to
#       - sid - SAP system ID
#       - ino - Instance number
#  Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
#       - 3 - Clustered instance found, but cluster control disabled
####
sub lsr($$$) {
        my ($out, $sid, $ino) = @_;
        #if the file exists, we'll delete and it will override it
        if (-e $out) { unlink $out; }
        my @matches=getResFromSINO($sid,$ino);
        my $returnRC=0;
        foreach my $res (@matches){
                my $sapObj=sap->new("$sid","$res",'','');
                my $id=$sapObj->SAPLKSAPID();
                my $msg=getMsgFromID("lsr",$id);
                if ($msg eq '') { exit 1; }
                dbgConnector "$sid:$ino:$msg";
                my $rc=FileHandler($out, 1, "$sid:$ino:$msg", $sapObj);
                if ($rc == 1) { $returnRC = 1; }
        }
        return $returnRC;
}

####
#  Description
#       - Connector cpa ("Check Pending Action") function
#       - This routine is called to determine whether an
#       - action (start/stop/migrate) is already in progress.
#  Input
#       - res - Cluster resource ID
#       - act - Action to check for pending status
#  Returns
#       - 0 - Found pending action of specified type
#       - 1 - No pending action of specified type found
#       - 2 - Bad parameters
####
sub cpa($$) {
        my ($res, $act) = @_;
        my $filename = "$sap::APPSTORE/$res.$act";

        # bug7669 - Checking flags in progress and returning 0 to let cpa call to know that
        # a restore or remove are in progress. Note: Couldnt use FlagInProgress function,
        # because we dont have the sapObj container and not enough info to build one.
        my $tag = getTagFromID($res);
        if ($tag eq '') { return 1;}

        my $remove_flag = "$SAP_FLAG_REMOVE_IN_PROGRESS${tag}";
        my $restore_flag = "$SAP_FLAG_RESTORE_IN_PROGRESS${tag}";
        if (($act =~ /stop/i) && (LK::flg_test('',$remove_flag) == 0)) {
                msg ('inprogress', $res, 'remove', $sap::me);
                return 0;
        }
        if (($act =~ /start/i) && (LK::flg_test('',$restore_flag) == 0)) {
                msg ('inprogress', $res, 'restore', $sap::me);
                return 0;
        }

        #In case the file is empty, e.g. file was created, but the process died
        if ( -e $filename && -z $filename ){
            dbgConnector "Filename=$filename was empty";
            unlink $filename;
            next;
        }
        return 1 if (!open(SAPCPAWORKFILE, $filename));
        #Check to see if pid is valid
        my $pid=<SAPCPAWORKFILE>;
        close(SAPCPAWORKFILE);
        chomp($pid);
        if (($pid eq '') || ($pid == 1))
        {
             dbgConnector "Filename=$filename was removed since pid is invalid";
             system "rm $filename";
             return 1;
        }
        my $cmd="ps -wwwfp $pid |";
        open (PSINFO, $cmd);
        my @pidoutput = <PSINFO>;
        close (PSINFO);
        if (($pidoutput[1]) && ($pidoutput[1] =~ m/sap_sios_cluster_connector/)) {
             dbgConnector "Filename=$filename and there is a pid $pidoutput[1] running as us";
             return 0;
        }else{
             dbgConnector "Filename=$filename and this is not our fra, so we delete it\n";
             #This is not our fra, delete the file
             system "rm $filename";
        }
        return 1;
}

####
#  Description
#       - Connector smm ("SAP Maintenance Mode") function
#       - Called to enable, disable, or check status of HA
#       - maintenance mode. Status is written to the file
#       - specified by --out.
#  Input
#       - out - Output file
#       - sid - SAP system ID
#       - ino - Instance number
#       - mod - -1 = check, 0 = disable, 1 = enable
#  Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
#       - 3 - No corresponding resource found in the cluster
####
sub smm($$$$) {
        my ($sid, $ino, $mod, $out) = @_;
        dbgConnector "Maintenance mode $mod is getting called for SID=$sid, Instance=$ino";
        # if the file exists, we'll delete and it will override it
        if (-e $out) { unlink $out; }
        my @matches=getResFromSINO($sid,$ino);
        my $returnRC=0;
        # There really should be only 1 match to SID/INO
        foreach my $res (@matches){
                my $sapObj=sap->new("$sid","$res",'','');
                if ( ref($sapObj) && $sapObj->SAPLKSAPLK() ne '') {
                        my $tag=$sapObj->SAPTAG();
                        my $rc=0;
                        system "$ENV{'LKROOT'}/lkadm/subsys/appsuite/sap/bin/lk_maintenance_mode --tag=$tag --mode=$mod --cluster";
                        $rc = $?;
                        my $msg="Executed command to put resource $tag, SAP SID=$sid, instance=$ino with a return code $rc";
                        #If return code is not 0, then set returnRC to be not OK
                        if ($rc) {
                           $returnRC=-1;
                        }
                        #Write to out file, exist ok
                        FileHandler($out, 1, "$msg", $sapObj);
                }
        }
        # SAP Certification requires a rc=0 from check all the time, failed or not.
        if ($mod =~/^check$/i) {
            return 0;
        }
        return $returnRC;
}

####
#  Description
#       - Connector fra ("Fire Resource Action") function
#  Input
#       - res - Cluster resource ID
#       - act - Action (start/stop/migrate)
#       - nod - (Optional) Target node for migration
#  Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
####
sub fra($$$) {
        my ($res, $act, $nod) = @_;
        my $tag = getTagFromID($res);
        if ($tag eq '') { return 1;}

        # If the user specified a target migration node, the migrate action will
        # fail if the given node is not alive.
        if (defined($nod) && ($nod ne '') && (LK::sys_getstate('', $nod) != $LK::ALIVE)) {
                dbgConnector "The given target migration node $nod is not alive, returning 1 in fra on $me";
                error ('pairdown', $nod, $tag, $me);
                return 1;
        }

        my $actSaved=$act;
        my $waitOnPid=0;
        if ($actSaved =~ /start_internal/i ){
                $act="start";
                $waitOnPid=1;
                dbgConnector "The fra was called with $actSaved which was mapped to $act and triggered the setting of the waitOnPid flag on $me";
        }
        my $filename = "$sap::APPSTORE/$res.$act";

        # If files exist with pattern of $res.* it means there is a command on that res already
        # If we are being called again, it is possible that this call is a duplicate call
        my $safetyCheck=sap::checkFileandPid($res,$act);
        return 1 if ($safetyCheck);
        dbgConnector "The flag files under $sap::APPSTORE/$res.* do not exist on $me";

        # Now we need a way to check for the RKU/RKS without causing a massive process
        # cleanup at the end via DESTROY. We also have to make sure that we do not
        # block the entire fra process of the parent

        my $conPid=fork();
        # if we were unable to fork a child, we assume no RKS/RKU and continue
        # otherwise, in the child process run the checks
        if ($conPid == 0 ){     # Start of Child_1
                dbgConnector "[Child_1] Trying to create a sapObj from $tag for the fra $act and the RKU checks on $me";
                my $sapObj=sap->new("","","$tag");
                my $rc=0;
                if ( ref($sapObj) && $sapObj->SAPLKSAPLK() ne '' ){
                        dbgConnector "[Child_1] Created sapObj from $tag for fra on $me";
                        if ( $sapObj->isSAPRKSRunning() ) {
                                $rc=1;
                                dbgConnector "[Child_1] An RKU/RKS is in-progress.  Skipping the fra $act on $me";
                        }
                }
                exit $rc; #End of Child_1
        }

        # Parent
        # The above section could become a function, in any regard, we need to wait for the RKU
        # checks to complete
        dbgConnector "[Parent] Waiting for $conPid to complete the RKU checks on $me";
        waitpid($conPid,0);
        my $rkuRet=$? >> 8; #According to the perldocs the return of the forked child command is in $?
        if ($rkuRet ){
                dbgConnector "[Parent] A confirmed RKU/RKS is in-progress (child returned:$rkuRet) on $me";
                return 1;
        }
        # Second child process is forked so that we can do the actually called action for the sscc
        # In the first child we needed to check for the RKU/RKS.  We could not use the same process
        # because the parent thread cannot wait for the completion of the sscc action.  It can
        # and should wait on the completion of the RKU/RKS checks.

        my $pid = fork();
        #Error
        if (not defined $pid) {
                dbgConnector "[Parent] Unable to fork a new child process.  The pid ($pid) is not defined, so we must return 1 on $me";
                return 1;
        }
        elsif ($pid == 0) { # Start of Child_2

                #I thought I should put the output to dev/null, I wasnt sure, because it shows lots of output
                #If debug is set, then we will display all of the output, for debug purposes
                if(!$ENV{'SAP_DEBUG'}) {
                        open STDOUT, '>>/dev/null';
                        open STDERR, '>>/dev/null';
                }

                my $actRet=0;
                dbgConnector "[Child_2] BEGIN: Action $act for tag $tag started on $me";
                if ($act =~/start/i)
                {
                        # 12.15.15 - SAP certification revision 2 is failing w/ restore -b
                        # changing to restore
                        system "$ENV{'LKROOT'}/bin/perform_action -t $tag -a restore";
                        $actRet=($? >> 8);
                }
                elsif ($act =~/stop/i)
                {
                        system "$ENV{'LKROOT'}/bin/perform_action -t $tag -a remove";
                        $actRet=($? >> 8);
                }
                elsif ($act =~/migrate/i )
                {

                        # summary of migrate and version 2.x:
                        # It seems that v2 of an fra migrate wants to perform a SAP instance only
                        # failover.  The instance should not be stopped (i.e. removed) but
                        # forcefully ended and then restored on the other node (*/!^%?*)
                        # Below is the way to do a failover (non-remove remove) that also lets the tool know that
                        # we are in progress and to not issue another stop or migrate.
                        # We create a .stop file first
                        my $tmpfilename = "$sap::APPSTORE/$res.stop";
                        system ("echo $$ > $tmpfilename");
                        dbgConnector "[Child_2] Trying to create a sapObj from $tag on $me";
                        my $sapObj=sap->new("","","$tag");
                        my $rc=0;
                        if ( ! ref($sapObj) || $sapObj->SAPLKSAPLK() eq '' ){
                                dbgConnector "[Child_2] Failed to create a sapObj from $tag for fra on $me";
                                exit 1;
                        }

                        # A new requirement for version 2.x, the --node can be empty.  If it is empty
                        # the connector must pick the next node for the migrate action
                        if ($nod eq '' ){
                                dbgConnector "[Child_2] No node parameter was specified for the migrate action, calling fraNextNode on $me";
                                $nod = ($sapObj->fraNextNode())[0];
                                dbgConnector "[Child2_2 No node parameter was passed in, the fra $act has chosen the next best node as $nod on $me";
                        }

                        my $savedResState = (split /\x01/, LK::ins_list('', '', '', '', $tag, ''))[6];

                        # We set the resource state to OSU so that the remove script won't run (which would restart sapstartsrv)
                        # and also so that quickStart won't run while we're killing the processes.
                        LK::ins_setstate('', $tag, $LK::OSU);

                        # Pause: This is cause for alarm.  According to the specification, a migrate
                        # actually involves a failover of the SAP resource w/o the actual stopping of the
                        # resource on the in-service node. This is to preserve lock tables.
                        # 1) We turn off recovery and remove (see below w/ protection level)
                        # 2) We kill the associated sapstart process, running instance processes, and sapstartsrv
                        my $savedProtectionLevel=$sapObj->SAPPROTECTION();
                        dbgConnector "[Child_2] Obtained savedProtection=$savedProtectionLevel on $me";
                        my $newLevelP="$P_LEVEL_0";
                        $sapObj->SAPPROTECTION($newLevelP);
                        $sapObj->SAPUpdateLKInfoField();
                        my $currPLevel=$sapObj->SAPPROTECTION();
                        dbgConnector "[Child_2] Updated protection levels; currentLevel=$currPLevel savedProtection=$savedProtectionLevel on $me";

                        dbgConnector "[Child_2] Calling KillInstanceProcesses to kill running processes associated to instance on $me";
                        my $kipRet = $sapObj->KillInstanceProcesses();
                        if ($kipRet == 0) {
                                $sapObj->SAPPROTECTION($savedProtectionLevel);
                                $sapObj->SAPUpdateLKInfoField();
                                LK::ins_setstate('', $tag, $savedResState);
                                dbgConnector "[Child_2] At least one required process for the instance was not killed by KillInstanceProcesses, restoring protection levels: currentLevel=$currPLevel savedProtection=$savedProtectionLevel on $me";
                                exit 1;
                        }

                        # 3) After the kill we use lcdremexec to start on the other node, skipping the actual SAP remove
                        # lcdremexec to the other node using the sap_sios_cluster_connector-main with the option start_internal
                        # and we add the blocking / waitOnPid stuff to make it safe and secure
                        # We start the instance on the next node and wait for the lcdremexec to complete.  This locks out duplicates
                        # while the remove/restore are in progress
                        dbgConnector "[Child_2] Running cd /tmp; $lkroot/bin/lcdremexec -d $nod $sap::APPBIN/sap_sios_cluster_connector-main fra --res $res --act start_internal on $me";
                        system "cd /tmp; $lkroot/bin/lcdremexec -d $nod $sap::APPBIN/sap_sios_cluster_connector-main fra --res $res --act start_internal";
                        $actRet=($? >> 8);

                        # If the flag file is left over, then here is the log message you want to look for in the lifekeeper.log file
                        # If you don't see it, then the lcdremexec was killed by forceumount or some other thing
                        dbgConnector "[Child_2] +++++ Finished lcdremexec -d $nod done +++++ on $me ";
                        $sapObj->SAPPROTECTION($savedProtectionLevel);
                        $sapObj->SAPUpdateLKInfoField();
                        $currPLevel=$sapObj->SAPPROTECTION();
                        dbgConnector "[Child_2] currentLevel=$currPLevel savedProtection=$savedProtectionLevel on $me";
                        system ("rm -vf $tmpfilename ");
                        dbgConnector "End of migrate on $me";
                } # End of migrate

                # Keep up, we are still in the child process.  Every thing above could be a routing fraChildWorker
                dbgConnector "[Child_2] Action $act for tag $tag ended with $actRet on $me ";
                # Once we get here, we also have to remove the filename created by the parent
                if ( -f $filename ){
                        dbgConnector "[Child_2] removing the file $filename on $me";
                        system "rm -vf $filename";
                } else{
                        dbgConnector "[Child_2] WARNING: The file $filename does not exist on $me";
                }
                # The child is finally done; exit 0 (or we could actually exit with actRet.
                # Due to testing success we need to stick with 0
                exit(0)
        } # End of Child_2

        # Parent
        dbgConnector "[Parent] Forked $pid from parent $$ to perform fra $act on $me";
        dbgConnector "[Parent] creating filename $filename for fra $act (res=$res, tag=$tag) and saving pid $pid on $me ";
        open SAPWORKFILE,">$filename";
        print SAPWORKFILE "$pid\n";
        close SAPWORKFILE;
        my $childRet=0;
        dbgConnector "[Parent] Finished creating filename $filename for fra $act (res=$res, tag=$tag) on $me ";
         if ( $waitOnPid ){
                dbgConnector "[Parent] was called for $actSaved and $act with a wait for child flag, waiting on $pid on $me";
                waitpid($pid,0);
                $childRet=$?;
                dbgConnector "[Parent] The wait for $pid is over ( $actSaved and $act ) on $me";
         }
        dbgConnector "[Parent] Finished fra $act (res=$res, tag=$tag) and exiting ($childRet) on $me ";
        return 0;
}

####
#  Description
#       - Function to return the next highest priority alive node, or myself.
#       - If called for a TYPE_CS resource with SAPENQ_VERSION=2 or
#       - a TYPE_NEW_ERS resource with SAPENQREP_VERSION=2, this routine
#       - will attempt to not return the node where the corresponding
#       - enqueue resource is running if another live node is available.
#       - e.g., If ASCS with ENSA2 is running on NodeA, ERSv2 is running
#       - on NodeB, and NodeC is available to failover to, fraNextNode
#       - will return NodeC even if NodeB has higher priority for the
#       - ASCS hierarchy.
#  Input
#       - res - The cluster ID of the SAP resource that migrate is being called for
#  Return
#       - next node available for resource switchover
#       - myself - no other alive nodes exist, or $res was empty
#
sub fraNextNode
{
        my $func = (caller(0))[3];
        dbgConnector "=> Trace:Entering $func on $me";

        my $self = shift;
        my $res = shift;

        if (!ref($self) || ($self->SAPLKSAPLK() eq '')) {
                dbgConnector "Unable to get resource info, returning \"$me\" in $func on $me";
                dbgConnector "<= Trace:Leaving $func on $me";
                return ($me);
        }

        if (!defined($res) || ($res eq '')) {
                $res = $self->SAPLKSAPID();
                dbgConnector "No resource cluster ID specified, setting res=$res in $func on $me";
        }

        my @viableMigrationTargets = ();

        my $msg = getMsgFromID("fra", $res);
        # Node resource is currently ISP on, will be "NotActive" if not ISP anywhere.
        my $currentNode = (split /:/, $msg)[2];
        # Systems resource has been extended to, ordered by priority.
        my @priSortedEqvSys = split /,/, (split /:/, $msg)[3];

        # In certain cases where there is not a "good" choice of where to
        # migrate the resource to, we will by default either return:
        # 1. The node where the resource is currently ISP, or
        # 2. The highest priority node that the resource has been extended to
        #    if it is not ISP anywhere.
        my $defaultBestChoice = (($currentNode eq "NotActive") ? @priSortedEqvSys[0] : $currentNode);
        dbgConnector "Setting defaultBestChoice=$defaultBestChoice in $func on $me";

        my $tag = $self->SAPTAG();

        # Loop over nodes that the resource has been extended to,
        # in failover priority order and add nodes to the @viableMigrationTargets
        # array, excluding nodes where:
        # 1. The resource is currently ISP
        # 2. The node is not alive
        # 3. The canfailover script indicates that failover to that node
        #    would be unsuccessful
        foreach my $node (@priSortedEqvSys) {
                dbgConnector "Checking node => $node";
                if ($node eq "$currentNode") {
                        dbgConnector "Skipping $node as a possible migration target since resource $tag is currently ISP there";
                        next;
                }
                if (InspectCluster($node) != 0) {
                        dbgConnector "Skipping $node as a possible migration target since it is not alive";
                        next;
                }
                system("$lkroot/lkadm/bin/canfailover -t $tag -s $node --ignore-flags");
                if ($? != 0) {
                        dbgConnector "Skipping $node as a possible migration target since the script \'$lkroot/lkadm/bin/canfailover -t $tag -s $node --ignore-flags\' returned 1 on $me";
                        next;
                }
                push @viableMigrationTargets, $node;
        }

        if (!@viableMigrationTargets) {
                dbgConnector "No viable migration targets were found, returning $defaultBestChoice";
                dbgConnector "=> Trace:Leaving $func on $me\n";
                return ($defaultBestChoice);
        }

        my $viableMigrationTargetsCSV = join ", ", @viableMigrationTargets;
        dbgConnector "Viable migration targets found for resource $tag: $viableMigrationTargetsCSV";

        # If this is a TYPE_CS resource running ENSA2 or a TYPE_NEW_ERS
        # resource running ERSv2 and more than one viable migration node
        # was found, we will attempt to further exclude the node where
        # the corresponding enqueue resource is currently ISP.
        # This is a requirement of the S/4-HA-CLU-1.0 SAP certification.
        my $type = $self->SAPINSTTYPE();
        my $isENSA2 = (($type == $TYPE_CS) && ($self->SAPENQ_VERSION() == 2));
        my $isERS2 = (($type == $TYPE_NEW_ERS) && ($self->SAPENQREP_VERSION() == 2));
        if (($isENSA2 || $isERS2) && (@viableMigrationTargets > 1)) {
                my $corrEnqResTag = $self->CorrespondingEnqResourceTag();
                if (!defined($corrEnqResTag) || ($corrEnqResTag eq '')) {
                        dbgConnector "No corresponding enqueue resource found for resource $tag, not attempting to avoid corresponding enqueue resource";
                } else {
                        dbgConnector "Resource $corrEnqResTag was found to be the enqueue resource corresponding to $tag, attempting to exclude the node where $corrEnqResTag is currently ISP";
                        my @filteredViableMigrationTargets;
                        foreach my $node (@viableMigrationTargets) {
                                next if ($self->IsTagISP($corrEnqResTag, $node));
                                push @filteredViableMigrationTargets, $node;
                        }
                        my $filteredViableMigrationTargetsCSV = join ", ", @filteredViableMigrationTargets;
                        dbgConnector "After attempting to exclude the node where $corrEnqResTag is currently ISP, the remaining viable migration targets for resource $tag are: $filteredViableMigrationTargetsCSV";
                        dbgConnector "=> Trace:Leaving $func on $me\n";
                        return (@filteredViableMigrationTargets ? @filteredViableMigrationTargets : ($defaultBestChoice));
                }
        }

        # If this was not a TYPE_CS or TYPE_NEW_ERS resource running ENSA2/ERSv2 or
        # there was only one viable choice for the migration node, then no extra
        # filtering was needed. In this case we can just return @viableMigrationTargets.
        dbgConnector "=> Trace:Leaving $func on $me\n";
        return @viableMigrationTargets;
}

####
#  Description
#       - Connector hcc ("HA Configuration Check") function
#       - This routine performs HA configuration checks and
#       - writes the results to the output file. Each test
#       - contributes a line in the output file of the form:
#       - CHK-STATE:CHK-CATEGORY:CHK-DESCRIPTION:CHK-COMMENT
#  Input
#       - out - Output file
#       - sid - SAP system ID
#       - ino - Instance number
# Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
####
sub hcc($$$) {
        my $func = (caller(0))[3];
        dbgConnector "=> Trace:Entering $func on $me\n";
        my ($out, $sid, $ino) = @_;

        # The basic format of the output for sapha version 2
        # (2015 certification version)
        # CHK-STATE:CHK-CATEGORY:CHK-Description:CHK-Comment where
        # CHK-STATE (0-2):CHK-CATEGORY (0-3)
        # The Description and comment fields are free from
        # CHK-STATE 0 - success, 1 - warning, 2 - error
        # CHK-CATEGORY 0 - SAP config, 1 - sap state, 2 - ha config, 3 - ha state

        # if the file exists, we'll delete and it will overwrite it
        if (-e $out) { unlink $out; }

        my @output = ();
        my $sapState = 0;
        my $comment = "Cluster node $me is operational";
        my $ret = LK::sys_getstate('', $me);
        my $lkInitDone = (-f "$ENV{'LKROOT'}/config/LK_INITDONE");
        if ($ret != $LK::ALIVE) {
                $comment = "Cluster node $me is not operational";
                $sapState = 2;
        } elsif (!$lkInitDone) {
                $comment = "LifeKeeper initialization not complete on cluster node $me";
                $sapState = 2;
        }
        push @output, "$sapState:3:HA State Check:$comment";
        my $sapObj;
        my $tag;
        # if LK is running and operational, perform the additional
        # HA checks
        if (($ret == $LK::ALIVE) && $lkInitDone) { #extended hcc check for the tag and comm paths
                my @matches = getResFromSINO($sid, $ino); #returns the matching instance name
                foreach my $res (@matches){
                        dbgConnector "In $func preparing to call new to generate an object for use in $func on $me for the checking of ISP in the cluster\n";
                        $sapObj = sap->new("$sid", "$res", '', '');
                        if (!ref($sapObj) || $sapObj->SAPLKSAPLK() eq '') {
                                dbg "In $func and unable to create a new object for checking the resource state on $me";
                                dbgConnector "=> Trace:Leaving $func on $me\n";
                                # Return 1 (no success/internal error) since we can't create the object.
                                # HACheckFailoverConfig will report "HA Checks Failed" in this case.
                                return 1;
                        }

                        $tag = $sapObj->SAPTAG();
                        dbgConnector "Found tag $tag in $func on $me\n";

                        $ret = $sapObj->isTagClusterISP($tag);
                        dbgConnector "isTagClusterISP returned $ret for the tag $tag in $func on $me\n";

                        # The resource is not ISP on any node when ret=0
                        # This has to be set as an SAP error = 2
                        if ($ret == 0) {
                                $sapState = 2;
                                $comment = "SAP instance ${sid}${ino} ($tag) is not running";
                        } else {
                                $sapState = 0;
                                $comment = "SAP instance ${sid}${ino} ($tag) is running";
                        }

                        push @output, "$sapState:1:SAP Resource Instance Check:$comment";
                }
                # Check if the comm paths are functional
                $sapState = 0;
                $comment = "All cluster nodes are alive and all comm paths are functional";
                system("$lkroot/lkadm/bin/test_lk");
                if ($? != 0) {
                        $comment = "One or more cluster nodes or comm paths is currently down";
                        $sapState = 2;
                }
                push @output, "$sapState:2:HA Config Check:$comment";
        }
        my $msg = join "\n", @output;
        my $rc = FileHandler($out, 1, "$msg", $sapObj);
        dbgConnector "=> Trace:Leaving $func on $me\n";
        return $rc;
}

####
#  Description
#       - Checks to see if the specified tag is ISP somewhere in the LK
#       - cluster
####
#  Input:
#       - self - reference to the object
#       - tag - the tag to check (ours or another resource (NFS, Oracle)
# Outputs:
#       1 - ISP on some node in the cluster
#       0 - not ISP anywhere in the cluster
#
sub isTagClusterISP
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";
        my $self=shift;
        my $tag=shift;
        my $ret=0;
        my %eqvs;
        my %fulleqvs;
        $self->GetLKEquiv(\%eqvs, $tag, \%fulleqvs);
        foreach my $sys (keys %fulleqvs){
                my $eqvTag=$fulleqvs{$sys}{'Tag'};
                dbg "[input tag=$tag] Checking if the equivalent tag $eqvTag on $sys is ISP on $sys. cmd running on $me\n";
                next if ($self->IsTagISP($eqvTag, $sys) == 0 );
                $ret=1;
        }
        dbg "The check on equivalent systems for IsTagISP returned [$ret] on $me\n";
        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;
}

####
#  Description
#       - Connector gvi ("Get Version Info") function
#       - This routine prints the version info of the cluster
#       - software and the HA connector, as well as a link to
#       - the cluster product documentation and an SAP note.
#       - The information written to the output file has the format:
#       - HASCRIPTCO-VERS
#       - HAPROD
#       - HAPROD-SAP
#       - HAPROD-DOC
#  Input
#       - out - Output file
#  Returns
#       - 0 - Success
#       - 1 - No success / internal error
#       - 2 - Bad parameters
####
sub gvi($) {
        my $func = (caller(0))[3];
        dbgConnector "=> Trace:Entering $func on $me\n";
        my ($out) = @_;
        #if the file exists, we'll delete and it will override it
        if (-e $out) { unlink $out; }
        my @PKGS=($CORE_PKG, $SSCC_PKG);
        my @allTags=sap::GetAllSAPTags();
        my $objTag;
        my $sapObj;
        my $rpmError=0;
        if (@allTags){ # if we found a single tag, then use it
                $objTag=$allTags[0];
                $sapObj=sap->new("","",$objTag,'');
        }
	#bug7622 - we now support version 3 of HA
        my $hascriptver=3;
        my @vString=();
        push @vString, $hascriptver;
        foreach my $lkpkg (@PKGS){
                chomp (my $ver=`rpm -q --qf '%{VERSION}-%{RELEASE}' $lkpkg 2>/dev/null`);
                if ($ver eq '' ){ $rpmError=1; next; }  #capture rpm errors
                push @vString,"\"SIOS Protection Suite for Linux\" $lkpkg $ver";
                next if ($lkpkg !~ /$CORE_PKG$/ );
        }

        # bug7678 - Change the doc page to point to a static link.
        push @vString, "$SAP_DOCS_PAGE";
        my $msgString=join "\n", @vString;
        ($rpmError) && dbgConnector "Error finding rpm information on one or more packages on $me\n";
        my $rc=FileHandler($out, 1, "$msgString", $sapObj);
        # FileHandler should take care of making the file with the correct permissions
        # However, 12/2015, we encountered an SAP HA Tool use case where the permissions were
        # not correct.  This will make sure the file can be read by both sapadm and root
        system ("chmod a+rw $out");
        dbgConnector "=> Trace:Leaving $func on $me\n";
        return $rc;
}

####
#  Description
#       - This function uses the sapcontrol utility with the -function GetSystemUpdateList
#       - to determine if a Rolling Kernel Upgrade is in progress for any
#       - of the SAP Instances of a given SAP SID
#       - The sapcontrol utility will check all processes for the SAP Instance
#       - In order to use this utility startsapsrv must be running first
#       - but, it cannot be started by the function if it is not running
#       - because this will cause the Rolling Kernel Upgrade (RKU) to fail.
#       - The RKU is detected by the presence of the Rolling Kernel Switch (RKS)
####
#  Input:
#       - [RunLocal] - run on this or another server
#       - [display] - populate object for sap-display
####
#  Output:
#       - sapcontrol output on status
####
#  Returns
#       - 0 - not running
#       - 1 - running
sub isSAPRKSRunning
{
        my $func = (caller(0))[3];
        dbg "=> Trace:Entering $func on $me\n";

        my $self=shift;
        my $action=shift;       #added for consistency with all other sapcontrol/kit interfaces
        my $RunLocal=shift;
        my $nodisplay=shift;
        my $Instance=$self->SAPINST();
        my $InstanceNum=$self->SAPINSTNUM();
        $nodisplay=($nodisplay eq '' ) ? 0 : $nodisplay;
        $RunLocal=($RunLocal eq '' ) ? 0: $RunLocal;
        my $sapadmin=$self->SAPADMIN();
        my $tmpfile=$self->SAPTMPFILE();
        my $sid=$self->SAPSID();
        my $Type=$self->SAPINSTTYPE();
        my $Tag=$self->SAPTAG();
        my $ret=0;
        my $ignoreByInstanceNum=$ENV{'SAP_RKS_CHECK_IGNORE_' . $InstanceNum};
        if ( $SAP_RKS_CHECK_IGNORE || $ignoreByInstanceNum ){
                msg ('disabled', 'RKS checks', $func, $Tag, $me);
                dbg "<= Trace:Leaving $func on $me\n";
                return 0;
        }

        dbg "Entry to $func: RunLocal=$RunLocal, Instance=$Instance, InstanceNum=$InstanceNum, sid=$sid and Type=$Type nodisplay=$nodisplay on $me\n";
        if ($Type == $TYPE_ERS && $RunLocal == 0 ){
                $ret=$self->SAPRemExec($func, $nilValue, $nodisplay);
                dbg "<= Trace:Leaving $func on $me\n";
                return $ret;
        }

        if ($InstanceNum eq '' ){
                error ('valueempty', 'InstanceNum', $func, $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1; #assume it is running?
        }

        # 0. Make sure basic NFS shares are accessible
        my $retNFS=$self->NFSCheck();
        if ( $retNFS == 0 ){
                error('nonfs', $me );
                dbg "<= Trace:Leaving $func on $me\n";
                return 1; # #should we return 1 or 0?
        }
        $self->LocalNFSMounts();


        # 1. We cannot start the SAPSRV during an
        # Rolling Kernel Switch (RKS) check b/c it
        # causes the Rolling Kernel Upgrade (RKU) to fail.
        # Our only workaround is to wait until the RKU
        # has restarted the SAPSRV, which it stops during the
        # upgrade process.  Upon the wait loop completion,
        # if it is still not running then we assume that
        # no RKU is in progress.  Otherwise, if it is running
        # or has restarted in the interval, we assume
        # that an RKU is in progress and proceed
        # to check for the RKS
        my $waitRKSMax=$self->SAPRKSWAIT();
        my $waitRKSCount=0;
        my $rksInProgress=0;
        while ($waitRKSCount <= $waitRKSMax ){
                if ( ! $self->SAPSRVUtils('status') ){
                        dbg "Waiting to see if the SAPSRV will restart in $waitRKSMax ($waitRKSCount)";
                        $waitRKSCount+=5;
                        sleep 5;
                }else{
                        $rksInProgress=1;
                        last;
                }
        }
        if ( ! $rksInProgress ){
                # if the sapstartsrv command is still not running
                # then it is likely that it has died, rather than
                # simply having been stopped by the RKU process.
                # In which case, we will eventually "assume" no RKU is in
                # progress
                # perform one last check to be sure
                if ( ! $self->SAPSRVUtils('status') ){
                        dbg "<= SAPSRVUtils in $func indicates that sapstartsrv was not running on $me\n";
                        dbg "<= Trace:Leaving $func on $me\n";
                        return 0;
                }
        }

        # 2. Get the VIP
        my $vip=$self->SAPVIP();

        # 3. GetSystemUpdateList
        # This command requires that the SAPSRV be running, but during
        # a Rolling Kernel Upgrade (RKU) we cannot actually start the SAPSRV
        # *if* we get to this point, we assume the SAPSRV is running and we
        # issue the sapcontrol -function command to check for the
        # Rolling Kernel Switch (RKS)
        my $sapSystemUpdateList="$SAP_CONTROL_CMD -host $vip -nr $InstanceNum -function GetSystemUpdateList";
        my $resfile=$self->SAPRESULTFILE();
        msg ('userstatus', $sapSystemUpdateList, $me );

        my @out=grep /status/i, $self->NFSCheckBeforeCmd("su - $sapadmin -c \"sh -c 'LANG=C $sapSystemUpdateList 2>&1; echo \\\$? > $resfile' \"");
        if (@out){
                dbg "<= GetSystemUpdateList returned @out, setting ret to $ret on $me\n";
                $ret=1;
        }
        chomp (my $cmdRet=`cat $resfile`);
        msg ('commandReturned', "SID=$sid, tag=$Tag on $sapSystemUpdateList", $cmdRet, $me);
        unlink $resfile;

        dbg "<= Trace:Leaving $func on $me\n";
        return $ret;

}

####
#  Description
#       - Given the ID of a resource ($res) and a fra action ($act)
#       - which is either 'start', 'stop', or 'migrate', this routine
#       - determines whether the flag file $res.$act exists and contains
#       - a valid, running PID.  If the file exists but the PID it contains
#         no longer belongs to a relevant LifeKeeper process, it is deleted.
#  Input
#       - res - Cluster ID of the resource (SAPID-<SID>-<INST>)
#       - act - fra action to check for (start, stop, or migrate)
#  Returns
#       - 0 - No file found, or contains a PID that is not associated to a LifeKeeper process
#       - 1 - File found and contains a PID associated to a running LifeKeeper process
####
sub checkFileandPid
{
        my $func = (caller(0))[3];
        my $res=shift;
        my $act=shift;
        dbgConnector "=> Entering $func for $res and $act on $me";

        my @fileList = ("$sap::APPSTORE/$res.$act");

        # Since the beginning of a migration action will forcibly terminate all instance
        # processes without using the remove script, we do not allow a migrate action to
        # be performed if any action is currently in-progress. Doing so could allow this
        # forcible instance termination to interfere with the action already in-progress.
        if ($act =~ /migrate/i) {
                push @fileList, "$sap::APPSTORE/$res.start";
                push @fileList, "$sap::APPSTORE/$res.stop";
        }

        my $ret = 0;
        my $pid;
        my $psOut;
        my @filteredPSOut;
        my $psPid;
        foreach my $filename (@fileList) {
                next if (! -f $filename);
                chomp($pid = `cat $filename 2>/dev/null`);
                dbgConnector "[$func] File $filename contains the pid $pid on $me";

                # If the PID is not valid, delete the file and move to the next one.
                if (($pid <= 1) || ($pid eq '')) {
                        dbgConnector "[$func] File $filename does not contain a valid pid (pid=$pid), removing $filename on $me";
                        unlink $filename;
                        next;
                }

                chomp($psOut = `ps --no-headers -wwf -p $pid`);

                # Determine whether the ps output associated to the PID corresponds to a relevant LifeKeeper process.
                if ($act =~ /start/i) {
                        @filteredPSOut = grep /$SSCC_SCRIPT/ || /lifekeeper/i || /restore/, $psOut;
                } elsif ($act =~ /stop/i) {
                        @filteredPSOut = grep /$SSCC_SCRIPT/ || /lifekeeper/i || /remove/, $psOut;
                } elsif ($act =~ /migrate/i) {
                        @filteredPSOut = grep /$SSCC_SCRIPT/ || /lifekeeper/i || /restore/ || /remove/, $psOut;
                }

                ($psPid = $filteredPSOut[0]) =~ s/^\S+\s*(\d+)\s+.*$/$1/g;
                if (($psPid <= 1) || ($psPid eq '')) {
                        dbgConnector "[$func] File $filename does not contain a valid/running pid (filepid=$pid, psPid=$psPid, [psOut=$psOut]), removing $filename on $me";
                        unlink $filename;
                        next;
                }

                dbgConnector "[$func] File $filename contains a valid/running pid (filepid=$pid, psPid=$psPid, [psOut=$psOut]), setting ret=1";
                $ret = 1;
        }
        if ($ret == 0) {
                dbgConnector "[$func] No files with names [fileList=@fileList] contained valid/running pid's";

                # If the remove or restore script is currently running but there is no *.stop or *.start file (which will
                # be the case when remove/restore is initiated via perform_action), we do not allow a migrate action to proceed.
                # This is because the migrate code will forcibly kill all instance processes, which could interfere
                # with the remove/restore already in-progress.
                my $tag = getTagFromID($res); 
                if (($act =~ /migrate/i) && ((LK::flg_test('', "$SAP_FLAG_REMOVE_IN_PROGRESS${tag}") == 0) || (LK::flg_test('', "$SAP_FLAG_RESTORE_IN_PROGRESS${tag}") == 0))) {
                        dbgConnector "[$func] A restore or remove action is currently in-progress for resource $tag, setting ret=1. Reattempt the migrate action after this action completes.";
                        $ret = 1;
                }
        }

        dbgConnector "<= Leaving $func for $res and $act with return code $ret on $me";
        return $ret;
}

################################################################################
##############################--- End of sap.pm  ---############################
################################################################################

1;
################################################################################
##########################--- ********************** ---########################
################################################################################


################################################################################
##########################--- ********************** ---########################
################################################################################



# TODO LIST:
#       Add support for restart/recover and failover counters
#       Add support for Failures/Interval
#       Determine if GetRealIP needs IPv6 updates
#       FindActualPath Integration with the core
#                       - Consider merging this NFS logic into core/mounted
#                       - via additional options
#                       - Example: mounted -n <dir> -q <check shared>
#       LocalNFSMounts Integration with NAS
#                       - LifeKeeper really needs the concept of Active/Active within a
#                       - single resource.  This code should really be removed and
#                       - an Active/Active Network Attached Storage (NAS)
#                       - resource incorporated to handle making sure the mounts remain
#                       - mounted
#       Code Refactor
#                       -refactor SAP HA code

# Dev/Test Notes On Order of Operations:
#       * All Application Servers must stop before a CI/Central Service
#       *       is taken OSU and restarted after the CI/Central Services
#       *       are restarted
#

